/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.dao.impl;

import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.dao.impl.abstr.AbstractDAOImpl;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.stereotype.Repository;

import javax.sql.DataSource;
import java.sql.Types;
import java.time.Instant;
import java.time.temporal.ChronoUnit;

import java.util.LinkedList;
import java.util.List;
import java.util.Objects;

@SuppressWarnings("SqlResolve")
@Repository
public class OrganDonorDAOImpl extends AbstractDAOImpl implements OrganDonorDAO {

    private final String TABLE_NAME = "OrganDonor";

    @Autowired
    public OrganDonorDAOImpl(DataSource organDonorDataSource) {
        super(organDonorDataSource);
    }

    @Override
    public Integer createOrganDonorRegistration(OrganDonorVO organDonorVO, Instant nowTimestamp) throws DAOException {
        assignFieldsForNew(organDonorVO, nowTimestamp);

        try {
            final String sql = "INSERT INTO `OrganDonor` " +
                    "(`PersonIdentifier`, `ModifiedDateTime`, `ValidFrom`, `ValidTo`, `PermissionType`, `Heart`, `Kidneys`, `Lungs`, `Corneas`, `Liver`, `SmallIntestine`, `Pancreas`, `Skin`, `RelativeAcceptanceRequired`, `Version`) " +
                    "VALUES (:PersonIdentifier, :ModifiedDateTime, :ValidFrom, :ValidTo, :PermissionType, :Heart, :Kidneys, :Lungs, :Corneas, :Liver, :SmallIntestine, :Pancreas, :Skin, :RelativeAcceptanceRequired, :Version)";

            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", organDonorVO.getPersonIdentifier(), Types.VARCHAR)
                    .addValue("ModifiedDateTime", java.sql.Timestamp.from(organDonorVO.getModifiedDateTime().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ValidFrom", java.sql.Timestamp.from(organDonorVO.getValidFrom().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("ValidTo", java.sql.Timestamp.from(organDonorVO.getValidTo().truncatedTo(ChronoUnit.SECONDS)), Types.TIMESTAMP)
                    .addValue("PermissionType", organDonorVO.getPermissionType().getDatabaseValue(), Types.VARCHAR)
                    .addValue("Heart", organDonorVO.getHeart(), Types.BIT)
                    .addValue("Kidneys", organDonorVO.getKidneys(), Types.BIT)
                    .addValue("Lungs", organDonorVO.getLungs(), Types.BIT)
                    .addValue("Corneas", organDonorVO.getCorneas(), Types.BIT)
                    .addValue("Liver", organDonorVO.getLiver(), Types.BIT)
                    .addValue("SmallIntestine", organDonorVO.getSmallIntestine(), Types.BIT)
                    .addValue("Pancreas", organDonorVO.getPancreas(), Types.BIT)
                    .addValue("Skin", organDonorVO.getSkin(), Types.BIT)
                    .addValue("RelativeAcceptanceRequired", organDonorVO.getRelativeAcceptanceRequired(), Types.BIT)
                    .addValue("Version", organDonorVO.getVersion(), Types.INTEGER);
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            namedParamJdbcTemplate.update(sql, parameterSource, generatedKeyHolder);
            final int pid = Objects.requireNonNull(generatedKeyHolder.getKey()).intValue();
            organDonorVO.setOrganDonorPID(pid);
            return pid;
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred inserting Organdonor", e);
        }
    }

    @Override
    public OrganDonorVO getOrganDonorRegistration(String personIdentifier, Instant nowTimestamp) throws DAOException {
        List<OrganDonorVO> organDonorVO = getByPersonIdAndDate(personIdentifier, nowTimestamp);
        return organDonorVO.isEmpty() ? null : organDonorVO.get(0);
    }

    @Override
    public boolean existsOrganDonorRegistrationRecords(String personIdentifier) throws DAOException {
        try {
            final String sql = "SELECT COUNT(*) FROM OrganDonor " +
                    "WHERE PersonIdentifier = :PersonIdentifier";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                    .addValue("PersonIdentifier", personIdentifier, Types.VARCHAR);
            final long count = Objects.requireNonNull(namedParamJdbcTemplate.queryForObject(sql, parameterSource, Long.class));
            return count > 0;
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred counting OrganDonor", e);
        }
    }

    @Override
    public void updateOrganDonorRegistration(Integer organDonorPID, OrganDonorVO organDonorVO, Instant nowTimestamp) throws DAOException {
        updateValidToTimestamp(getIdColumnName(), organDonorPID, nowTimestamp);
        createOrganDonorRegistration(organDonorVO, nowTimestamp);
    }

    @Override
    public void deleteOrganDonorRegistration(Integer pidToUpdate, Instant nowTimestamp) throws DAOException {
        updateValidToTimestamp(getIdColumnName(), pidToUpdate, nowTimestamp);
    }

    @Override
    public void deleteOrganDonorRegistrationRecords(String personIdentifier) throws DAOException {
        try {
            String sql = "DELETE FROM " + getTableName() + " WHERE PersonIdentifier = :personId";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("personId", personIdentifier, Types.VARCHAR);
            executeUpdate(sql, parameters);
        } catch (DAOException e) {
            throw new DAOException("An error occurred deleting from " + getTableName(), e);
        }
    }

    @Override
    public LinkedList<String> getallPersonsFromODR() throws DAOException {
        try {
            final String sql = "SELECT DISTINCT PersonIdentifier FROM OrganDonor";
            return new LinkedList<>(namedParamJdbcTemplate.query(sql, (resultSet, i) -> resultSet.getNString("PersonIdentifier")));
        } catch (DataAccessException | NullPointerException e) {
            throw new DAOException("An error occurred fetching all rows from OrganDonor", e);
        }
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }


    @Override
    protected RowMapper getDefaultRowMapper() {
        return new OrganDonorRowMapper();
    }

}
