/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.dao.impl.abstr;

import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;

import javax.sql.DataSource;
import java.sql.Types;
import java.time.Instant;
import java.time.Year;
import java.time.ZoneOffset;
import java.time.temporal.ChronoUnit;
import java.util.List;

public abstract class AbstractDAOImpl {

    public static final Instant VALID_TO_DEFAULT = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;

    protected AbstractDAOImpl(DataSource dataSource) {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }

    public abstract String getTableName();
    protected abstract RowMapper getDefaultRowMapper();

    protected String getIdColumnName() {
        return getTableName() + "PID";
    }

    protected <T> List<T> getByPersonIdAndDate(String personIdentifier, Instant date) throws DAOException{
        try {
            final String sql = "SELECT * FROM " + getTableName() + " o where o.PersonIdentifier = :personIdentifier" +
                    " AND o.ValidFrom <= :date AND o.ValidTo > :date";
            final MapSqlParameterSource parameterSource = new MapSqlParameterSource().
                    addValue("personIdentifier", personIdentifier, Types.VARCHAR)
                    .addValue("date", java.sql.Timestamp.from(date), Types.TIMESTAMP);
            final List<T> list = queryForList(sql, parameterSource, getDefaultRowMapper());
            return list;
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + getTableName() + " by personid and date", e);
        }
    }

    protected void updateValidToTimestamp(String idColumnName, Integer idList, Instant timestamp) throws DAOException {
        if (idList == null) {
            return;
        }

        try {
            final String tableName = getTableName();
            String sql = "UPDATE " + tableName + " SET ValidTo = :ValidTo WHERE " + idColumnName + " IN (:Ids)";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("ValidTo", java.sql.Timestamp.from(timestamp), Types.TIMESTAMP);
            parameters.addValue("Ids", idList, Types.INTEGER);
            executeUpdate(sql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred updating valid-to timestamps", e);
        }
    }

    protected <T> T fetchSingleRowMatchingSingleColumn(String columnName, String filterValue) throws DAOException {
        final MapSqlParameterSource parms = new MapSqlParameterSource().addValue("val", filterValue, Types.VARCHAR);

        List<T> result = queryForList("SELECT * FROM " + getTableName() + " WHERE " + columnName + "=:val", parms, getDefaultRowMapper());
        if (result.size() == 0) {
            return null;
        } else if (result.size() > 1) {
            throw new DAOException("Multiple rows were found in table " + getTableName()
                    + " looking in column " + columnName + " for the value " + filterValue);
        } else {
            return result.get(0);
        }
    }

    protected <T> List<T> queryForList(String sql, SqlParameterSource parms, RowMapper rowMapper) throws DAOException {
        try {
            final List<T> list = namedParamJdbcTemplate.query(sql, parms, rowMapper);
            return list;
        } catch (Exception e) {
            throw new DAOException("An error occurred executing query: " + sql, e);
        }
    }

    protected void executeUpdate(String updateSql, SqlParameterSource parameters) throws DAOException {
        try {
            namedParamJdbcTemplate.update(updateSql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred executing update: " + updateSql, e);
        }
    }

    protected void assignFieldsForNew(OrganDonorVO organDonorVO, Instant nowTimestamp) {
        if (nowTimestamp != null && organDonorVO.getValidTo() == null && organDonorVO.getModifiedDateTime() == null) {
            organDonorVO.setValidFrom(nowTimestamp.truncatedTo(ChronoUnit.SECONDS));
            organDonorVO.setModifiedDateTime(nowTimestamp.truncatedTo(ChronoUnit.SECONDS));
        }
        if (organDonorVO.getValidTo() == null) {
            organDonorVO.setValidTo(AbstractDAOImpl.VALID_TO_DEFAULT.truncatedTo(ChronoUnit.SECONDS));
        }
    }
}
