/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.util;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.PropertySource;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.time.Duration;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Component
@PropertySource("classpath:application.properties")
public class Properties {
    private static final Logger log = LogManager.getLogger(Properties.class);

    @Value("${datasource.odr.jndi-name:}")
    private String dataSourceOdrJNDIName;

    @Value("${datasource.stamdata.jndi-name:}")
    private String dataSourceStamdataJNDIName;

    @Value("${app.endpoint:}")
    private String appEndpoint;

    @Value("${dcc.endpoint:}")
    private String dccEndpoint;

    @Value("${minlog.read-activity-text:}")
    private String minLogReadActivityText;
    
    @Value("${minlog.create-activity-text:}")
    private String minLogCreateActivityText;
    
    @Value("${minlog.delete-activity-text:}")
    private String minLogDeleteActivityText;
    
    @Value("${minlog.update-activity-text:}")
    private String minLogUpdateActivityText;

    @Value("${schemavalidation.validate-requests:}")
    private boolean schemavalidationValidateRequests;

    @Value("${schemavalidation.validate-responses:}")
    private boolean schemavalidationValidateResponse;

    @Value("${schemavalidation.fail-on-response-error:}")
    private boolean schemavalidationFailOnResponseError;

    @Value("${cprexists.url:#{null}}")
    String cprExistsUrl;

    @Value("${cprexists.validationlevel}")
    private String cprExistsValidationLevel;

    @Value("${cprexists.minage:15}")
    private int cprExistsMinAge;

    @Value("#{'${cprexists.inactive.status:30,50,60}'.split(',')}")
    private List<Integer> inactiveStatusList;
    
    @Value("${cprexists.maxTotalConnections:200}")
    private int maxTotalConnections;

    @Value("${cprexists.defaultMaxConnectionsPerRoute:20}")
    private int defaultMaxConnectionsPerRoute;

    @Value("${whitelisted.level3.cvrs:}")
    private String whitelistedLevel3Cvrs;

    @Value("${allowed.idws.audience:https://fsk}")
    private String allowedIdwsAudience;
    @Value("${nas.app.name}")
    private String appName;

    @Value("${nas.app.shortName}")
    private String appShortName;

    @Value("${nas.fail.threshold}")
    private String nasFailThreshold;

    @Value("${nas.connect.timeout.millis}")
    private String nasConnectTimeoutMillis;

    @Value("${nas.read.timeout.millis}")
    private String nasReadTimeoutMillis;

    @Value("${nas.max.total.connections}")
    private String nasMaxTotalConnections;

    @Value("${nas.default.max.connections.per.route}")
    private String nasDefaultMaxConnectionsPerRoute;

    @Value("${nas.service.url}")
    private String nasServiceUrl;

    @Value("${nas.idcard.subject.id.type}")
    private String idcardSubjectIdType;

    @Value("${nas.idcard.subject.id}")
    private String idcardSubjectId;

    @Value("${nas.idcard.subject.name}")
    private String idcardSubjectName;

    @Value("${nas.idcard.system.name}")
    private String idcardSystemName;

    @Value("${nas.sts.endpoint}")
    private String stsEndpoint;

    @Value("${NSP_TEST_FEDERATION:false}")
    private boolean stsTestMode;

    @Value("${nas.sts.keystore}")
    private String stsKeystore;

    @Value("${nas.sts.keystore.password}")
    private String stsKeystorePassword;

    @Value("${jobs.delete.max.duration}")
    private String jobsDeleteMaxDuration;

    @PostConstruct
    private void init() {
        log.info(toString());
    }

    @Override
    public String toString() {
        return "Configuration properties: {" +
                "dataSourceOdrJNDIName=" + getDataSourceOdrJNDIName() +
                ", dataSourceStamdataJNDIName=" + getDataSourceStamdataJNDIName() +
                ", dccEndpoint=" + getDccEndpoint() +
                ", minLogReadActivityText=" + getMinLogReadActivityText() +
                ", minLogCreateActivityText=" + getMinLogCreateActivityText() +
                ", minLogDeleteActivityText=" + getMinLogDeleteActivityText() +
                ", minLogUpdateActivityText=" + getMinLogUpdateActivityText() +
                ", schemavalidationValidateRequests=" + isSchemavalidationValidateRequests() +
                ", schemavalidationValidateResponse=" + isSchemavalidationValidateResponse() +
                ", schemavalidationFailOnResponseError=" + isSchemavalidationFailOnResponseError() +
                ", jobsDeleteMaxDuration=" + getJobsDeleteMaxDuration() +
                ", whitelistedLevel3Cvrs=" + getWhitelistedLevel3Cvrs() +
                '}';
    }

    public String getDataSourceOdrJNDIName() {
        return dataSourceOdrJNDIName;
    }

    public String getDataSourceStamdataJNDIName() {
        return dataSourceStamdataJNDIName;
    }

    public String getDccEndpoint() {
        return dccEndpoint;
    }

    public String getMinLogReadActivityText() {
        return minLogReadActivityText;
    }
    
    public String getMinLogCreateActivityText() {
        return minLogCreateActivityText;
    }
    
    public String getMinLogDeleteActivityText() {
        return minLogDeleteActivityText;
    }
    
    public String getMinLogUpdateActivityText() {
        return minLogUpdateActivityText;
    }

    public boolean isSchemavalidationValidateRequests() {
        return schemavalidationValidateRequests;
    }

    public boolean isSchemavalidationValidateResponse() {
        return schemavalidationValidateResponse;
    }

    public boolean isSchemavalidationFailOnResponseError() {
        return schemavalidationFailOnResponseError;
    }

    public String getCprExistsUrl() {
        return cprExistsUrl;
    }

    public String getCprExistsValidationLevel() {
        return cprExistsValidationLevel;
    }

    public int getCprExistsMinAge() {
        return cprExistsMinAge;
    }
    
    public List<Integer> getInactiveStatusList() {
        return inactiveStatusList;
    }

    public int getMaxTotalConnections() {
        return maxTotalConnections;
    }

    public int getDefaultMaxConnectionsPerRoute() {
        return defaultMaxConnectionsPerRoute;
    }

    public List<String> getWhitelistedLevel3Cvrs() {
        return Arrays.stream(whitelistedLevel3Cvrs.split(",")).map(String::trim).filter(cvr -> !cvr.isEmpty()).collect(Collectors.toList());
    }

    public String getAllowedIdwsAudience() {
        return allowedIdwsAudience;
    }
    public String getAppName() {
        return appName;
    }

    public String getAppShortName() {
        return appShortName;
    }

    public String getNasFailThreshold() {
        return nasFailThreshold;
    }

    public String getNasConnectTimeoutMillis() {
        return nasConnectTimeoutMillis;
    }

    public String getNasReadTimeoutMillis() {
        return nasReadTimeoutMillis;
    }

    public String getNasMaxTotalConnections() {
        return nasMaxTotalConnections;
    }

    public String getNasDefaultMaxConnectionsPerRoute() {
        return nasDefaultMaxConnectionsPerRoute;
    }

    public String getNasServiceUrl() {
        return nasServiceUrl;
    }

    public String getIdcardSubjectIdType() {
        return idcardSubjectIdType;
    }

    public String getIdcardSubjectId() {
        return idcardSubjectId;
    }

    public String getIdcardSubjectName() {
        return idcardSubjectName;
    }

    public String getIdcardSystemName() {
        return idcardSystemName;
    }

    public String getStsEndpoint() {
        return stsEndpoint;
    }

    public boolean getStsTestMode() {
        return stsTestMode;
    }

    public String getStsKeystore() {
        return stsKeystore;
    }

    public String getStsKeystorePassword() {
        return stsKeystorePassword;
    }

    public Duration getJobsDeleteMaxDuration() { return Duration.parse(jobsDeleteMaxDuration); }
}
