/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.slettejob;

import java.time.LocalDateTime;
import java.util.Collections;
import java.util.LinkedList;

import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.PersonSimplifiedDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class DeleteOrganDonorRegistrationJob {
    private static final Logger log = LogManager.getLogger(DeleteOrganDonorRegistrationJob.class);
    private final OrganDonorDAO organDonorDAO;
    private final PersonSimplifiedDAO personSimplifiedDAO;
    private final Properties properties;
    private Boolean lastExecutionSucceeded;
    private LocalDateTime timeOfLastExecution;
    private Exception lastException;
    private LinkedList<String> workQueue;

    @Autowired
    public DeleteOrganDonorRegistrationJob(OrganDonorDAO organDonorDAO, PersonSimplifiedDAO personSimplifiedDAO, Properties properties) {
        this.organDonorDAO = organDonorDAO;
        this.personSimplifiedDAO = personSimplifiedDAO;
        this.properties = properties;

        workQueue = new LinkedList<>();
    }

    public void runDeleteOrganDonorRegistration() {
        log.trace("Deletion Job starting");
        try {
            if (loadWorkQueue()) {
                log.info("Cleanup new workqueue=" +  workQueue.size());
            } else {
                // Start stopuret
                long startMillis = System.currentTimeMillis();

                int hasBeenDeleted = 0;

                while (belowTimeLimit(startMillis)) {

                    String personIdentifier = findNextItemInWorkQueue();

                    if (personIdentifier != null) {
                        if (isPersonDeceasedOneYearAgo(personIdentifier)) {
                            deletePersonFromODR(personIdentifier);
                            hasBeenDeleted++;
                        }
                    } else {
                        break;
                    }
                }

                // Log status efter sletning
                log.info("Cleanup iteration deleted=" + hasBeenDeleted + " remaining=" + workQueue.size() + " done.");
            }

            setRunStatusOk();
        } catch (Exception e) {
            log.error("Deletion Job unable to run because of an error.", e);

            setRunStatusError(e);
        }
        
        log.trace("Deletion Job done");
    }

    private synchronized boolean loadWorkQueue() throws DAOException {
        if (workQueue.isEmpty()) {
            workQueue = organDonorDAO.getallPersonsFromODR();

            Collections.shuffle(workQueue);
            return true;
        }
        return false;
    }

    private synchronized String findNextItemInWorkQueue() {
        return workQueue.size() > 0 ? workQueue.removeFirst() : null;
    }

    private synchronized void setRunStatusOk() {
        timeOfLastExecution = LocalDateTime.now();
        lastExecutionSucceeded = true;
        lastException = null;
    }

    private synchronized void setRunStatusError(Exception e) {
        timeOfLastExecution = LocalDateTime.now();
        lastExecutionSucceeded = false;
        lastException = e;
    }

    private boolean isPersonDeceasedOneYearAgo(String personIdentifier) throws DAOException {
        return personSimplifiedDAO.isPersonDeceasedOneYearAgo(personIdentifier);
    }

    private void deletePersonFromODR(String personIdentifier) throws DAOException {
        log.info("Now deleting Organdonorregistration for person with personId: [" + personIdentifier + "]");
        organDonorDAO.deleteOrganDonorRegistrationRecords(personIdentifier);
    }

    public Exception getLastException() {
		return lastException;
	}
    
    public Boolean getLastExecutionSucceeded() {
		return lastExecutionSucceeded;
	}
    
    public LocalDateTime getTimeOfLastExecution() {
		return timeOfLastExecution;
	}

    private boolean belowTimeLimit(long startMillis) {
        return ((System.currentTimeMillis() - startMillis) < properties.getJobsDeleteMaxDuration().toMillis());
    }
}
