/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.slettejob;

import dk.sundhedsdatastyrelsen.organdonor.conf.TestConfig;
import dk.sundhedsdatastyrelsen.organdonor.conf.TestDAOConfig;
import dk.sundhedsdatastyrelsen.organdonor.conf.TestDBScriptPreprocessor;
import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import javax.sql.DataSource;
import java.time.Instant;
import java.time.Year;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

@RunWith(SpringRunner.class)
@ActiveProfiles("development")
@ContextConfiguration(classes = {TestConfig.class, TestDAOConfig.class, TestDBScriptPreprocessor.class})
@Transactional("stamdataDataSourceTxManager")
public class DeleteOrganDonorRegistrationJobTest extends ServiceTestHelper {

    @Autowired
    DeleteOrganDonorRegistrationJob deleteOrganDonorRegistrationJob;

    @Autowired
    private DataSource stamdataDataSource;

    @Autowired
    OrganDonorDAO organDonorDAO;

    private final String DEFAULT_PERSONID = "1111111111";
    private final String DEFAULT_PERSONID2 = "1111111112";
    private final String DEFAULT_PERSONID3 = "1111111113";
    private final Instant VALID_TO_DEFAULT = Year.of(8099).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
    private Instant now;

    @Before
    public void setup() throws DAOException {
        this.namedParamJdbcTemplate = new NamedParameterJdbcTemplate(stamdataDataSource);

        now = Instant.now();
    }

    @Test
    public void deleteOdrJob_PersonIsDeadNotDead() throws DAOException, InterruptedException {
        // Arrange
        organDonorDAO.createOrganDonorRegistration(getOrganDonorRegistrationExample(DEFAULT_PERSONID2), now);
        boolean existsRecordBeforeDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID2);
        assertTrue(existsRecordBeforeDeletion);

        registerDeath(lessThanAYearAgo(), VALID_TO_DEFAULT); // <--- less than a year ago

        // Act - must be run twice. First run collects persons to be deleted and next run deletes the persons
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();

        // Assert
        boolean existsRecordAfterDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID2);
        assertTrue(existsRecordAfterDeletion); // must still exist
    }

    @Test
    public void deleteOdrJob_PersonIsDeadOneYearAgo_odrDeleted() throws DAOException, InterruptedException {
        // Arrange
        organDonorDAO.createOrganDonorRegistration(getOrganDonorRegistrationExample(DEFAULT_PERSONID), now);
        boolean existsRecordBeforeDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID);
        assertTrue(existsRecordBeforeDeletion);

        registerDeath(moreThanAYearAgo(), VALID_TO_DEFAULT); // <--- more than a year ago

        // Act - must be run twice. First run collects persons to be deleted and next run deletes the persons
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();

        // Assert
        boolean existsRecordAfterDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID);
        assertFalse(existsRecordAfterDeletion); // must not exist
    }

    @Test
    public void deleteOdrJob_PersonIsDeadOneYearAgoCurrentlyAlive_odrNotDeleted() throws DAOException, InterruptedException {
        // Arrange
        organDonorDAO.createOrganDonorRegistration(getOrganDonorRegistrationExample(DEFAULT_PERSONID3), now);
        boolean existsRecordBeforDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID3);
        assertTrue(existsRecordBeforDeletion);

        registerDeath(moreThanAYearAgo(), now);  // <--- more than a year ago
        createAlivePeople(VALID_TO_DEFAULT); // <--- make alive again

        // Act - must be run twice. First run collects persons to be deleted and next run deletes the persons
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();
        deleteOrganDonorRegistrationJob.runDeleteOrganDonorRegistration();

        // Assert
        boolean existsRecordAfterDeletion = organDonorDAO.existsOrganDonorRegistrationRecords(DEFAULT_PERSONID3);
        assertTrue(existsRecordAfterDeletion); // must still exist
    }

    private void createAlivePeople(Instant validTo) {
        createPerson(DEFAULT_PERSONID, "01", Year.of(1935).atMonth(1).atDay(1).atStartOfDay(ZoneId.systemDefault()).toInstant(), now, validTo);
        createPerson(DEFAULT_PERSONID3, "01", Year.of(1935).atMonth(1).atDay(1).atStartOfDay(ZoneId.systemDefault()).toInstant(), now, validTo);
        createPerson("2222222222", "01", Year.of(1975).atMonth(5).atDay(5).atStartOfDay(ZoneId.systemDefault()).toInstant(), now, validTo);
        createPerson("3333333333", "01", Year.of(2001).atMonth(12).atDay(24).atStartOfDay(ZoneId.systemDefault()).toInstant(), now, validTo);
    }

    private Instant lessThanAYearAgo() {
    	ZonedDateTime zonedDateTime = now.atZone(ZoneId.systemDefault());
        zonedDateTime = zonedDateTime.minusYears(1).plusSeconds(10); // a little less than a year ago!
        return zonedDateTime.toInstant();
    }

    private Instant moreThanAYearAgo() {
    	ZonedDateTime zonedDateTime = now.atZone(ZoneId.systemDefault());
        zonedDateTime = zonedDateTime.minusYears(1).minusSeconds(10);// a little more than a year ago!
        return zonedDateTime.toInstant();
    }

    private void registerDeath(Instant when, Instant validTo) {
        closeCprRecord(DEFAULT_PERSONID, when);
        createPerson(DEFAULT_PERSONID,
                "90",
                Year.of(1935).atMonth(1).atDay(1).atStartOfDay(ZoneId.systemDefault()).toInstant(),
                when,
                validTo);
    }
}
