/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.wildfly;

import dk.nsp.test.idp.model.Identity;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.*;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.ToHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.wildfly.idptest.NspServiceException;
import dk.sundhedsdatastyrelsen.organdonor.wildfly.idptest.SOAPOperation;
import dk.sundhedsdatastyrelsen.organdonor.wildfly.idptest.ServiceCaller;
import org.apache.cxf.binding.soap.interceptor.AbstractSoapInterceptor;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.odr.PermissionType;
import org.hl7.v3.II;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;

import java.net.URI;
import java.net.URISyntaxException;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

abstract class AbstractITCase {

    protected static final String TEST_CPRNR = "0501792275";

    public static final String tooYoungCprNr = "1234567890";
    private final ObjectFactory odrFactory = new ObjectFactory();
    private final ToHL7CDAMapper toHL7CDAMapper = new ToHL7CDAMapper();
    protected boolean useAdminForWriteOperations;
    private ServiceCaller serviceCaller;

    protected Identity actor;


    private String uri = System.getProperty("itOdrAddress", "http://localhost:8080/odr") + "/odr";
    private String adminUri = System.getProperty("itOdrAddress", "http://localhost:8080/odr") + "/odrAdmin";

    @Before
    public void setUp() {
        serviceCaller = new ServiceCaller();
    }

    @Test
    public void testOrganDonorRegistration_Request() throws Exception, NspServiceException {
        boolean hasRegistration = testHasOrganDonorRegistration_Request();
        if (hasRegistration) {
            testDeleteOrganDonorRegistration_Request();
        }
        testCreateOrganDonorRegistration_Request();
        testGetOrganDonorRegistration_Request();
        testDeleteOrganDonorRegistration_Request();
    }

    @Test
    public void getRequestShouldBeDeniedWhenPatientIsTooYoung() throws JAXBException {
        final GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(tooYoungCprNr);
        Assert.assertThrows(NspServiceException.class, () -> serviceCaller.call(
                actor,
                request,
                SOAPOperation.GetOrganDonorRegistration.getSoapAction(),
                new URI(uri),
                GetOrganDonorRegistrationRequest.class,
                GetOrganDonorRegistrationResponse.class));
    }

    private boolean testHasOrganDonorRegistration_Request() throws URISyntaxException, JAXBException, NspServiceException {
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(TEST_CPRNR);
        final HasOrganDonorRegistrationResponse response = serviceCaller.call(
                actor,
                request,
                SOAPOperation.HasOrganDonorRegistration.getSoapAction(),
                new URI(uri),
                HasOrganDonorRegistrationRequest.class,
                HasOrganDonorRegistrationResponse.class);

        assertNotNull(response);
        return response.isRegistrationExists();
    }

    private void testCreateOrganDonorRegistration_Request() throws Exception, NspServiceException {
        final CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest(TEST_CPRNR);
        final CreateOrganDonorRegistrationResponse response = serviceCaller.call(
                actor,
                request,
                SOAPOperation.CreateOrganDonorRegistration.getSoapAction(),
                new URI(useAdminForWriteOperations ? adminUri : uri),
                CreateOrganDonorRegistrationRequest.class,
                CreateOrganDonorRegistrationResponse.class);

        // The response is empty, but should not be null
        assertNotNull(response);
    }

    private void testGetOrganDonorRegistration_Request() throws JAXBException, URISyntaxException, NspServiceException {
        final GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(TEST_CPRNR);
        final GetOrganDonorRegistrationResponse response = serviceCaller.call(
                actor,
                request,
                SOAPOperation.GetOrganDonorRegistration.getSoapAction(),
                new URI(uri),
                GetOrganDonorRegistrationRequest.class,
                GetOrganDonorRegistrationResponse.class);

        assertNotNull(response);
        assertEquals(TEST_CPRNR, response.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension());
    }

    private void testDeleteOrganDonorRegistration_Request() throws JAXBException, URISyntaxException, NspServiceException {
        final DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRegistrationRequest(TEST_CPRNR);
        final DeleteOrganDonorRegistrationResponse response =  serviceCaller.call(
                actor,
                request,
                SOAPOperation.DeleteOrganDonorRegistration.getSoapAction(),
                new URI(useAdminForWriteOperations ? adminUri : uri),
                DeleteOrganDonorRegistrationRequest.class,
                DeleteOrganDonorRegistrationResponse.class);

        assertNotNull(response);
    }

    private OrganDonorRegistration createDefaultOrganDonorRegistration(PermissionType permissionType, boolean heart, boolean lungs,
                                                                       boolean liver, boolean pancreas, boolean kidneys,
                                                                       boolean cornea, boolean smallIntestine, boolean skin,
                                                                       boolean relativeAcceptance) {
        OrganDonorRegistration odr = new OrganDonorRegistration();
        odr.setPermissionType(permissionType);
        odr.setPermissionForHeart(heart);
        odr.setPermissionForLungs(lungs);
        odr.setPermissionForLiver(liver);
        odr.setPermissionForPancreas(pancreas);
        odr.setPermissionForKidneys(kidneys);
        odr.setPermissionForCornea(cornea);
        odr.setPermissionForSmallIntestine(smallIntestine);
        odr.setPermissionForSkin(skin);
        odr.setRequiresRelativeAcceptance(relativeAcceptance);
        return odr;
    }

    private II createDefaultII(String personIdentifier) {
        final String AUTHORITYNAME = "CPR";
        final String ROOT = "1.2.208.176.1.2";
        return toHL7CDAMapper.createId(ROOT, personIdentifier, AUTHORITYNAME);
    }

    private CreateOrganDonorRegistrationRequest createCreateOrganDonorRegistrationRequest(String personIdentifier) {
        final PermissionType permissionType = PermissionType.FULL;
        final boolean heart = true;
        final boolean lungs = true;
        final boolean liver = false;
        final boolean pancreas = true;
        final boolean kidneys = false;
        final boolean cornea = false;
        final boolean smallIntestine = true;
        final boolean skin = false;
        final boolean relativeAcceptance = true;
        return createCreateOrganDonorRegistrationRequest(personIdentifier, permissionType,
                heart, lungs, liver, pancreas, kidneys, cornea, smallIntestine, skin, relativeAcceptance);
    }

    @SuppressWarnings("SameParameterValue")
    private CreateOrganDonorRegistrationRequest createCreateOrganDonorRegistrationRequest(String personIdentifier, PermissionType permissionType, boolean heart, boolean lungs,
                                                                                          boolean liver, boolean pancreas, boolean kidneys,
                                                                                          boolean cornea, boolean smallIntestine, boolean skin,
                                                                                          boolean relativeAcceptance) {
        final CreateOrganDonorRegistrationRequest createRequest = odrFactory.createCreateOrganDonorRegistrationRequest();
        createRequest.setId(createDefaultII(personIdentifier));
        createRequest.setOrganDonorRegistration(createDefaultOrganDonorRegistration(permissionType, heart, lungs, liver, pancreas, kidneys, cornea,
                smallIntestine, skin, relativeAcceptance));
        return createRequest;
    }

    private DeleteOrganDonorRegistrationRequest createDeleteOrganDonorRegistrationRequest(String personIdentifier) {
        DeleteOrganDonorRegistrationRequest request = odrFactory.createDeleteOrganDonorRegistrationRequest();
        request.setId(createDefaultII(personIdentifier));
        return request;
    }

    private GetOrganDonorRegistrationRequest createGetOrganDonorRegistrationRequest(String personIdentifier) {
        GetOrganDonorRegistrationRequest request = odrFactory.createGetOrganDonorRegistrationRequest();
        request.setId(createDefaultII(personIdentifier));
        return request;
    }

    private HasOrganDonorRegistrationRequest createHasOrganDonorRegistrationRequest(String personIdentifier) {
        HasOrganDonorRegistrationRequest request = odrFactory.createHasOrganDonorRegistrationRequest();
        request.setId(createDefaultII(personIdentifier));
        return request;
    }
}
