/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.wildfly;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

public class HttpTestClient {


	private final URL url;
	
	public HttpTestClient(String url) throws MalformedURLException {
		this.url = new URL(url);
	}
	
	/**
	 * This will send the body as content of the HTTP request to the
	 * <code>sts.url</code>
	 * 
	 * @param soapEnvolope
	 * @return the response parsed as a <code>Document</code>
	 */
	public Document execute(String soapEnvolope) {
		SoapClient sc = new SoapClient(url);
		return sc.execute(soapEnvolope);
	}


	/**
	 * This is the SoapClient from Seal's TestTools, but the encoding is forced to utf-8.
	 * The header and body would otherwise disagree.
	 */
	private static class SoapClient {
		private final URL url;

		private HttpURLConnection httpConn;

		private final static Logger logger = Logger.getLogger(SoapClient.class);

		public SoapClient(URL SOAPUrl) {
			url = SOAPUrl;
		}

		public Document execute(String soapEnvelope) {
			logger.debug("SOAPClient.execute:" + soapEnvelope);
			URLConnection connection;

			try {
				connection = url.openConnection();
				httpConn = (HttpURLConnection) connection;
				final byte[] bytes = soapEnvelope.getBytes("UTF-8");

				httpConn.setRequestProperty("Content-Length", String
						.valueOf(bytes.length));
				httpConn.setRequestProperty("Content-Type",
											"text/xml; charset=utf-8");
				httpConn.setRequestProperty("SOAPAction", "\"\"");
				httpConn.setRequestMethod("POST");
				httpConn.setDoOutput(true);
				httpConn.setDoInput(true);

				OutputStream out = httpConn.getOutputStream();
				out.write(bytes);
				out.close();

				final String contentType = httpConn.getContentType();
				if (!(contentType.contains("text/xml") && contentType.contains("utf-8"))) {
					throw new IllegalStateException("Unexpected content type '" + contentType + "'. Expected 'text/xml' and 'charset=utf-8'.");
				}
				InputSource inputSource = new InputSource((httpConn.getResponseCode() >= 500) ? httpConn.getErrorStream()
																: httpConn.getInputStream());

				DocumentBuilderFactory factory = DocumentBuilderFactory
						.newInstance();
				factory.setValidating(false);
				factory.setNamespaceAware(true);
				Document document;
				try {
					DocumentBuilder builder = factory.newDocumentBuilder();
					document = builder.parse(inputSource);

				} catch (SAXException sxe) {

					Exception x = sxe;
					if (sxe.getException() != null) {
						x = sxe.getException();
					}
					throw new RuntimeException(x.getMessage(),x);
				} catch (ParserConfigurationException pce) {
					throw new RuntimeException(pce.getMessage(),pce);
				}
				return document;
			} catch (IOException e) {
				throw new RuntimeException("Error during soap invocation at " + url + " : "
													+ e.getMessage(), e);
			}
		}
	}

}
