/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service;

import dk.sundhedsdatastyrelsen.organdonor.conf.AbstractHttpServlet;
import dk.sundhedsdatastyrelsen.organdonor.service.util.HealthIndicator;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OdrContextListener;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import javax.servlet.annotation.WebServlet;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.time.Instant;
import java.util.Collection;
import java.util.Properties;

@WebServlet(name="IsAliveServlet", urlPatterns={"/isAlive"})
public class IsAliveServlet extends AbstractHttpServlet {

    private BuildData buildData;

    @Autowired
    private DataSource organDonorDataSource;

    @Autowired
    private DataSource stamdataDataSource;

    @Autowired(required = false)
    private Collection<HealthIndicator> healthIndicators;

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        int status;
        String statusMessage;

        if (datasourceResponding()) {
            status = 200;
            statusMessage = " OK";
            // Check other health indicators
            if (healthIndicators != null) {
                for (HealthIndicator healthIndicator : healthIndicators) {
                    if(healthIndicator != null) {
                        int indicatorStatus = healthIndicator.getStatusCode();
                        if (indicatorStatus > status) {
                            status = indicatorStatus;
                            statusMessage = " Non-Authoritative Information from: " + healthIndicator.getName();
                        }
                    }
                }
            }
            response.setStatus(status);
            response.setContentType("text/plain");
        } else {
            status = 500;
            statusMessage = "Internal Server Error: Database check failed";
        }

        response.getWriter().println(String.format("%s %s", status, statusMessage));
        response.getWriter().println();

        buildData = getBuildData(request.getServletContext());
        response.getWriter().println("Title: " + buildData.title);
        response.getWriter().println("Deployed: " + OdrContextListener.DEPLOY_TIME);
        response.getWriter().println("Build-Date: " + buildData.date);
        response.getWriter().println("Build-Version: " + buildData.version);
        response.getWriter().println("Builder: " + buildData.builder);
        response.getWriter().println("Display time: " + Instant.now());
        response.getWriter().println("\n");
        response.getWriter().println(getHealth());
    }

    private boolean datasourceResponding() {
        try {
            new JdbcTemplate(organDonorDataSource).execute("SELECT 1");
            new JdbcTemplate(stamdataDataSource).execute("SELECT 1");
        } catch (Exception e) {
            return false;
        }
        return true;
    }

    private BuildData getBuildData(ServletContext servletContext) throws IOException {
        if (buildData == null) {
            InputStream stream = servletContext.getResourceAsStream("/META-INF/MANIFEST.MF");
            if (stream == null) {
                // Local
                stream = this.getClass().getResourceAsStream("/META-INF/MANIFEST.MF");
            }
            if (stream != null) {
                Properties prop = new Properties();
                prop.load(new InputStreamReader(stream, "UTF-8"));
                stream.close();
                buildData = new BuildData(prop.getProperty("Implementation-Title"), prop.getProperty("Implementation-Version"), prop.getProperty("Built-Date"), prop.getProperty("Built-By"));
            }
        }
        return buildData;
    }

    private class BuildData {
        String title;
        String version;
        String date;
        String builder;

        BuildData(String title, String version, String date, String builder) {
            this.title = title;
            this.version = version;
            this.date = date;
            this.builder = builder;
        }
    }

    private String getHealth() {
        String health = "";
        if(healthIndicators != null) {
            for (HealthIndicator healthIndicator : healthIndicators) {
                if(healthIndicator != null) {
                    health = health + healthIndicator.getName()+" : \""+healthIndicator.getStatusText()+"\"\n";
                }
            }
        }
        return health;
    }
}
