/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service;

import dk.sundhedsdatastyrelsen.organdonor.conf.AbstractHttpServlet;
import dk.sundhedsdatastyrelsen.organdonor.service.util.WSDLHelper;
import org.apache.commons.io.Charsets;
import org.apache.commons.io.FilenameUtils;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.util.FileCopyUtils;

import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.StandardCharsets;

@WebServlet(name="WSDLServlet", urlPatterns={"/wsdl", "/wsdl/idws", "/wsdl/dgws", "/schema/*"})
public class WSDLServlet extends AbstractHttpServlet {

    private static final Logger logger = LogManager.getLogger(WSDLServlet.class);

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        if (request.getServletPath().equals("/wsdl")) {
            logger.debug("WSDL request");
            response.setContentType("text/html");
            response.setStatus(200);
            response.setCharacterEncoding(Charsets.UTF_8.toString());
            String resp = getWSDL();
            response.getWriter().write(resp);

        } else if (request.getServletPath().equals("/wsdl/dgws")) {
            logger.debug("WSDL dgws request");
            response.setContentType("text/xml");
            response.setStatus(200);
            response.setCharacterEncoding(Charsets.UTF_8.toString());
            String resp = getWsdlDGWS(request);
            response.getWriter().write(resp);

        } else if (request.getServletPath().equals("/wsdl/idws")) {
            logger.debug("WSDL idws request");
            response.setContentType("text/xml");
            response.setStatus(200);
            response.setCharacterEncoding(Charsets.UTF_8.toString());
            String resp = getWsdlIDWS(request);
            response.getWriter().write(resp);

        } else if (request.getServletPath().equals("/schema")) {
            logger.debug("Schema request: " + request.getPathInfo());
            String resp = getSchemaFile(request);
            if (resp == null){
                response.setStatus(404);
                response.setCharacterEncoding(Charsets.UTF_8.toString());
            }
            else{
                response.setContentType("text/xml");
                response.setStatus(200);
                response.setCharacterEncoding(Charsets.UTF_8.toString());
                response.getWriter().write(resp);
            }
        }
    }


    private String getWSDL() throws IOException {
        InputStream in = getClass().getResourceAsStream("/static/wsdl_overview.html");
        Reader reader = new InputStreamReader(in, StandardCharsets.UTF_8);
        return FileCopyUtils.copyToString(reader);
    }

    private String getWsdlDGWS(HttpServletRequest request) throws IOException {
        return outputString("/wsdl/organdonor_dgws_service_2018_05_01.wsdl", request.getRequestURL().toString());
    }

    private String getWsdlIDWS(HttpServletRequest request) throws IOException {
        return outputString("/wsdl/organdonor_idws_service_2018_05_01.wsdl", request.getRequestURL().toString());
    }

    private String getSchemaFile(HttpServletRequest request) throws IOException {
        try{
        String servletPath = FilenameUtils.normalize(request.getPathInfo());
        return outputString("/schema/"+servletPath, request.getRequestURL().toString());}
        catch (NullPointerException e ){
            return null;
        }
    }

    private String outputString(String resource, String requestURL) throws IOException {
        InputStream in = getClass().getResourceAsStream(resource);
        Reader reader = new InputStreamReader(in, StandardCharsets.UTF_8);
        String text = FileCopyUtils.copyToString(reader);

        if (requestURL != null) {
            text = WSDLHelper.replaceSchemaLocations(text, requestURL);
        }

        return text;
    }

}
