/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.auditlog;

import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.audit.AuditBuilder.TypeOfInformation;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;

import java.util.function.Supplier;

public class AuditLogger {
    private static final String COMPONENT = "ODR";

    private static final String CREATE_ORGAN_DONOR_REGISTRATION_CONTEXT = "createOrganDonorRegistration";
    private static final String UPDATE_ORGAN_DONOR_REGISTRATION_CONTEXT = "updateOrganDonorRegistration";
    private static final String DELETE_ORGAN_DONOR_REGISTRATION_CONTEXT = "deleteOrganDonorRegistration";
    private static final String GET_ORGAN_DONOR_REGISTRATION_CONTEXT = "getOrganDonorRegistration";
    private static final String HAS_ORGAN_DONOR_REGISTRATION_CONTEXT = "hasOrganDonorRegistration";
    private static final String VALIDATE_CPR_CONTEXT = "validateCpr";

    private final Supplier<AuditBuilder> auditBuilderSupplier;

    public AuditLogger() {
        this(Audit::createAuditBuilder);
    }

    public AuditLogger(Supplier<AuditBuilder> auditBuilderSupplier) {
        this.auditBuilderSupplier = auditBuilderSupplier;
    }

    public void logCreateOrganDonorRegistration(OrganDonorVO organDonor, ServiceActorVO serviceActorVO) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = CREATE_ORGAN_DONOR_REGISTRATION_CONTEXT;
        logOrganDonor(organDonor, auditBuilder, context);
        logServiceActor(auditBuilder, context, serviceActorVO);
    }

    public void logUpdateOrganDonorRegistration(OrganDonorVO organDonor, ServiceActorVO serviceActorVO) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = UPDATE_ORGAN_DONOR_REGISTRATION_CONTEXT;
        logOrganDonor(organDonor, auditBuilder, context);
        logServiceActor(auditBuilder, context, serviceActorVO);
    }

    public void logDeleteOrganDonorRegistration(String personIdentifier, ServiceActorVO serviceActorVO) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = DELETE_ORGAN_DONOR_REGISTRATION_CONTEXT;
        log(auditBuilder, context, Type.CPR, personIdentifier);
        logServiceActor(auditBuilder, context, serviceActorVO);
    }

    public void logGetOrganDonorRegistration(OrganDonorVO organDonor, ServiceActorVO serviceActorVO) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = GET_ORGAN_DONOR_REGISTRATION_CONTEXT;
        logOrganDonor(organDonor, auditBuilder, context);
        logServiceActor(auditBuilder, context, serviceActorVO);
    }

    public void logHasOrganDonorRegistration(String personIdentifier, ServiceActorVO serviceActorVO) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        String context = HAS_ORGAN_DONOR_REGISTRATION_CONTEXT;
        log(auditBuilder, context, Type.CPR, personIdentifier);
        logServiceActor(auditBuilder, context, serviceActorVO);
    }

    public void logCprDoesNotExist(String cpr) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        log(auditBuilder, VALIDATE_CPR_CONTEXT, Type.CPR_DOES_NOT_EXIST, cpr);
    }

    public void logCprInactive(String cpr) {
        AuditBuilder auditBuilder = auditBuilderSupplier.get();
        log(auditBuilder, VALIDATE_CPR_CONTEXT, Type.CPR_INACTIVE, cpr);
    }

    private void logOrganDonor(OrganDonorVO organDonor, AuditBuilder auditBuilder, String context) {
        log(auditBuilder, context, Type.PID, organDonor.getOrganDonorPID());
        log(auditBuilder, context, Type.CPR, organDonor.getPersonIdentifier());
        log(auditBuilder, context, Type.VERSION, organDonor.getVersion());
        log(auditBuilder, context, Type.MODIFIED, organDonor.getModifiedDateTime());
        log(auditBuilder, context, Type.VALID_FROM, organDonor.getValidFrom());
        log(auditBuilder, context, Type.VALID_TO, organDonor.getValidTo());
        log(auditBuilder, context, Type.PERMISSION_TYPE, organDonor.getPermissionType());
        log(auditBuilder, context, Type.HEART, organDonor.getHeart());
        log(auditBuilder, context, Type.KIDNEYS, organDonor.getKidneys());
        log(auditBuilder, context, Type.LUNGS, organDonor.getLungs());
        log(auditBuilder, context, Type.CORNEAS, organDonor.getCorneas());
        log(auditBuilder, context, Type.LIVER, organDonor.getLiver());
        log(auditBuilder, context, Type.SMALL_INTESTINE, organDonor.getSmallIntestine());
        log(auditBuilder, context, Type.PANCREAS, organDonor.getPancreas());
        log(auditBuilder, context, Type.SKIN, organDonor.getSkin());
        log(auditBuilder, context, Type.RELATIVE_ACCEPTANCE_REQUIRED, organDonor.getRelativeAcceptanceRequired());
    }

    private void logServiceActor(AuditBuilder auditBuilder, String context, ServiceActorVO serviceActor) {
        log(auditBuilder, context, Type.ACTOR_CPR, serviceActor.getPersonIdentifier());
        log(auditBuilder, context, Type.ACTOR_USERTYPE, serviceActor.getUserType());
    }

    private void log(AuditBuilder auditBuilder, String context, Type type, Object value) {
        if (value != null) {
            auditBuilder.addAuditInformation(COMPONENT, context, type.typeOfInformation, type.value, value);
        }
    }

    private enum Type {
        PID("pid", TypeOfInformation.NonPersonalInformation),
        CPR("cpr", TypeOfInformation.SensitivePersonalInformation),
        VERSION("version", TypeOfInformation.NonPersonalInformation),
        MODIFIED("modified", TypeOfInformation.NonPersonalInformation),
        VALID_FROM("validFrom", TypeOfInformation.NonPersonalInformation),
        VALID_TO("validTo", TypeOfInformation.NonPersonalInformation),
        PERMISSION_TYPE("permissionType", TypeOfInformation.SensitivePersonalInformation),
        HEART("heart", TypeOfInformation.SensitivePersonalInformation),
        KIDNEYS("kidneys", TypeOfInformation.SensitivePersonalInformation),
        LUNGS("lungs", TypeOfInformation.SensitivePersonalInformation),
        CORNEAS("corneas", TypeOfInformation.SensitivePersonalInformation),
        LIVER("liver", TypeOfInformation.SensitivePersonalInformation),
        SMALL_INTESTINE("smallIntestine", TypeOfInformation.SensitivePersonalInformation),
        PANCREAS("pancreas", TypeOfInformation.SensitivePersonalInformation),
        SKIN("skin", TypeOfInformation.SensitivePersonalInformation),
        RELATIVE_ACCEPTANCE_REQUIRED("relativeAcceptanceRequired", TypeOfInformation.SensitivePersonalInformation),
        ACTOR_CPR("actor-cpr", TypeOfInformation.SensitivePersonalInformation),
        ACTOR_USERTYPE("actor-usertype", TypeOfInformation.RegularPersonalInformation),
        CPR_DOES_NOT_EXIST("cpr-does-not-exist", TypeOfInformation.SensitivePersonalInformation),
        CPR_INACTIVE("cpr-inactive", TypeOfInformation.SensitivePersonalInformation);

        private final String value;
        private final TypeOfInformation typeOfInformation;

        Type(String value, TypeOfInformation typeOfInformation) {
            this.value = value;
            this.typeOfInformation = typeOfInformation;
        }
    }
}
