/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.business;

import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.nas.organdonor.NasException;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.nas.organdonor.NotificationSender;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultUtil;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;

@Service
public class OrganDonorServiceImpl implements OrganDonorService {

    private static final Logger log = LogManager.getLogger(OrganDonorServiceImpl.class);

    static final String ODR_SERVICE_IDENTIFIER = "Organdonorregister";

    private final Properties properties;
    private final AccessControl accessControl;
    private final MinLogDataWrapper minLogDataWrapper;
    private final Validator validator;
    private final OrganDonorDAO organDonorDAO;

    @Autowired
    private NotificationSender notificationSender;

    static final int INITIAL_VERSION = 1;

    @Autowired
    public OrganDonorServiceImpl(Properties properties, MinLogDataWrapper minLogDataWrapper, AccessControl accessControl, Validator validator, OrganDonorDAO organDonorDAO, NotificationSender notificationSender) {
        this.properties = properties;
        this.minLogDataWrapper = minLogDataWrapper;
        this.accessControl = accessControl;
        this.validator = validator;
        this.organDonorDAO = organDonorDAO;
        this.notificationSender = notificationSender;
    }

    @Override
    @Transactional
    public void createOrganDonorRegistration(OrganDonorVO organDonorVO, RequestContextVO context) {
        accessControl.checkWriteAccess(context);
        try {
            final String personIdentifier = context.getPersonIdentifier();
            final Instant nowTimestamp = context.getTime();
            validator.validateOrganDonorRegistrationData(organDonorVO);
            OrganDonorVO existingOrganDonorRegistration = organDonorDAO.getOrganDonorRegistration(personIdentifier, nowTimestamp);
            validator.validatePersonIdOnCreate(existingOrganDonorRegistration, personIdentifier);
            organDonorVO.setVersion(INITIAL_VERSION);
            organDonorDAO.createOrganDonorRegistration(organDonorVO, nowTimestamp);
            minLogIfNotSystemUser(context, properties.getMinLogCreateActivityText());

            // Send a notification to NAS, indicating that the Organ Donor has changed.
            sendUpdateNotificationToNas(context, nowTimestamp);

        } catch (DAOException | MinLogException | NasException e) {
            FaultUtil.throwODRFault(FaultCode.CREATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    @Transactional
    public void updateOrganDonorRegistration(OrganDonorVO organDonorVO, RequestContextVO context) {
        accessControl.checkWriteAccess(context);
        try {
            final String personIdentifier = context.getPersonIdentifier();
            Instant nowTimestamp = context.getTime();
            validator.validateOrganDonorRegistrationData(organDonorVO);
            OrganDonorVO existingOrganDonorRegistration = organDonorDAO.getOrganDonorRegistration(personIdentifier, nowTimestamp);
            validator.validatePersonId(existingOrganDonorRegistration, personIdentifier);
            final int existingVersion = existingOrganDonorRegistration == null ? 0 : existingOrganDonorRegistration.getVersion();
            final int nextVersion = existingVersion + 1;
            organDonorVO.setVersion(nextVersion);
            organDonorDAO.updateOrganDonorRegistration(existingOrganDonorRegistration.getOrganDonorPID(), organDonorVO, nowTimestamp);
            minLogIfNotSystemUser(context, properties.getMinLogUpdateActivityText());

            // Send a notification to NAS, indicating that the Organ Donor has changed.
            sendUpdateNotificationToNas(context, nowTimestamp);

        } catch (DAOException | MinLogException | NasException e) {
            FaultUtil.throwODRFault(FaultCode.UPDATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    @Transactional
    public void deleteOrganDonorRegistration(RequestContextVO context) {
        accessControl.checkWriteAccess(context);
        try {
            final String personIdentifier = context.getPersonIdentifier();
            final Instant nowTimestamp = context.getTime();
            OrganDonorVO existingOrganDonorRegistration = organDonorDAO.getOrganDonorRegistration(personIdentifier, nowTimestamp);
            validator.validatePersonId(existingOrganDonorRegistration, personIdentifier);
            if (existingOrganDonorRegistration.getOrganDonorPID() != null) {
                organDonorDAO.deleteOrganDonorRegistration(existingOrganDonorRegistration.getOrganDonorPID(), nowTimestamp);

                // Send a notification to NAS, indicating that the Organ Donor has changed.
                sendUpdateNotificationToNas(context, nowTimestamp);

            }
            minLogIfNotSystemUser(context, properties.getMinLogDeleteActivityText());
        } catch (DAOException | MinLogException | NasException e) {
            FaultUtil.throwODRFault(FaultCode.DELETE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    public OrganDonorVO getOrganDonorRegistration(RequestContextVO context) {
        accessControl.checkGetAccess(context);
        OrganDonorVO returnedOrganDonorRegistration = null;

        try {
            final String personIdentifier = context.getPersonIdentifier();
            returnedOrganDonorRegistration = organDonorDAO.getOrganDonorRegistration(personIdentifier, context.getTime());
            validator.validatePersonId(returnedOrganDonorRegistration, personIdentifier);

            // Log to MinLog. If MinLog fails, it throws an exception
            minLogIfNotSystemUserAndNotCitizen(context, properties.getMinLogReadActivityText());
        } catch (DAOException | MinLogException e) {
            FaultUtil.throwODRFault(FaultCode.GET_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e);
        }
        return returnedOrganDonorRegistration;
    }

    @Override
    public boolean hasOrganDonorRegistration(RequestContextVO context) {
        accessControl.checkHasAccess(context);
        OrganDonorVO returnedOrganDonorRegistration = null;

        try {
            final String personIdentifier = context.getPersonIdentifier();
            returnedOrganDonorRegistration = organDonorDAO.getOrganDonorRegistration(personIdentifier, context.getTime());
        } catch (DAOException e) {
            FaultUtil.throwODRFault(FaultCode.HAS_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e);
        }
        return returnedOrganDonorRegistration != null;
    }

    private void minLogIfNotSystemUser(RequestContextVO context, String minLogText) throws MinLogException {
        if (context.getServiceActor().getUserType() != UserType.SystemUser) {
            logActivityToMinLog(context, minLogText);
        }
    }
    
    private void minLogIfNotSystemUserAndNotCitizen(RequestContextVO context, String minLogText) throws MinLogException {
        if (context.getServiceActor().getUserType() != UserType.SystemUser && context.getServiceActor().getUserType() != UserType.Citizen) {
            logActivityToMinLog(context, minLogText);
        }
    }
    
    private void logActivityToMinLog(RequestContextVO context, String minLogText) throws MinLogException {
        minLogDataWrapper.logActivity(context, minLogText, ODR_SERVICE_IDENTIFIER);
    }

    private void sendUpdateNotificationToNas(RequestContextVO context, Instant currentTime) throws NasException {
        notificationSender.sendOrganDonorNotification(context.getPersonIdentifier(), currentTime);
    }
}
