/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.business;

import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.ValidationErrors;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;

@Component
public class Validator {
    @Autowired
    OrganDonorDAO organDonorDAO;

    public void validatePersonIdOnCreate(OrganDonorVO organDonorRegistration, String personId) {
        List<String> validationErrors = new ArrayList<>();
        if (organDonorRegistration != null) {
            validationErrors.add(format(ValidationErrors.PERSON_ID_EXISTS, organDonorRegistration == null ? personId : organDonorRegistration.getPersonIdentifier()));
        }
        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    public void validatePersonId(OrganDonorVO organDonorRegistration, String personId) throws DAOException {
        List<String> validationErrors = new ArrayList<>();
        if (organDonorRegistration == null) {
            validationErrors.add(format(ValidationErrors.PERSON_ID_INVALID, personId));
        }
        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    public void validateOrganDonorRegistrationData(OrganDonorVO organDonorVO) {
        List<String> validationErrors = new ArrayList<>();
        validatePidNotProvided(organDonorVO, validationErrors);
        validatePersonIdentifierProvided(organDonorVO.getPersonIdentifier(), validationErrors);

        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    private void validatePersonIdentifierProvided(String personIdentifier, List<String> validationErrors) {
        if (personIdentifier == null) {
            validationErrors.add(format(ValidationErrors.PERSONID_REQUIRED));
        }
    }

    public void validatePidNotProvided(OrganDonorVO organDonorVO, List<String> validationErrors) {
        if (organDonorVO == null) return;

        if (organDonorVO.getOrganDonorPID() != null) {
            validationErrors.add(format(ValidationErrors.PID_FORBIDDEN, organDonorVO.getOrganDonorPID().toString()));
        }
    }

    private void throwExceptionInCaseOfValidationErrors(List<String> validationErrors) {
        if (validationErrors.size() > 0) {
            throw new IllegalArgumentException(String.join(", ", validationErrors));
        }
    }

    public String format(String error, String... expressions) {
        if (expressions != null) {
            for (int i=0; i<expressions.length; i++) {
                if (expressions[i] == null) {
                    expressions[i] = "";
                }
            }
        }
        String formatted = String.format(error, expressions);
        return formatted;
    }
}
