/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice;

import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.ws.SLALoggingServletFilter;
import dk.sundhedsdatastyrelsen.organdonor.service.AdminServlet;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultUtil;
import dk.sundhedsdatastyrelsen.organdonor.service.util.ODRException;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.jws.HandlerChain;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.xml.ws.BindingType;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.Addressing;
import java.text.ParseException;

import static javax.xml.ws.soap.SOAPBinding.SOAP11HTTP_BINDING;

@BindingType(value = SOAP11HTTP_BINDING)
@Addressing(enabled = true, required = false)
@HandlerChain(file = "handler-chain.xml")
public abstract class AbstractWebService {

    private static final Logger logger = LogManager.getLogger(AbstractWebService.class);

    @Resource
    protected WebServiceContext context;

    @PostConstruct
    public void init() throws ServletException {
        ServletContext servletContext = (ServletContext) context.getMessageContext().get(MessageContext.SERVLET_CONTEXT);
        Object o = servletContext.getAttribute(OdrContextListener.ANNOTATION_CONFIG_CONTEXT);
        if(o instanceof AnnotationConfigApplicationContext) {
            ((AnnotationConfigApplicationContext)o).getAutowireCapableBeanFactory().autowireBean(this);
        }
    }

    boolean isAdmin() {
        ServletRequest servletRequest = (ServletRequest) context.getMessageContext().get(MessageContext.SERVLET_REQUEST);
        Boolean isAdmin = (Boolean) servletRequest.getAttribute(AdminServlet.ADMIN_KEY);
        return isAdmin != null ? isAdmin : false;
    }

    SLALogItem getSLALogItem() {
        ServletRequest servletRequest = (ServletRequest) context.getMessageContext().get(MessageContext.SERVLET_REQUEST);
        return (SLALogItem) servletRequest.getAttribute(SLALoggingServletFilter.CURRENT_LOGITEM_REQUEST_KEY);
    }

    protected void commonExceptionHandler(Exception e, FaultCode validationErrorCode, FaultCode internalServerErrorCode) {
        if (e instanceof IllegalArgumentException || e instanceof ParseException) {
            // Ok to reveal detailed information

            logger.info("Service call threw the following request validation exception: " + e.getMessage());

            FaultUtil.throwODRFault(validationErrorCode,
                    validationErrorCode.getCode() + ": " + validationErrorCode.getMessage()
                            + " Details: " + e.getMessage());

        } else if (e instanceof ODRException) {
            // Ok to reveal detailed information (two error codes will be present)
            throw (ODRException) e;
        }

        // Not ok to reveal detailed information
        logger.error("Service call produced the an unrecognized exception with code", e);
        FaultUtil.throwODRFault(internalServerErrorCode,
                internalServerErrorCode.getCode() + ": " + internalServerErrorCode.getMessage());
    }

    /**
     * For testing
     */
    public void setContext(WebServiceContext context) { this.context = context; }
}
