/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice;

import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.*;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.organdonor.service.business.OrganDonorService;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.context.RequestContextProvider;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.FromHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.ToHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.CprExists;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.RequestValidator;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.beans.factory.annotation.Autowired;

import javax.jws.WebService;
import javax.servlet.annotation.WebServlet;

@WebServlet(name="OrganDonorRegistrationService", urlPatterns={"/odr"})
@WebService(serviceName = "OrganDonorRegistrationService", wsdlLocation = "wsdl/organdonor_service_2018_05_01.wsdl", targetNamespace = "http://sundhedsdatastyrelsen.dk/organdonor/2018/05/01/", portName = "OrganDonorRegistrationPort")
@SchemaValidation
public class OrganDonorWebService extends AbstractWebService implements OrganDonorRegistrationPortType {
    private static final Logger log = LogManager.getLogger(OrganDonorWebService.class);

    @Autowired
    public RequestContextProvider requestContextProvider;

    @Autowired
    private FromHL7CDAMapper fromHL7CDAMapper;

    @Autowired
    private OrganDonorService businessService;

    @Autowired
    private ToHL7CDAMapper toHL7CDAMapper;

    @Autowired
    private CprExists cprExists;

    @Autowired
    private Properties properties;

    @Autowired
    private AuditLogger auditLogger;

    @Override
    public CreateOrganDonorRegistrationResponse createOrganDonorRegistration20180501(CreateOrganDonorRegistrationRequest request) {
        final CreateOrganDonorRegistrationResponse response = new ObjectFactory().createCreateOrganDonorRegistrationResponse();
        try {
            RequestContextVO requestContext = requestContextProvider.getRequestContext(request.getId().getExtension(), isAdmin(), getSLALogItem());
            new RequestValidator(cprExists, properties).validate(request);
            OrganDonorVO organDonorVO = fromHL7CDAMapper.mapOrganDonorFromRequest(request);
            businessService.createOrganDonorRegistration(organDonorVO, requestContext);
            auditLogger.logCreateOrganDonorRegistration(organDonorVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_ORGANDONORREGISTRATION_VALIDATION_ERROR, FaultCode.CREATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR);
        }
        return response;
    }

    @Override
    public UpdateOrganDonorRegistrationResponse updateOrganDonorRegistration20180501(UpdateOrganDonorRegistrationRequest request) {
        final UpdateOrganDonorRegistrationResponse response = new ObjectFactory().createUpdateOrganDonorRegistrationResponse();
        try {
            RequestContextVO requestContext = requestContextProvider.getRequestContext(request.getId().getExtension(), isAdmin(), getSLALogItem());
            new RequestValidator(cprExists, properties).validate(request);
            OrganDonorVO organDonorVO = fromHL7CDAMapper.mapOrganDonorFromRequest(request);
            businessService.updateOrganDonorRegistration(organDonorVO, requestContext);
            auditLogger.logUpdateOrganDonorRegistration(organDonorVO, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_ORGANDONORREGISTRATION_VALIDATION_ERROR, FaultCode.UPDATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR);
        }
        return response;
    }

    @Override
    public DeleteOrganDonorRegistrationResponse deleteOrganDonorRegistration20180501(DeleteOrganDonorRegistrationRequest request) {
        DeleteOrganDonorRegistrationResponse response = new ObjectFactory().createDeleteOrganDonorRegistrationResponse();
        try {
            RequestContextVO requestContext = requestContextProvider.getRequestContext(request.getId().getExtension(), isAdmin(), getSLALogItem());
            new RequestValidator(cprExists, properties).validate(request);
            businessService.deleteOrganDonorRegistration(requestContext);
            auditLogger.logDeleteOrganDonorRegistration(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_ORGANDONORREGISTRATION_VALIDATION_ERROR, FaultCode.DELETE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR);
        }
        return response;
    }

    @Override
    public GetOrganDonorRegistrationResponse getOrganDonorRegistration20180501(GetOrganDonorRegistrationRequest request) {
        GetOrganDonorRegistrationResponse response = new ObjectFactory().createGetOrganDonorRegistrationResponse();
        try {
            RequestContextVO requestContext = requestContextProvider.getRequestContext(request.getId().getExtension(), isAdmin(), getSLALogItem());
            new RequestValidator(cprExists, properties).validate(request);
            OrganDonorVO returnedOrganDonorRegistration = businessService.getOrganDonorRegistration(requestContext);
            final POCDMT000040ClinicalDocument clinicalDocument = toHL7CDAMapper.toClinicalDocument(returnedOrganDonorRegistration.getPersonIdentifier(), returnedOrganDonorRegistration, requestContext);
            response.setClinicalDocument(clinicalDocument);
            auditLogger.logGetOrganDonorRegistration(returnedOrganDonorRegistration, requestContext.getServiceActor());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_ORGANDONORREGISTRATION_VALIDATION_ERROR, FaultCode.GET_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR);
        }
        return response;
    }

    @Override
    public HasOrganDonorRegistrationResponse hasOrganDonorRegistration20180501(HasOrganDonorRegistrationRequest request) {
        HasOrganDonorRegistrationResponse response = new ObjectFactory().createHasOrganDonorRegistrationResponse();
        RequestValidator requestValidator = new RequestValidator(cprExists, properties);

        try {
            RequestContextVO requestContext = requestContextProvider.getRequestContext(request.getId().getExtension(), isAdmin(), getSLALogItem());
            requestValidator.validate(request);
            boolean odrExists = businessService.hasOrganDonorRegistration(requestContext);
            response.setRegistrationExists(odrExists);
            auditLogger.logHasOrganDonorRegistration(requestContext.getPersonIdentifier(), requestContext.getServiceActor());
        } catch (Exception e) {
            boolean isCprValidationError = requestValidator.getValidationErrors().stream()
                    .map(rve -> rve.cprValidation)
                    .reduce((val1, val2) -> val1 && val2)
                    .orElse(false);
            if (isCprValidationError) { response.setRegistrationExists(false); return response; }
            commonExceptionHandler(e, FaultCode.HAS_ORGANDONORREGISTRATION_VALIDATION_ERROR, FaultCode.HAS_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR);
        }
        return response;
    }
}
