/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.context;

import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.Organisation.OrganisationIdentifierFormat;
import dk.sds.nsp.security.SecurityContext.User;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultUtil;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;

@Component
public class ServiceActorProvider {

	private static final Logger logger = LogManager.getLogger(ServiceActorProvider.class);

	private final String allowedAudience;
	private final List<String> whiteListedLevel3Cvrs;

	@Autowired
	public ServiceActorProvider(Properties properties) {
		this(properties.getAllowedIdwsAudience(), properties.getWhitelistedLevel3Cvrs());
	}
	
	public ServiceActorProvider(String allowedAudience, List<String> whiteListedLevel3Cvrs) {
		this.allowedAudience = allowedAudience;
		this.whiteListedLevel3Cvrs = whiteListedLevel3Cvrs;
	}
	

	public ServiceActorVO getServiceActor(SecurityContext securityContext, boolean wantsAdmin) {
		return mapSecurityContextToValidatedServiceActorVO(securityContext, wantsAdmin);
	}

	private ServiceActorVO mapSecurityContextToValidatedServiceActorVO(SecurityContext securityContext, boolean wantsAdmin) {
		// Gennemløb de understøttede aktørtyper for denne service og valider og map
		
		// Tjekke for sundhedsfaglig og admin brugere
		if (securityContext.getActingUser().filter(u -> User.UserType.HealthcareProfessional.equals(u.getUserType())).isPresent()) {
			// Ingen delegering
			if (securityContext.getPrincipalUser().isPresent()) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}

			return mapToServiceActor(securityContext, wantsAdmin ? UserType.Admin : UserType.HealthCareProfessional, true);
		}
		// Borgerkald
		if (securityContext.getActingUser().filter(u -> User.UserType.Citizen.equals(u.getUserType())).isPresent()) {
			if (wantsAdmin) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED);
			}
			// Ingen fuldmagt for borgere
			if (securityContext.getPrincipalUser().isPresent()) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}
			// Der skal være audience tilstede
			if (!securityContext.getTicket().get().getAudience().isPresent()) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_INVALID_SECURITY_TICKET);
			}

	        // Validering af audience
			String audience = securityContext.getTicket().get().getAudience().get();
			if (!audience.equals(allowedAudience)) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_WRONG_AUDIENCE, "Gyldigt audience: " + allowedAudience);
			}

			return mapToServiceActor(securityContext, UserType.Citizen, false);
		}
		// Systembruger
		if (!securityContext.getActingUser().isPresent() && securityContext.getOrganisation().isPresent()) {
			
			if (wantsAdmin) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED);
			}

			ServiceActorVO result = mapToServiceActor(securityContext, UserType.SystemUser, true);
			
			// Er whitelisting ok?
			if (!whiteListedLevel3Cvrs.contains(result.getOrganizationId())) {
				FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_CVR_NOT_WHITELISTED);
			}
			return result;
		}

		FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_USERTYPE_NOT_SPECIFIED);
		return null;
	}


	private ServiceActorVO mapToServiceActor(SecurityContext securityContext, UserType userType, boolean requireCvr) {
		ServiceActorVO result = new ServiceActorVO(userType);
		if (securityContext.getOrganisation().isPresent()) {
			mapDataFromOrganisation(result, securityContext.getOrganisation().get());
		} else if (requireCvr) {
			FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_INGEN_CVR);
		}
		if (securityContext.getActingUser().isPresent()) {
			mapAndValidateDataFromUser(result, securityContext.getActingUser().get());
		}
		
		return result;
	}


	private void mapDataFromOrganisation(ServiceActorVO serviceActor, SecurityContext.Organisation organisation) {
		if (OrganisationIdentifierFormat.CVR.equals(organisation.getIdentifierFormat())) {
			serviceActor.setOrganizationId(organisation.getIdentifier());			
			serviceActor.setOrganizationIdSource("CVR");
			serviceActor.setOrganizationName(organisation.getName().orElse(null));			
		} else {
			FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_INGEN_CVR);
		}
	}

	private void mapAndValidateDataFromUser(ServiceActorVO serviceActor, SecurityContext.User user) {
		if (SecurityContext.User.PersonIdentifierFormat.CPR.equals(user.getIdentifierFormat())) {
			serviceActor.setPersonIdentifier(user.getIdentifier());
		} else {
			FaultUtil.throwODRFault(FaultCode.AUTHORIZATION_ERROR_USER_NOT_CPR);
		}
		serviceActor.setFirstName(user.getGivenName().orElse(null));
		serviceActor.setLastName(user.getSurname().orElse(null));
		if (user.getCredentials().isPresent()) {
			SecurityContext.User.Credentials credentials = user.getCredentials().get();
			serviceActor.setAuthorisationNumber(credentials.getAuthorizationCode().orElse(null));
			serviceActor.setEducationCode(credentials.getEducationCode().orElse(null));
		}
	}
}
