/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.CreateOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.UpdateOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.odr.OrganDonorRegistration;
import org.springframework.stereotype.Component;

import javax.xml.bind.ValidationException;

@Component
public class FromHL7CDAMapper extends AbstractHL7Mapper {

    private static final Logger log = LogManager.getLogger(FromHL7CDAMapper.class);

    public OrganDonorVO mapOrganDonorFromRequest(CreateOrganDonorRegistrationRequest request) throws ValidationException {
        return mapRequestToOrganDonorVO(request.getId().getExtension(), request.getOrganDonorRegistration());
    }

    public OrganDonorVO mapOrganDonorFromRequest(UpdateOrganDonorRegistrationRequest request) throws ValidationException {
        return mapRequestToOrganDonorVO(request.getId().getExtension(), request.getOrganDonorRegistration());
    }

    private OrganDonorVO mapRequestToOrganDonorVO(String personIdentifier, OrganDonorRegistration organDonorRegistration) throws ValidationException {
        OrganDonorVO organDonorVO = new OrganDonorVO();
        organDonorVO.setPersonIdentifier(personIdentifier);

        final org.hl7.odr.PermissionType permissionType = organDonorRegistration.getPermissionType();
        organDonorVO.setPermissionType(fromHL7PermissionType(permissionType));
        switch (permissionType) {
            case FULL:
            case FULL_WITH_RESEARCH:
                organDonorVO.setRelativeAcceptanceRequired(organDonorRegistration.isRequiresRelativeAcceptance());
                break;
            case LIMITED:
            case LIMITED_WITH_RESEARCH:
                setValuesLimitedPermission(organDonorVO, organDonorRegistration);
                organDonorVO.setRelativeAcceptanceRequired(organDonorRegistration.isRequiresRelativeAcceptance());
                break;
        }

        return organDonorVO;
    }

    private PermissionType fromHL7PermissionType(org.hl7.odr.PermissionType permissionType) throws ValidationException {
        switch (permissionType) {
            case FULL:
                return PermissionType.FULL;
            case FULL_WITH_RESEARCH:
                return PermissionType.FULL_WITH_RESEARCH;
            case LIMITED:
                return PermissionType.LIMITED;
            case LIMITED_WITH_RESEARCH:
                return PermissionType.LIMITED_WITH_RESEARCH;
            case NONE:
                return PermissionType.NONE;
            case DONT_KNOW:
                return PermissionType.DONT_KNOW;
            default:
                log.error("Unable to map schema enum value for permissionType element");
                throw new ValidationException("Value of permissionType element not matched");
        }
    }

    private void setValuesLimitedPermission(OrganDonorVO organDonorVO, OrganDonorRegistration organDonorRegistration) {
        organDonorVO.setHeart(organDonorRegistration.isPermissionForHeart());
        organDonorVO.setCorneas(organDonorRegistration.isPermissionForCornea());
        organDonorVO.setLungs(organDonorRegistration.isPermissionForLungs());
        organDonorVO.setKidneys(organDonorRegistration.isPermissionForKidneys());
        organDonorVO.setSkin(organDonorRegistration.isPermissionForSkin());
        organDonorVO.setLiver(organDonorRegistration.isPermissionForLiver());
        organDonorVO.setPancreas(organDonorRegistration.isPermissionForPancreas());
        organDonorVO.setSmallIntestine(organDonorRegistration.isPermissionForSmallIntestine());

    }
}
