/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda;

import dk.sundhedsdatastyrelsen.organdonor.OID;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.v3.*;
import org.springframework.stereotype.Component;

import java.math.BigInteger;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;

@Component
public class ToHL7CDAMapper extends AbstractHL7Mapper {

    private static final Logger log = LogManager.getLogger(ToHL7CDAMapper.class);

    protected static ObjectFactory o = new ObjectFactory();

    public POCDMT000040ClinicalDocument toClinicalDocument(String personIdentifier, OrganDonorVO odr, RequestContextVO context) {
        final POCDMT000040ClinicalDocument doc = createDocument();
        doc.setEffectiveTime(createTimestamp(context.getTime()));
        doc.setVersionNumber(createVersionNumber(odr.getVersion()));
        doc.getRecordTargets().add(createRecordTarget(personIdentifier));
        doc.getAuthors().add(createAuthor(personIdentifier, odr.getModifiedDateTime()));
        doc.setCustodian(createCustodian());
        doc.setComponent(createComponent(odr));
        return doc;
    }

    private POCDMT000040ClinicalDocument createDocument() {
        final POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();
        doc.setClassCode(ActClinicalDocument.DOCCLIN);
        doc.getMoodCodes().add(MOOD_CODE_EVENT);

        final CS dk = new CS();
        dk.setCode(REALM_CODE_DK);
        doc.getRealmCodes().add(dk);
        doc.setTypeId(createTypeId(OID.HL7_REGISTEREDE_RMIMS, "POCD_HD000040"));

        final II templateId = new II();
        templateId.setRoot(OID.ORGANDONOR_CODE_SYSTEM);
        doc.getTemplateIds().add(templateId);
        doc.setId(createId(OID.ISO_MEMBERBODY_DK_MEDCOM, "NA", MEDCOM));
        doc.setCode(createCode(ORGANDONOR, OID.ORGANDONOR_CODE_SYSTEM, ORGANDONORREGISTRATION));
        doc.setTitle(createText(ORGANDONORREGISTRATION));
        doc.setConfidentialityCode(createCode(CONFIDENTIALITY_CODE, OID.CONFIDENTIALITYCODE));
        doc.setLanguageCode(createCode(LANGUAGE_CODE_DK));
        return doc;
    }

    private POCDMT000040Component2 createComponent(OrganDonorVO organDonorVO) {
        final POCDMT000040Component2 component = o.createPOCDMT000040Component2();
        final POCDMT000040StructuredBody structuredBody = new POCDMT000040StructuredBody();
        final POCDMT000040Component3 bodyComponent = new POCDMT000040Component3();
        final POCDMT000040Section section = new POCDMT000040Section();
        final StrucDocText odr = new StrucDocText();
        odr.getContent().add("Organdonor registrering");
        section.setText(odr);

        POCDMT000040Observation observation = o.createPOCDMT000040Observation();
        observation.getClassCodes().add(AbstractHL7Mapper.OBSERVATION_CLASSCODE);
        observation.setMoodCode(XActMoodDocumentObservation.EVN);

        observation.setCode(createCode(ORGANDONOR, OID.ORGANDONOR_CODE_SYSTEM, ORGANDONORREGISTRATION));
        observation.getValues().add(mapOrganDonorRegistrationValues(organDonorVO));
        POCDMT000040Entry entry = o.createPOCDMT000040Entry();
        entry.setObservation(observation);
        section.getEntries().add(entry);
        bodyComponent.setSection(section);
        structuredBody.getComponents().add(bodyComponent);
        component.setStructuredBody(structuredBody);

        return component;
    }

    private OrganDonorRegistration mapOrganDonorRegistrationValues(OrganDonorVO organDonorVO) {
        OrganDonorRegistration odr = new OrganDonorRegistration();
        odr.setPermissionType(toHL7PermissionType(organDonorVO.getPermissionType()));
        odr.setPermissionForHeart(organDonorVO.getHeart());
        odr.setPermissionForLungs(organDonorVO.getLungs());
        odr.setPermissionForLiver(organDonorVO.getLiver());
        odr.setPermissionForPancreas(organDonorVO.getPancreas());
        odr.setPermissionForKidneys(organDonorVO.getKidneys());
        odr.setPermissionForCornea(organDonorVO.getCorneas());
        odr.setPermissionForSmallIntestine(organDonorVO.getSmallIntestine());
        odr.setPermissionForSkin(organDonorVO.getSkin());
        odr.setRequiresRelativeAcceptance(organDonorVO.getRelativeAcceptanceRequired());

        return odr;
    }

    private org.hl7.odr.PermissionType toHL7PermissionType(PermissionType permissionType) {
        switch (permissionType) {
            case FULL:
                return org.hl7.odr.PermissionType.FULL;
            case FULL_WITH_RESEARCH:
                return org.hl7.odr.PermissionType.FULL_WITH_RESEARCH;
            case LIMITED:
                return org.hl7.odr.PermissionType.LIMITED;
            case LIMITED_WITH_RESEARCH:
                return org.hl7.odr.PermissionType.LIMITED_WITH_RESEARCH;
            case NONE:
                return org.hl7.odr.PermissionType.NONE;
            case DONT_KNOW:
                return org.hl7.odr.PermissionType.DONT_KNOW;
            default:
                log.error("Unable to map permissionType to permissionType element");
                throw new IllegalStateException("Internal value of permissionType element not matched");
        }
    }

    private POCDMT000040Author createAuthor(String personIdentifier, Instant timestamp) {
        final POCDMT000040Author author = o.createPOCDMT000040Author();

        final POCDMT000040AssignedAuthor assignedAuthor = o.createPOCDMT000040AssignedAuthor();
        assignedAuthor.getIds().add(createId(OID.CPR, personIdentifier, CPR));
        author.setAssignedAuthor(assignedAuthor);
        author.setTime(createTimestamp(timestamp));
        return author;
    }

    private void addEmptyComponent(POCDMT000040ClinicalDocument clinicalDocument) {
        final POCDMT000040Component2 component = new POCDMT000040Component2();
        final POCDMT000040NonXMLBody nonXMLBody = new POCDMT000040NonXMLBody();
        final ST st = new ST();
        st.getNullFlavors().add("NA");
        nonXMLBody.setText(st);
        component.setNonXMLBody(nonXMLBody);
        clinicalDocument.setComponent(component);
    }

    private POCDMT000040RecordTarget createRecordTarget(String personIdentifier) {
        final POCDMT000040RecordTarget recordTarget = o.createPOCDMT000040RecordTarget();
        recordTarget.getTypeCodes().add("RCT");
        recordTarget.setContextControlCode("OP");

        final POCDMT000040PatientRole patientRole = o.createPOCDMT000040PatientRole();
        patientRole.getClassCodes().add("PAT");

        // cprnr
        if (personIdentifier != null) {
            patientRole.getIds().add(createId(OID.CPR, personIdentifier, CPR));
        }
        recordTarget.setPatientRole(patientRole);

        return recordTarget;
    }

    private INT createVersionNumber(int version) {
        final INT versionNumber = o.createINT();
        versionNumber.setValue(BigInteger.valueOf(version));
        return versionNumber;
    }

    private POCDMT000040Custodian createCustodian() {
        final POCDMT000040Custodian custodian = o.createPOCDMT000040Custodian();
        final POCDMT000040AssignedCustodian assignedCustodian = o.createPOCDMT000040AssignedCustodian();
        final POCDMT000040CustodianOrganization custodianOrganisation = o.createPOCDMT000040CustodianOrganization();
        final II id = o.createII();
        id.setRoot("NA");
        custodianOrganisation.getIds().add(id);
        assignedCustodian.setRepresentedCustodianOrganization(custodianOrganisation);
        custodian.setAssignedCustodian(assignedCustodian);

        return custodian;
    }

    public TS createTimestamp(Instant instant) {
        final TS ts = o.createTS();
        final DateTimeFormatter formatter = DateTimeFormatter.ofPattern(TS_FORMAT).withZone(ZoneId.systemDefault());
        String format = formatter.format(instant);
        ts.setValue(format);
        return ts;
    }

    private ST createText(String text) {
        final ST st = o.createST();
        st.getContent().add(text);
        return st;
    }

    private CS createCode(String code) {
        final CS c = o.createCS();
        c.setCode(code);
        return c;
    }

    protected CE createCode(String code, String codeSystem) {
        return createCode(code, codeSystem, null);
    }

    protected CE createCode(String code, String codeSystem, String displayName) {
        final CE c = o.createCE();
        c.setCode(code);
        c.setCodeSystem(codeSystem);
        c.setDisplayName(displayName);
        return c;
    }

    public II createId(String root, String extension, String assigningAuth) {
        final II id = o.createII();
        if (root != null) {
            id.setRoot(root);
        }
        if (extension != null) {
            id.setExtension(extension);
        }
        if (assigningAuth != null) {
            id.setAssigningAuthorityName(assigningAuth);
        }
        return id;
    }

    private POCDMT000040InfrastructureRootTypeId createTypeId(String root, String extension) {
        final POCDMT000040InfrastructureRootTypeId typeId = new POCDMT000040InfrastructureRootTypeId();
        typeId.setRoot(root);
        typeId.setExtension(extension);
        return typeId;
    }
}
