/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation;

import dk.sundhedsdatastyrelsen.organdonor.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationServiceException;
import org.apache.log4j.Logger;

public class CprExistsImpl implements CprExists {

	private static final Logger LOGGER = Logger.getLogger(CprExistsImpl.class);

	private CprExistsServiceClient cprExistsServiceClient;

	private final CprExists.ValidationLevel validationLevel;

	private final AuditLogger auditLogger;

	private final int cprMinValidationAge;

	public CprExistsImpl(CprExists.ValidationLevel validationLevel, CprExistsServiceClientFactory cprExistsServiceClientFactory, AuditLogger auditLogger, int cprMinValidationAge) {
		this.auditLogger = auditLogger;
		this.validationLevel = validationLevel;
		this.cprExistsServiceClient = cprExistsServiceClientFactory.createClient();
		this.cprMinValidationAge = cprMinValidationAge;
	}

	@Override
	public void setServiceClient(CprExistsServiceClient serviceClient) { cprExistsServiceClient = serviceClient; }

	@Override
	public boolean validateCpr(String patientId) throws ValidationException {

		if (!ValidationLevel.OFF.equals(validationLevel)) {
			try {
				CprStatus cprStatus = cprExistsServiceClient.validateCpr(patientId);
				boolean cprOk = (cprStatus != null && !cprStatus.isInactive());
				if (!cprOk && ValidationLevel.WARNING.equals(validationLevel)) {
					if (cprStatus == null) {
						auditLogger.logCprDoesNotExist(patientId);
					} else {
						auditLogger.logCprInactive(patientId);
					}

					return false;
				} else if (!cprOk && ValidationLevel.REJECT.equals(validationLevel)) {
					if (cprStatus == null) {
						throw new ValidationException("Patientid eller status eksisterer ikke");
					} else {
						throw new ValidationException("Patientens status er inaktiv");
					}
				}
			} catch (ValidationServiceException e) {
				LOGGER.error("Kaldet til Cprexist fejlede", e);
				throw new ValidationException(e.getMessage());
			}
		}
		return true;
	}

	@Override
	public void validateAge(String patientId) throws ValidationException {
		int age = cprExistsServiceClient.getAge(patientId);
		if (age < cprMinValidationAge) {
				throw new ValidationException("Patienten afvises, da patienten er " + age +
						" år gammel, men minimum skal være " + cprMinValidationAge + " år gammel.");
			}
	}
}
