/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.util.HttpClientFactory;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationServiceException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.List;
import java.util.Map;

public class CprExistsServiceClient {

	public static final String APPLICATION_JSON = "application/json";

	private static final String CONTENT_TYPE = "Content-Type";

	private static final Logger LOGGER = Logger.getLogger(CprExistsServiceClient.class);

	private final ObjectMapper om = new ObjectMapper();

	private final CprExistsServiceClientFactory.ServiceErrorCallback serviceErrorCallback;

	private final HttpClientFactory httpClientFactory;

	private final String url;
	
	private List<Integer> inactiveStatusList;

	public CprExistsServiceClient(HttpClientFactory httpClientFactory, String url, CprExistsServiceClientFactory.ServiceErrorCallback serviceErrorCallback, List<Integer> inactiveStatusList) {
		this.url = url;
		this.httpClientFactory = httpClientFactory;
		this.serviceErrorCallback = serviceErrorCallback;
		this.inactiveStatusList = inactiveStatusList;
	}

	public CprStatus validateCpr(String patientId) {

		try {
			HttpUriRequest cprRequest = buildStatusRequest(patientId);
			CprStatus cprStatus = (CprStatus)invokeService(cprRequest);
			serviceErrorCallback.registerServiceSucces();
			return cprStatus;
		} catch (ValidationServiceException e) {
			LOGGER.error("Fejl i kald til CprExists: ", e);
			serviceErrorCallback.registerServiceError();
			throw e;
		}
	}

	public int getAge(String patientId) {
		try {
			HttpUriRequest ageRequest = buildAgeRequest(patientId);
			int age = (int)invokeService(ageRequest);
			serviceErrorCallback.registerServiceSucces();
			return age;
		} catch (ValidationServiceException e) {
			LOGGER.error("Fejl i kald til CprExists: ", e);
			serviceErrorCallback.registerServiceError();
			throw e;
		}
	}

	private HttpUriRequest buildStatusRequest(String cpr) {
		// Example request string: http://localhost:8086/stamdata-cprexists/person/1212121212/status
		return new HttpGet(String.format("%s/person/%s/status", url, cpr));
	}

	private HttpUriRequest buildAgeRequest(String cpr) {
		// Example request string: http://localhost:8086/stamdata-cprexists/person/1212121212/age
		return new HttpGet(String.format("%s/person/%s/age", url, cpr));
	}


	private Object invokeService(HttpUriRequest request) {
		// Construct client
		HttpClient client = getClient();

		// Send request
		try {
			// Using a ResponseHandler makes the httpclient framework handle resource allocation/deallocation for us.
			// If you decide to not use a ResponseHandler anymore, you must ensure that resources (connections) are properly released!
			return client.execute(request, this::handleResponse);
		}
		catch(IOException e) {        	
			throw new ValidationServiceException("CPR-nummeret kunne ikke valideres, fordi kaldet til cprexists-servicen gik galt.", e);
		}
	}

	private HttpClient getClient() {
		return httpClientFactory.getClient();
	}

	private Object handleResponse(HttpResponse response) {
		int statusCode = response.getStatusLine().getStatusCode();
		CprStatus cprStatus = null; //return null if cpr or status not found. Otherwise set status to inactive or not
		final String errorDescription =
				"CPR-nummeret eller alder kunne ikke valideres, fordi kaldet til cprexists-servicen returnerede en " +
						"uventet statuskode: %d";
		switch(statusCode) {
		case HttpStatus.SC_OK: {
			boolean contentTypeOk = contentTypeOk(response);
			Map<String, Object> responseJson = parseResponseAsJson(response);
			if  (!contentTypeOk || (!responseJson.containsKey("status") && !responseJson.containsKey("age"))) {
				LOGGER.error("Fejl i retursvar fra CPRExist (200)");
				throw new ValidationServiceException("Fejl i retursvar fra CPRExist");
			}
			if (responseJson.get("status") instanceof Integer) {
				int status = (int) responseJson.get("status");
				LOGGER.debug("Status found for cpr is: " + status);
				cprStatus = new CprStatus(inactiveStatusList.contains(status));
				return cprStatus;
			}
			if (responseJson.get("age") instanceof Integer) {
				int age = (int) responseJson.get("age");
				LOGGER.debug("Age found for cpr is: " + age);
				return age;
			}
			LOGGER.error("Status found for cpr is not a number.");
			return cprStatus;
			
		}
		case HttpStatus.SC_NOT_FOUND: {
			boolean contentTypeOk = contentTypeOk(response);
			if  (!contentTypeOk) {
				LOGGER.error("Fejl i retursvar fra CPRExist (404) forkert contenttype");
				throw new ValidationServiceException("Fejl i retursvar fra CPRExist");
			}
			parseResponseAsJson(response);
			throw new ValidationServiceException(String.format(errorDescription, statusCode));
		}
		default:
			throw new ValidationServiceException(String.format(errorDescription, statusCode));
		}
	}

	private Map<String, Object> parseResponseAsJson(HttpResponse response) {
		try {
			Map<String, Object> responseJson = om.readValue(response.getEntity().getContent(), Map.class);
			return responseJson;
		} catch (UnsupportedOperationException | IOException e) {
			LOGGER.error("Kunne ikke parse response body til json", e);
			throw new ValidationServiceException("Fejl i fortolkning af svar fra CPRExist", e);
		}
	}

	private boolean contentTypeOk(HttpResponse response) {
		Header headers = response.getFirstHeader(CONTENT_TYPE);
		return (headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON));
	}
}
