/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation;

import dk.sundhedsdatastyrelsen.organdonor.OID;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.*;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ParseException;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.AbstractHL7Mapper;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.odr.PermissionType;
import org.hl7.v3.*;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.AbstractHL7Mapper.CPR;
import static dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.AbstractHL7Mapper.TS_FORMAT;

public class RequestValidator {
    private static final Logger log = LogManager.getLogger(RequestValidator.class);

    private final Properties properties;
    protected CprExists cprExists;
    private List<RequestValidatorError> validationErrors;

    public static class RequestValidatorError {
        public final String description;
        public final Boolean cprValidation;

        public RequestValidatorError(String description, Boolean cprValidation) {
            this.description = description;
            this.cprValidation = cprValidation;
        }
    }

    public RequestValidator(CprExists cprExists, Properties properties) {
        this.cprExists = cprExists;
        this.properties = properties;
        clearValidationErrors();
    }

    public List<RequestValidatorError> getValidationErrors() {
        return validationErrors;
    }

    public void clearValidationErrors() {
        validationErrors = new LinkedList<>();
    }

    public String getCombinedErrorText() {
        return String.join(", ",
                validationErrors.stream()
                        .map(rve -> rve.description)
                        .collect(Collectors.toList())
        );
    }

    public void validate(CreateOrganDonorRegistrationRequest request) {
        try {
            validateCreateOrganDonorRegistration(request, "Opret Organdonorregistrering");
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Opret Organdonorregistrering", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(UpdateOrganDonorRegistrationRequest request) {
        try {
            validateUpdateOrganDonorRegistration(request, "Updater Organdonorregistrering");
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Updater Organdonorregistrering", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(GetOrganDonorRegistrationRequest request) {
        try {
            validateId(request.getId(), "Hent Organdonorregistrering");
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Hent Organdonorregistrering", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(DeleteOrganDonorRegistrationRequest request) {
        try {
            validateId(request.getId(), "Slet Organdonorregistrering");
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Slet Organdonorregistrering", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(HasOrganDonorRegistrationRequest request) {
        try {
            validateId(request.getId(), "Har Organdonorregistrering");
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Har Organdonorregistrering", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validateWithoutThrow(GetOrganDonorRegistrationResponse response) {
        try {
            validateClinicalDocument(response.getClinicalDocument(), "clinicalDocument");
        } catch (Exception e) {
            validationErrors.add(buildUnspecificErrorMessage("Organdonorregistrering response", e.getMessage()));
        }
    }

    private void validateCreateOrganDonorRegistration(CreateOrganDonorRegistrationRequest request, String errorMessage) {
        validateOrganDonorRegistration(request.getId(), request.getOrganDonorRegistration(), errorMessage
        );
    }

    private void validateUpdateOrganDonorRegistration(UpdateOrganDonorRegistrationRequest request, String errorMessage) {
        validateOrganDonorRegistration(request.getId(), request.getOrganDonorRegistration(), errorMessage
        );
    }

    private void validateOrganDonorRegistration(II id, OrganDonorRegistration organDonorRegistration,
                                                String errorMessage) {
        try {
            validateId(id, "Id");
            validateOrganDonorRegistration(organDonorRegistration, "Organdonorregistrering");
        } catch (Exception e) {
            validationErrors.add(new RequestValidatorError(errorMessage + e.getMessage(), false));
        }
    }

    private void validateOrganDonorRegistration(OrganDonorRegistration odr, String elementName) {
        if (!checkOdrExists(odr, elementName)) return;
        validateValuesInRequest(odr, elementName);
    }

    private void validateValuesInRequest(OrganDonorRegistration value, String elementName) {
        if (validatePermissionType(value.getPermissionType(), elementName)) {
            validateValuesBasedOnPermissionType(value, elementName);
        }
    }

    private void validateValuesBasedOnPermissionType(OrganDonorRegistration value, String elementName) {
        String permissionTypeValue = value.getPermissionType().value();

        if (permissionTypeValue.equals(PermissionType.LIMITED.value()) || permissionTypeValue.equals(PermissionType.LIMITED_WITH_RESEARCH.value())) {
            validateLimitedPermissionType(value, elementName);
        } else if (permissionTypeValue.equals(PermissionType.FULL.value()) || permissionTypeValue.equals(PermissionType.FULL_WITH_RESEARCH.value())) {
            validateFullPermissionTypes(value, elementName);
        }
    }

    private void validateLimitedPermissionType(OrganDonorRegistration value, String elementName) {
        if (!validateRelativeAcceptance(value.isRequiresRelativeAcceptance(), elementName)) {
            return;
        }
        validateOrganDonorValues(value, elementName);
    }

    private void validateFullPermissionTypes(OrganDonorRegistration value, String elementName) {
        if (value.getPermissionType().value().equals(PermissionType.FULL.value()) || value.getPermissionType().value().equals(PermissionType.FULL_WITH_RESEARCH.value())) {
            validatePermissionTypes(value, elementName, true);
            return;
        }
    }

    private void validatePermissionTypes(OrganDonorRegistration value, String elementName, boolean isRequired) {
        if (isRequired) {
            if (!validateRelativeAcceptance(value.isRequiresRelativeAcceptance(), elementName)) {
                return;
            }
        }
        validateOrganDonorValues(value, elementName);

    }

    private boolean validatePermissionType(PermissionType permissionType, String elementName) {
        if (checkThatExists(permissionType.value(), elementName + ".persmissionType")) {
            List<String> listOfEnumValues = Stream.of(PermissionType.values()).map(Enum::name).collect(Collectors.toList());
            if (checkThatExistsAndContainsSupportedValue(permissionType.value(), listOfEnumValues, elementName + ".permissionType")) {
                return true;
            }
        }
        return false;
    }

    private boolean validateRelativeAcceptance(Boolean requiresRelativeAcceptance, String elementName) {
        if (checkThatExist(requiresRelativeAcceptance, elementName + ".relativeAcceptance")) {
            if (checkThatExistsAndContainsSupportedValue(requiresRelativeAcceptance, Arrays.asList(true, false), elementName + ".relativeAcceptance")) {
                return true;
            }
        }
        return false;
    }

    private void validateOrganDonorValues(OrganDonorRegistration values, String elementName) {
        Map<String, Boolean> map = new HashMap<>();
        map.put(".heart", values.isPermissionForHeart());
        map.put(".cornea", values.isPermissionForCornea());
        map.put(".kidneys", values.isPermissionForKidneys());
        map.put(".liver", values.isPermissionForLiver());
        map.put(".lungs", values.isPermissionForLungs());
        map.put(".skin", values.isPermissionForSkin());
        map.put(".pancreas", values.isPermissionForPancreas());
        map.put(".smallIntestine", values.isPermissionForSmallIntestine());

        if (values.getPermissionType() == PermissionType.LIMITED || values.getPermissionType() == PermissionType.LIMITED_WITH_RESEARCH) {
            boolean contains = map.values().stream().allMatch(Objects::isNull);
            if (contains) {
                validationErrors.add(new RequestValidatorError(ValidationErrors.VALUE_REQUIRED, false));
                return;
            }
            map.forEach((s, value) -> {
                if (checkThatExist(value)) {
                    checkThatExistsAndContainsSupportedValue(value, Arrays.asList(true, false), elementName + s);
                }
            });
            return;
        }
    }

    private void validateClinicalDocument(POCDMT000040ClinicalDocument doc, String elementName) {
        final String classCodeElementName = elementName + ".classCode";
        if (checkThatExists(doc.getClassCode(), classCodeElementName)) {
            checkThatExistsAndContainsSupportedValue(doc.getClassCode().value(), AbstractHL7Mapper.CLASS_CODE_DOCCLIN, classCodeElementName);
        }
        checkThatListContainsSingleElementWithSpecificValue(doc.getMoodCodes(), AbstractHL7Mapper.MOOD_CODE_EVENT, elementName + ".moodCodes");

        validateCSListWithSingleElement(doc.getRealmCodes(), AbstractHL7Mapper.REALM_CODE_DK, elementName + ".realmCode");
        validateId(doc.getTypeId(), true, false, false,elementName + ".typeId");
        validateTemplateIds(doc.getTemplateIds(), elementName + ".templateIds");
        validateCD(doc.getCode(), OID.ORGANDONOR_CODE_SYSTEM, elementName + ".code");
        checkThatExists(doc.getTitle(), elementName + ".title");
        validateTime(doc.getEffectiveTime(), elementName + ".effectiveTime");
        validateCE(doc.getConfidentialityCode(), AbstractHL7Mapper.CONFIDENTIALITY_CODE, OID.CONFIDENTIALITYCODE, elementName + ".confidentialityCode");
        validateCS(doc.getLanguageCode(),  AbstractHL7Mapper.LANGUAGE_CODE_DK, elementName + ".languageCode");
        checkThatExists(doc.getVersionNumber(), elementName + ".versionNumber");
        validateRecordTargets(doc.getRecordTargets(), elementName + ".recordTargets");
        validateAuthors(doc.getAuthors(), elementName + ".authors");
        validateCustodian(doc.getCustodian(), elementName + ".custodian");
        validateClinicalDocumentComponent(doc.getComponent(), elementName + ".component");


        checkThatListDoesNotExistOrContainElements(doc.getAuthenticators(), elementName + ".authenticators");
        checkThatListDoesNotExistOrContainElements(doc.getAuthorizations(), elementName + ".authorizations");
        checkThatDoesNotExist(doc.getComponentOf(), elementName + ".componentOf");
        checkThatDoesNotExist(doc.getCopyTime(), elementName + "copyTime");
        checkThatDoesNotExist(doc.getDataEnterer(), elementName + ".dataEnterer");
        checkThatListDoesNotExistOrContainElements(doc.getDocumentationOves(), elementName + ".documentationOves");
        checkThatListDoesNotExistOrContainElements(doc.getInformants(), elementName + ".informants");
        checkThatListDoesNotExistOrContainElements(doc.getInformationRecipients(), elementName + ".informationRecipients");
        checkThatListDoesNotExistOrContainElements(doc.getInFulfillmentOves(), elementName + ".inFulfillmentOves");
        checkThatDoesNotExist( doc.getLegalAuthenticator(), elementName + ".legalAuthenticator");
        checkThatListDoesNotExistOrContainElements(doc.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(doc.getParticipants(), elementName + ".participants");
        checkThatListDoesNotExistOrContainElements(doc.getRelatedDocuments(), elementName + ".relatedDocuments");
    }

    private void validateCSListWithSingleElement(List<CS> codes, String code, String elementName) {
        if (checkThatListContainsSingleElement(codes, elementName)) {
            validateCS(codes.get(0), code, elementName);
        }
    }

    private void validateTemplateIds(List<II> templateIds, String elementName) {
        if (checkThatListContainsSingleElement(templateIds, elementName + ".templateId")) {
            validateId(templateIds.get(0), false, false, false,elementName + ".templateId");
        }
    }

    private void validateCS(CS cs, String code, String elementName) {
        checkThatExistsAndContainsSupportedValue(cs.getCode(), code, elementName + ".code");

        checkThatListDoesNotExistOrContainElements(cs.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(cs.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(cs.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(cs.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(cs.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(cs.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(cs.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(cs.getValueSetVersion(), elementName + ".valueSetVersion");
    }

    private void validateCD(CD cd, String expectedCodeSystem, String elementName) {
        if (!checkThatExists(cd, elementName)) return;
        checkThatExists(cd.getCode(), elementName + ".code");
        checkThatExists(cd.getDisplayName(), elementName + ".displayName");
        checkThatExistsAndContainsSupportedValue(cd.getCodeSystem(), Arrays.asList(expectedCodeSystem),
                elementName + ".codeSystem");

        checkThatListDoesNotExistOrContainElements(cd.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(cd.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(cd.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(cd.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(cd.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(cd.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(cd.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(cd.getValueSetVersion(), elementName + ".valueSetVersion");
    }

    private void validateTime(TS time, String elementName) {
        if (!checkThatExists(time, elementName)) return;
        if (checkThatExists(time.getValue(), elementName + ".value")) {
            checkDateTimeIsValid(time.getValue(), elementName + ".value");
        }

        checkThatListDoesNotExistOrContainElements(time.getNullFlavors(), elementName + ".nullFlavors");
    }

    private void validateCE(CE ce, String expectedCode, String expectedCodeSystem, String elementName) {
        validateCE(ce, Arrays.asList(expectedCode), expectedCodeSystem, elementName);
    }

    private void validateCE(CE ce, List<String> expectedCodes, String expectedCodeSystem, String elementName) {
        checkThatExistsAndContainsSupportedValue(ce.getCode(), expectedCodes, elementName + ".code");
        checkThatExistsAndContainsSupportedValue(ce.getCodeSystem(), expectedCodeSystem, elementName + ".codeSystem");

        checkThatListDoesNotExistOrContainElements(ce.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(ce.getCodeSystemName(), elementName + ".codeSystemName");
        checkThatDoesNotExist(ce.getCodeSystemVersion(), elementName + ".codeSystemVersion");
        checkThatDoesNotExist(ce.getOriginalText(), elementName + ".originalText");
        checkThatListDoesNotExistOrContainElements(ce.getQualifiers(), elementName + ".qualifiers");
        checkThatListDoesNotExistOrContainElements(ce.getTranslations(), elementName + ".translations");
        checkThatDoesNotExist(ce.getValueSet(), elementName + ".valueSet");
        checkThatDoesNotExist(ce.getValueSetVersion(), elementName + ".valueSetVersion");
    }

    private void validateAuthors(List<POCDMT000040Author> authors, String elementName) {
        if (checkThatListContainsSingleElement(authors, elementName)) {
            validateAuthor(authors.get(0), elementName);
        }
    }

    private void validateAuthor(POCDMT000040Author author, String elementName) {
        validateTime(author.getTime(), elementName + ".time");
        validateAssignedAuthor(author.getAssignedAuthor(), elementName + ".assignedAuthor");

        checkThatDoesNotExist(author.getFunctionCode(), elementName + ".functionCode");
        checkThatDoesNotExist(author.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(author.getTypeCodes(), elementName + ".typeCodes");
        checkThatListDoesNotExistOrContainElements(author.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(author.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(author.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateAssignedAuthor(POCDMT000040AssignedAuthor assignedAuthor, String elementName) {
        if (!checkThatExists(assignedAuthor, elementName)) return;
        validateAssignedAuthorId(assignedAuthor.getIds(), elementName);

        checkThatDoesNotExist(assignedAuthor.getAssignedAuthoringDevice(), elementName + ".assignedAuthoringDevice");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getTelecoms(), elementName + ".telecom");
        checkThatDoesNotExist(assignedAuthor.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(assignedAuthor.getCode(), elementName + ".code");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(assignedAuthor.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateAssignedAuthorId(List<II> ids, String elementName) {
        final String idElement = ".id";
        if (checkThatListContainsSingleElement(ids, elementName + idElement)) {
            validateCPR(ids.get(0), elementName, false);
        }
    }

    private void validateCustodian(POCDMT000040Custodian custodian, String elementName) {
        validateAssignedCustodian(custodian.getAssignedCustodian(), elementName + ".assignedCustodian");

        checkThatDoesNotExist(custodian.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(custodian.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(custodian.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(custodian.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateAssignedCustodian(POCDMT000040AssignedCustodian assignedCustodian, String elementName) {
        validateRepresentedCustodianOrganization(assignedCustodian.getRepresentedCustodianOrganization(), elementName + ".representedCustodianOrganization");

        checkThatDoesNotExist(assignedCustodian.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(assignedCustodian.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateRepresentedCustodianOrganization(POCDMT000040CustodianOrganization reprCustodianOrg, String elementName) {
        checkThatListContainsSingleElement(reprCustodianOrg.getIds(), elementName + ".id");

        checkThatDoesNotExist(reprCustodianOrg.getAddr(), elementName + ".addr");
        checkThatDoesNotExist(reprCustodianOrg.getTelecom(), elementName + ".telecom");
        checkThatDoesNotExist(reprCustodianOrg.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(reprCustodianOrg.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(reprCustodianOrg.getName(), elementName + ".name");
    }

    private void validateClinicalDocumentComponent(POCDMT000040Component2 component, String elementName) {
        validateStructuredBody(component.getStructuredBody(), elementName + ".structuredBody");

        checkThatDoesNotExist(component.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(component.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(component.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(component.getTemplateIds(), elementName + ".templateIds");

    }

    private void validateStructuredBody(POCDMT000040StructuredBody structuredBody, String elementName) {
        if (checkThatListContainsSingleElement(structuredBody.getComponents(), elementName + ".component")) {
            validateStructuredBodyComponent(structuredBody.getComponents().get(0), elementName + ".component");
        }

        checkThatListDoesNotExistOrContainElements(structuredBody.getClassCodes(), elementName + ".classCodes");
        checkThatDoesNotExist(structuredBody.getConfidentialityCode(), elementName + ".confidentialityCode");
        checkThatDoesNotExist(structuredBody.getLanguageCode(), elementName + ".languageCode");
        checkThatListDoesNotExistOrContainElements(structuredBody.getMoodCodes(), elementName + ".moodCodes");
        checkThatListDoesNotExistOrContainElements(structuredBody.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(structuredBody.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(structuredBody.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(structuredBody.getTypeId(), elementName + ".typeId");
    }

    private void validateStructuredBodyComponent(POCDMT000040Component3 component, String elementName) {
        validateComponentSection(component.getSection(), elementName + ".section");

        checkThatDoesNotExist(component.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(component.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(component.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(component.getTemplateIds(), elementName + ".templateIds");
    }

    private void validateComponentSection(POCDMT000040Section section, String elementName) {
        checkThatExists(section.getText(), elementName + ".text");
        validateEntry(section.getEntries(), elementName + ".entry");

        checkThatListDoesNotExistOrContainElements(section.getAuthors(), elementName + ".authors");
        checkThatListDoesNotExistOrContainElements(section.getClassCodes(), elementName + ".classCodes");
        checkThatDoesNotExist(section.getCode(), elementName + ".code");
        checkThatListDoesNotExistOrContainElements(section.getComponents(), elementName + ".components");
        checkThatDoesNotExist(section.getConfidentialityCode(), elementName + ".confidentialityCode");
        checkThatDoesNotExist(section.getId(), elementName + ".id");
        checkThatDoesNotExist(section.getId2(), elementName + ".id2");
        checkThatListDoesNotExistOrContainElements(section.getInformants(), elementName + ".informants");
        checkThatDoesNotExist(section.getLanguageCode(), elementName + ".languageCode");
        checkThatListDoesNotExistOrContainElements(section.getMoodCodes(), elementName + ".moodCodes");
        checkThatListDoesNotExistOrContainElements(section.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(section.getRealmCodes(), elementName + ".realmCodes");
        checkThatDoesNotExist(section.getSubject(), elementName + ".subject");
        checkThatListDoesNotExistOrContainElements(section.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(section.getTitle(), elementName + ".title");
        checkThatDoesNotExist(section.getTypeId(), elementName + ".typeId");
    }

    private void validateEntry(List<POCDMT000040Entry> entries, String elementName) {
        if (checkThatListContainsSingleElement(entries, elementName)) {
            validateEntry(entries.get(0), elementName);
        }
    }

    private void validateEntry(POCDMT000040Entry entry, String elementName) {
        if (checkThatExists(entry.getObservation(), elementName + ".observation")) {
            validateObservationInResponse(entry.getObservation(), elementName + ".observation");
        }

        checkThatDoesNotExist(entry.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(entry.getNullFlavors(), elementName + ".nullFlavors");
        checkThatListDoesNotExistOrContainElements(entry.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(entry.getTemplateIds(), elementName + ".templateIds");
        checkThatDoesNotExist(entry.getAct(), elementName + ".act");
        checkThatDoesNotExist(entry.getEncounter(), elementName + ".encounter");
        checkThatDoesNotExist(entry.getObservationMedia(), elementName + ".observationMedia");
        checkThatDoesNotExist(entry.getOrganizer(), elementName + ".organizer");
        checkThatDoesNotExist(entry.getProcedure(), elementName + ".procedure");
        checkThatDoesNotExist(entry.getRegionOfInterest(), elementName + ".regionOfInterest");
        checkThatDoesNotExist(entry.getSubstanceAdministration(), elementName + ".substanceAdministration");
        checkThatDoesNotExist(entry.getSupply(), elementName + ".supply");
    }

    private void validateObservationInResponse(POCDMT000040Observation observation, String elementName) {
        validateClassCodes(observation.getClassCodes(), elementName + ".classCodes");
        validateMoodCode(observation.getMoodCode(), elementName + ".moodCode");
        validateCD(observation.getCode(), OID.ORGANDONOR_CODE_SYSTEM, elementName + ".code");
        validateValues(observation.getValues(), elementName + ".values");

        checkThatListDoesNotExistOrContainElements(observation.getRealmCodes(), elementName + ".realmCodes");
        checkThatDoesNotExist(observation.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(observation.getTemplateIds(), elementName + ".templateIds");
        checkThatListDoesNotExistOrContainElements(observation.getIds(), elementName + ".ids");
        checkThatDoesNotExist(observation.getDerivationExpr(), elementName + ".derivationExpr");
        checkThatDoesNotExist(observation.getText(), elementName + ".text");
        checkThatDoesNotExist(observation.getStatusCode(), elementName + ".statusCode");
        checkThatDoesNotExist(observation.getEffectiveTime(), elementName + ".effectiveTime");
        checkThatDoesNotExist(observation.getPriorityCode(), elementName + ".priorityCode");
        checkThatDoesNotExist(observation.getRepeatNumber(), elementName + ".repeatNumber");
        checkThatDoesNotExist(observation.getLanguageCode(), elementName + ".languageCode");
        checkThatListDoesNotExistOrContainElements(observation.getInterpretationCodes(), elementName + ".interpretationCodes");
        checkThatListDoesNotExistOrContainElements(observation.getMethodCodes(), elementName + ".methodCodes");
        checkThatListDoesNotExistOrContainElements(observation.getTargetSiteCodes(), elementName + ".targetSiteCodes");
        checkThatDoesNotExist(observation.getSubject(), elementName + ".subject");
        checkThatListDoesNotExistOrContainElements(observation.getSpecimens(), elementName + ".specimens");
        checkThatListDoesNotExistOrContainElements(observation.getPerformers(), elementName + ".performers");
        checkThatListDoesNotExistOrContainElements(observation.getAuthors(), elementName + ".authors");
        checkThatListDoesNotExistOrContainElements(observation.getInformants(), elementName + ".informants");
        checkThatListDoesNotExistOrContainElements(observation.getParticipants(), elementName + ".participants");
        checkThatListDoesNotExistOrContainElements(observation.getEntryRelationships(), elementName + ".entryRelationships");
        checkThatListDoesNotExistOrContainElements(observation.getReferences(), elementName + ".references");
        checkThatListDoesNotExistOrContainElements(observation.getPreconditions(), elementName + ".preconditions");
        checkThatListDoesNotExistOrContainElements(observation.getReferenceRanges(), elementName + ".referenceRanges");
        checkThatDoesNotExist(observation.isNegationInd(), elementName + ".negationInd");

    }

    private void validateMoodCode(XActMoodDocumentObservation moodCode, String elementName) {
        checkThatExistsAndContainsSupportedValue(moodCode.value(), XActMoodDocumentObservation.EVN.value(), elementName);
    }

    private void validateClassCodes(List<String> classCodes, String elementName) {
        if(!checkThatListContainsSingleElement(classCodes, elementName)) return;
        String classCode = classCodes.get(0);
        checkThatExistsAndContainsSupportedValue(classCode, AbstractHL7Mapper.OBSERVATION_CLASSCODE, elementName);
    }

    private void validateValues(List<ANY> values, String elementName) {
        if(!checkThatExists(values, elementName)) return;
        if (!checkThatListContainsSingleElement(values, elementName)) return;
        OrganDonorRegistration organDonorRegistration = (OrganDonorRegistration) values.get(0);

        validatePermissionType(organDonorRegistration.getPermissionType(), elementName + ".permissionType");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForHeart(), elementName + ".heart");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForLungs(), elementName + ".lungs");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForLiver(), elementName + ".liver");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForPancreas(), elementName + ".pancreas");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForKidneys(), elementName + ".kidneys");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForCornea(), elementName + ".cornea");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForSmallIntestine(), elementName + ".smallIntestine");
        validateOdrBooleanValues(organDonorRegistration.isPermissionForSkin(), elementName + ".skin");
        validateOdrBooleanValues(organDonorRegistration.isRequiresRelativeAcceptance(), elementName + ".relativeAcceptance");
    }

    private void validateOdrBooleanValues(Boolean value, String elementName) {
        if (checkThatExist(value)) {
            checkThatExistsAndContainsSupportedValue(value, Arrays.asList(true, false), elementName);
        }
    }

    private void validateRecordTargets(List<POCDMT000040RecordTarget> recordTargets, String elementName) {
        if (checkThatListContainsSingleElement(recordTargets, elementName)) {
            validateRecordTarget(recordTargets.get(0), elementName);
        }
    }

    private void validateRecordTarget(POCDMT000040RecordTarget recordTarget, String elementName) {
        checkThatExistsAndContainsSupportedValue(recordTarget.getContextControlCode(), AbstractHL7Mapper.RECORD_TARGET_CTX_CTRL_CODE,
                elementName + ".contextControlCode");
        checkThatListContainsSingleElementWithSpecificValue(recordTarget.getTypeCodes(), AbstractHL7Mapper.RECORD_TARGET_TYPE_CODE,
                elementName + ".typeCode");
        validatePatientRole(recordTarget.getPatientRole(), elementName + ".patientRole");

        checkThatListDoesNotExistOrContainElements(recordTarget.getNullFlavors(), elementName + ".nullFlavors");
        checkThatDoesNotExist(recordTarget.getTypeId(), elementName + ".typeId");
        checkThatListDoesNotExistOrContainElements(recordTarget.getRealmCodes(), elementName + ".realmCodes");
        checkThatListDoesNotExistOrContainElements(recordTarget.getTemplateIds(), elementName + ".templateIds");
    }

    private void validatePatientRole(POCDMT000040PatientRole patientRole, String elementName) {
        if (checkThatListContainsSingleElement(patientRole.getIds(), elementName + ".id")) {
            validateCPR(patientRole.getIds().get(0), elementName + ".id", true);
        }
    }

    private void validateCPR(II id, String elementName, boolean checkCprExists) {
        validateId(id, Arrays.asList(OID.CPR), Arrays.asList(CPR),elementName + ".id");
        if(checkCprExists) {
            validateCprExists(id.getExtension());
        }
    }

    private void validateId(II id, List<String> supportedRoots, List<String> supportedAssigningAuthorities, String elementName) {
        if (!checkThatExists(id, elementName)) return;
        checkThatExists(id.getExtension(), elementName + ".extension");
        checkThatExistsAndContainsSupportedValue(id.getAssigningAuthorityName(), supportedAssigningAuthorities,
                elementName + ".assigningAuthorityName");
        checkThatExistsAndContainsSupportedValue(id.getRoot(), supportedRoots, elementName + ".root");

        checkThatDoesNotExist(id.isDisplayable(), elementName + ".displayable");
    }

    private <T> boolean checkThatListContainsSingleElement(List<T> list, String elementName) {
        return checkThatListContainsRightNumberOfElements(list, 1, 1, elementName);
    }

    private <T> boolean checkThatListContainsRightNumberOfElements(List<T> list, int minNumberOfElements,
                                                                   int maxNumberOfElements, String elementName) {
        if (list.size() < minNumberOfElements) {
            validationErrors.add(
                    new RequestValidatorError(list.size() + " elements were found but at least " + minNumberOfElements +
                            " elements are required: " + elementName,
                            false));
            return false;
        }
        if (list.size() > maxNumberOfElements) {
            validationErrors.add(
                    new RequestValidatorError(list.size() + " elements were found but at most " + maxNumberOfElements +
                            " elements are supported: " + elementName,
                            false));
            return false;
        }
        return true;
    }

    private void checkThatListContainsSingleElementWithSpecificValue(List<String> list, String expectedValue, String elementName) {
        if (checkThatListContainsSingleElement(list, elementName)) {
            checkThatExistsAndContainsSupportedValue(list.get(0), expectedValue, elementName);
        }
    }

    private boolean checkThatExistsAndContainsSupportedValue(String s, String supportedValue, String elementName) {
        return checkThatExistsAndContainsSupportedValue(s, Arrays.asList(supportedValue), elementName);
    }


    private boolean checkThatExistsAndContainsSupportedValue(Boolean s, List<Boolean> supportedValues, String elementName) {
        if (!checkThatExists(s, elementName)) {
            return false;
        }
        if (!supportedValues.contains(s)) {
            validationErrors.add(new RequestValidatorError(
                    "The " + elementName + " value " + s + " is not supported. Supported values are: "
                            + String.join(",", supportedValues.toString()),
                    false));
            return false;
        }
        return true;

    }

    private boolean checkThatExistsAndContainsSupportedValue(String s, List<String> supportedValues, String elementName) {
        if (!checkThatExists(s, elementName)) {
            return false;
        }
        if (!supportedValues.contains(s)) {
            validationErrors.add(new RequestValidatorError(
                    "The " + elementName + " value " + s + " is not supported. Supported values are: "
                            + String.join(",", supportedValues.toString()),
                    false));
            return false;
        }
        return true;

    }

    private void validateId(II id, String elementName) {
        validateId(id, true, true, true, elementName);
    }

    private void validateId(II id, boolean extensionRequired, boolean assigningAuthorityRequired, boolean personIdentifier, String elementName) {
        if (!checkThatExists(id, elementName)) return;

        if (extensionRequired) {
            if (!checkThatExists(id.getExtension(), elementName + ".extension")) {
                return;
            }
            if(personIdentifier) {
                checkPersonIdLength(id.getExtension(), elementName + ".extension");
                validateCprExists(id.getExtension());
            }
        }
        if (assigningAuthorityRequired)
            if(checkThatExists(id.getAssigningAuthorityName(), elementName + ".assigningAuthorityName"))
                checkThatExistsAndContainsSupportedValue(id.getAssigningAuthorityName(), Arrays.asList(AbstractHL7Mapper.CPR), elementName + ".assigningAuthorityName");

        checkThatExists(id.getRoot(), elementName + ".root");
        checkThatDoesNotExist(id.isDisplayable(), elementName + ".displayable");
    }

    private void checkPersonIdLength(String extension, String elementName) {
        if (extension == null || extension.length() != 10) {
            validationErrors.add(new RequestValidatorError(
                    "Person identifier not Valid. Requires 10 digits " + elementName + " [" + extension + "]",
                    false));
        }
    }

    private void validateCprExists(String extension) {
        try {
            cprExists.validateCpr(extension);
            cprExists.validateAge(extension);
        } catch (ValidationException e) {
            validationErrors.add(new RequestValidatorError(e.getMessage(), true));
        }
    }

    private void checkThatDoesNotExist(Object o, String elementName) {
        if (o != null) {
            validationErrors.add(new RequestValidatorError("Unsupported element found: " + elementName, false));
        }
    }

    private boolean checkThatExists(Object o, String elementName) {
        if (o == null || (o instanceof String && ((String) o).isEmpty())) {
            validationErrors.add(new RequestValidatorError("Required value was missing: " + elementName, false));
            return false;
        }
        return true;
    }

    private boolean checkThatExist(Boolean value) {
        if (value == null) {
            return false;
        }
        return true;
    }

    private boolean checkThatExist(Boolean value, String elementName) {
        if (value == null ||
                !checkThatExistsAndContainsSupportedValue(value, Arrays.asList(true, false), elementName)) {
            validationErrors.add(new RequestValidatorError("Required value was missing: " + elementName, false));
            return false;
        }
        return true;
    }

    private boolean checkOdrExists(Object odr, String elementName) {
        if (odr == null) {
            validationErrors.add(new RequestValidatorError("Required value was missing: " + elementName, false));
            return false;
        }
        return true;
    }

    private void throwExceptionInCaseOfValidationErrors() {
        if (validationErrors.size() > 0) {
            throw new IllegalArgumentException(getCombinedErrorText());
        }
    }

    private RequestValidatorError buildUnspecificErrorMessage(String requestName, String errorMessage) {
        return new RequestValidatorError("Parse-fejl ifm. " + requestName + ": " + errorMessage, false);
    }

    private <T> boolean checkThatListDoesNotExistOrContainElements(List<T> list, String elementName) {
        if (list != null && list.size() > 0) {
            validationErrors.add(new RequestValidatorError("Unsupported value was found: " + elementName, false));
            return false;
        }
        return true;
    }

    private void checkDateTimeIsValid(String dateTimeString, String elementName) {
        try {

            try {
            	DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern(TS_FORMAT);
                LocalDateTime.parse(dateTimeString, dateTimeFormatter);
            } catch (DateTimeParseException e) {
                throw new ParseException("Date time string " + dateTimeString + " does not comply with the required format: " + TS_FORMAT, e);
            }
        } catch (ParseException e) {
            validationErrors.add(
                    new RequestValidatorError("The timestamp " + dateTimeString + " for the element " + elementName +
                            " does not comply with the required format. " + e.getMessage(),
                            false));
        }
    }
}
