/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.auditlog;

import dk.sds.nsp.audit.AuditBuilder;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Instant;

public class AuditLoggerTest {

    private static final String COMPONENT = "ODR";

    private AuditBuilder auditBuilder;
    private AuditLogger auditLogger;

    @Before
    public void setUp() throws Exception {
        auditBuilder = Mockito.mock(AuditBuilder.class);
        auditLogger = new AuditLogger(() -> auditBuilder);
    }

    @Test
    public void logCreateOrganDonorRegistration_LogsOrganDonorAndActor() {
        OrganDonorVO organDonor = buildTestOrganDonor();
        ServiceActorVO serviceActor = buildServiceActor();

        auditLogger.logCreateOrganDonorRegistration(organDonor, serviceActor);

        String context = "createOrganDonorRegistration";
        verifyOrganDonorAuditLogged(organDonor, context);
        verifyServiceActorAuditLogged(serviceActor, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logUpdateOrganDonorRegistration_LogsOrganDonorAndActor() {
        OrganDonorVO organDonor = buildTestOrganDonor();
        ServiceActorVO serviceActor = buildServiceActor();

        auditLogger.logUpdateOrganDonorRegistration(organDonor, serviceActor);

        String context = "updateOrganDonorRegistration";
        verifyOrganDonorAuditLogged(organDonor, context);
        verifyServiceActorAuditLogged(serviceActor, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logDeleteOrganDonorRegistration_LogsPersonIdAndActor() {
        String personId = "1111111111";
        ServiceActorVO serviceActor = buildServiceActor();

        auditLogger.logDeleteOrganDonorRegistration(personId, serviceActor);

        String context = "deleteOrganDonorRegistration";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr", personId);
        verifyServiceActorAuditLogged(serviceActor, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logGetOrganDonorRegistration_LogsOrganDonorAndActor() {
        OrganDonorVO organDonor = buildTestOrganDonor();
        ServiceActorVO serviceActor = buildServiceActor();

        auditLogger.logGetOrganDonorRegistration(organDonor, serviceActor);

        String context = "getOrganDonorRegistration";
        verifyOrganDonorAuditLogged(organDonor, context);
        verifyServiceActorAuditLogged(serviceActor, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logHasOrganDonorRegistration_LogsPersonIdAndActor() {
        String personId = "2222222222";
        ServiceActorVO serviceActor = buildServiceActor();

        auditLogger.logHasOrganDonorRegistration(personId, serviceActor);

        String context = "hasOrganDonorRegistration";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr", personId);
        verifyServiceActorAuditLogged(serviceActor, context);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logCprDoesNotExist_LogsCprDoesNotExist() {
        String cpr = "1111111111";

        auditLogger.logCprDoesNotExist(cpr);

        String context = "validateCpr";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr-does-not-exist", cpr);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    @Test
    public void logCprInactive_LogsCprInactive() {
        String cpr = "1111111111";

        auditLogger.logCprInactive(cpr);

        String context = "validateCpr";
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr-inactive", cpr);
        Mockito.verifyNoMoreInteractions(auditBuilder);
    }

    private void verifyOrganDonorAuditLogged(OrganDonorVO organDonor, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "pid", organDonor.getOrganDonorPID());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "cpr", organDonor.getPersonIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "version", organDonor.getVersion());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "modified", organDonor.getModifiedDateTime());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "validFrom", organDonor.getValidFrom());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.NonPersonalInformation, "validTo", organDonor.getValidTo());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "permissionType", organDonor.getPermissionType());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "heart", organDonor.getHeart());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "kidneys", organDonor.getKidneys());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "lungs", organDonor.getLungs());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "corneas", organDonor.getCorneas());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "liver", organDonor.getLiver());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "smallIntestine", organDonor.getSmallIntestine());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "pancreas", organDonor.getPancreas());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "skin", organDonor.getSkin());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "relativeAcceptanceRequired", organDonor.getRelativeAcceptanceRequired());

    }

    private void verifyServiceActorAuditLogged(ServiceActorVO serviceActor, String context) {
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.SensitivePersonalInformation, "actor-cpr", serviceActor.getPersonIdentifier());
        Mockito.verify(auditBuilder).addAuditInformation(
                COMPONENT, context, AuditBuilder.TypeOfInformation.RegularPersonalInformation, "actor-usertype", serviceActor.getUserType());
    }

    private static OrganDonorVO buildTestOrganDonor() {
        OrganDonorVO organDonorVO = new OrganDonorVO();
        organDonorVO.setOrganDonorPID(5);
        organDonorVO.setPersonIdentifier("1111111111");
        organDonorVO.setVersion(4);
        organDonorVO.setHeart(true);
        organDonorVO.setLungs(false);
        organDonorVO.setLiver(false);
        organDonorVO.setCorneas(false);
        organDonorVO.setKidneys(true);
        organDonorVO.setSmallIntestine(false);
        organDonorVO.setPancreas(true);
        organDonorVO.setSkin(false);
        organDonorVO.setRelativeAcceptanceRequired(false);
        organDonorVO.setPermissionType(PermissionType.FULL);
        organDonorVO.setModifiedDateTime(Instant.now());
        organDonorVO.setValidFrom(Instant.now());
        organDonorVO.setValidTo(Instant.now());
        return organDonorVO;
    }
    public ServiceActorVO buildServiceActor() {
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        serviceActor.setOrganizationId("999999");
        serviceActor.setOrganizationIdSource("CVR");
        serviceActor.setOrganizationName("Test org");
        serviceActor.setPersonIdentifier("5555555555");
        serviceActor.setFirstName("Jens");
        serviceActor.setLastName("Jensen");
        serviceActor.setAuthorisationNumber("1234");
        serviceActor.setEducationCode("000");
        return serviceActor;
    }

}