/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.business;

import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.util.ODRException;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;
import org.springframework.transaction.annotation.Transactional;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

@Transactional
public class AccessControlTests extends ServiceTestHelper {

    private static final String personIdentifier = "1234567890";
    private static final String messageId = "8572";

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    private AccessControlImpl classUnderTest;

    @Before
    public void setUp() {
        classUnderTest = new AccessControlImpl();
    }

    @Test
    public void checkGetAccess_success() {
        // Arrange
        RequestContextVO requestContext1 = getRequestContextExample(personIdentifier, messageId);
        RequestContextVO requestContext2 = getRequestContextExample(personIdentifier, UserType.HealthCareProfessional, messageId);

        // Act
        classUnderTest.checkGetAccess(requestContext1);
        classUnderTest.checkGetAccess(requestContext2);
    }

    @Test
    public void checkGetAccess_fail() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, UserType.SystemUser, messageId);

        // Act
        try {
            classUnderTest.checkGetAccess(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkGetAccess_citizenWithDifferentId_fail() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, UserType.Citizen, messageId);
        requestContext.setPersonIdentifier("2222222222");

        // Act
        try {
            classUnderTest.checkGetAccess(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }

    @Test
    public void checkHasAccess_success() {
        // Arrange
        RequestContextVO requestContext1 = getRequestContextExample(personIdentifier, messageId);
        RequestContextVO requestContext2 = getRequestContextExample(personIdentifier, UserType.SystemUser, messageId);

        // Act
        classUnderTest.checkHasAccess(requestContext1);
        classUnderTest.checkHasAccess(requestContext2);
    }

    @Test
    public void checkWriteAccess_success() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, messageId);

        // Act
        classUnderTest.checkWriteAccess(requestContext);
    }

    @Test
    public void checkWriteAccess_fail() {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, UserType.HealthCareProfessional, messageId);

        // Act
        try {
            classUnderTest.checkWriteAccess(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED, e.getFaultCode());
        }
    }
}
