/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.business;

import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog.MinLogException;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.nas.organdonor.NasException;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.nas.organdonor.NotificationSender;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.util.ODRException;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import java.time.Instant;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.nullable;
import static org.mockito.Mockito.*;

public class OrganDonorServiceImplTest extends ServiceTestHelper {
    private static final String personIdentifier = "1234567890";

    private static final String messageId = "8572";

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private Properties properties;

    @Mock
    private MinLogDataWrapper minLogClient;

    @Mock
    private AccessControl accessControl;

    @Mock
    private Validator validator;

    @Mock
    private OrganDonorDAO organDonorDAO;

    @Mock
    private NotificationSender notificationSender;

    private OrganDonorServiceImpl serviceUnderTest;

    @Before
    public void setUp() {
        serviceUnderTest = new OrganDonorServiceImpl(properties, minLogClient, accessControl, validator, organDonorDAO, notificationSender);
    }

    @Test
    public void organDonorService_citizenCreateOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_createOrganDonorRegistration_success(UserType.Citizen, 1);
    }
    @Test
    public void organDonorService_adminCreateOrganDonorRegistration_success() throws DAOException, MinLogException, NasException {
    	organDonorService_createOrganDonorRegistration_success(UserType.Admin, 1);
    }
    @Test
    public void organDonorService_systemCreateOrganDonorRegistration_success() throws DAOException, MinLogException, NasException {
    	organDonorService_createOrganDonorRegistration_success(UserType.SystemUser, 0);
    }
    @Test
    public void organDonorService_healthCareProfCreateOrganDonorRegistration_success() throws DAOException, MinLogException, NasException {
    	organDonorService_createOrganDonorRegistration_success(UserType.HealthCareProfessional, 1);
    }
    
    private void organDonorService_createOrganDonorRegistration_success(UserType usertype, int minlogExpectedCalls) throws DAOException, NasException, MinLogException {
        OrganDonorVO organDonorVO = getOrganDonorRegistrationExample(personIdentifier);
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, usertype, messageId);

        serviceUnderTest.createOrganDonorRegistration(organDonorVO, requestContext);
        verify(accessControl).checkWriteAccess(requestContext);

        ArgumentCaptor<OrganDonorVO> argument = ArgumentCaptor.forClass(OrganDonorVO.class);
        verify(organDonorDAO).createOrganDonorRegistration(argument.capture(), any(Instant.class));
        
        verify(minLogClient, times(minlogExpectedCalls)).logActivity(eq(requestContext), any(), eq(OrganDonorServiceImpl.ODR_SERVICE_IDENTIFIER));

        assertEquals(organDonorVO.getPersonIdentifier(), argument.getValue().getPersonIdentifier());
        assertEquals(HEART, argument.getValue().getHeart());
        assertEquals(LUNGS, argument.getValue().getLungs());
        assertEquals(LIVER, argument.getValue().getLiver());
        verify(notificationSender).sendOrganDonorNotification(personIdentifier, requestContext.getTime());
    }
    
    @Test
    public void organDonorService_citizenUpdateOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_UpdateOrganDonorRegistration_success(UserType.Citizen, 1);
    }

    @Test
    public void organDonorService_adminUpdateOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_UpdateOrganDonorRegistration_success(UserType.Admin, 1);
    }
    
    @Test
    public void organDonorService_systemUpdateOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_UpdateOrganDonorRegistration_success(UserType.SystemUser, 0);
    }
    
    @Test
    public void organDonorService_healthCareProfUpdateOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_UpdateOrganDonorRegistration_success(UserType.HealthCareProfessional, 1);
    }

    
    private void organDonorService_UpdateOrganDonorRegistration_success(UserType usertype, int minlogExpectedCalls) throws DAOException, NasException, MinLogException {
        // Arrange
        int pid = 10;
        int version = 1;
        
        OrganDonorVO organDonorVO = getOrganDonorRegistrationExample(personIdentifier);
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, usertype, messageId);
        
        // Stub DAO method to return an object with a PID and Version
        OrganDonorVO organDonorVOFake = new OrganDonorVO();
        organDonorVOFake.setOrganDonorPID(pid);
        organDonorVOFake.setVersion(version);
        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenReturn(organDonorVOFake);

        // Act
        serviceUnderTest.updateOrganDonorRegistration(organDonorVO, requestContext);

        // Assert
        verify(accessControl).checkWriteAccess(requestContext);

        ArgumentCaptor<Integer> updatePid = ArgumentCaptor.forClass(Integer.class);
        ArgumentCaptor<OrganDonorVO> updateOdrArgument = ArgumentCaptor.forClass(OrganDonorVO.class);
        ArgumentCaptor<Instant> updateDateArgument = ArgumentCaptor.forClass(Instant.class);

        verify(organDonorDAO).updateOrganDonorRegistration(updatePid.capture(), updateOdrArgument.capture(), updateDateArgument.capture());
        
        verify(minLogClient, times(minlogExpectedCalls)).logActivity(eq(requestContext), any(), eq(OrganDonorServiceImpl.ODR_SERVICE_IDENTIFIER));

        assertEquals(pid, updatePid.getValue().intValue());
        assertNotNull(updateDateArgument.getValue());
        assertEquals(organDonorVO.getPersonIdentifier(), updateOdrArgument.getValue().getPersonIdentifier());
        assertEquals(OrganDonorServiceImpl.INITIAL_VERSION + 1, updateOdrArgument.getValue().getVersion().intValue());
        verify(notificationSender).sendOrganDonorNotification(personIdentifier, requestContext.getTime());
    }
    
    @Test
    public void organDonorService_citizenDeleteOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_deleteOrganDonorRegistration_success(UserType.Citizen, 1);
    }

    @Test
    public void organDonorService_adminDeleteOrganDonorRegistration_success() throws DAOException, NasException, MinLogException {
    	organDonorService_deleteOrganDonorRegistration_success(UserType.Admin, 1);
    }
    
    @Test
    public void organDonorService_systemDeleteOrganDonorRegistration_success() throws DAOException, NasException,  MinLogException {
    	organDonorService_deleteOrganDonorRegistration_success(UserType.SystemUser, 0);
    }
    
    @Test
    public void organDonorService_healthCareProfDeleteOrganDonorRegistration_success() throws DAOException, NasException,  MinLogException {
    	organDonorService_deleteOrganDonorRegistration_success(UserType.HealthCareProfessional, 1);
    }
    
    private void organDonorService_deleteOrganDonorRegistration_success(UserType usertype, int minlogExpectedCalls) throws DAOException, NasException, MinLogException {
        int pid = 10;
        int version = 1;

        RequestContextVO requestContext = getRequestContextExample(personIdentifier, usertype,  messageId);

        // Stub DAO method to return an object with a PID and Version
        OrganDonorVO organDonorVOFake = new OrganDonorVO();
        organDonorVOFake.setOrganDonorPID(pid);
        organDonorVOFake.setVersion(version);
        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenReturn(organDonorVOFake);

        // Act
        serviceUnderTest.deleteOrganDonorRegistration(requestContext);

        // Assert
        verify(accessControl).checkWriteAccess(requestContext);

        ArgumentCaptor<Integer> updateArgument = ArgumentCaptor.forClass(Integer.class);
        ArgumentCaptor<Instant> updateDateArgument = ArgumentCaptor.forClass(Instant.class);
        
        verify(organDonorDAO).deleteOrganDonorRegistration(updateArgument.capture(), updateDateArgument.capture());
        
        verify(minLogClient, times(minlogExpectedCalls)).logActivity(eq(requestContext), any(), eq(OrganDonorServiceImpl.ODR_SERVICE_IDENTIFIER));

        assertEquals(pid, updateArgument.getValue().intValue());
        assertNotNull(updateDateArgument);

        verify(notificationSender).sendOrganDonorNotification(personIdentifier, requestContext.getTime());
    }
    
    @Test
    public void organDonorService_citizenGetOrganDonorRegistration_success() throws DAOException, MinLogException {
    	organDonorService_getOrganDonorRegistration_success(UserType.Citizen, 0);
    }
    @Test
    public void organDonorService_adminGetOrganDonorRegistration_success() throws DAOException, MinLogException {
    	organDonorService_getOrganDonorRegistration_success(UserType.Admin, 1);
    }
    @Test
    public void organDonorService_systemGetOrganDonorRegistration_success() throws DAOException, MinLogException {
    	organDonorService_getOrganDonorRegistration_success(UserType.SystemUser, 0);
    }
    @Test
    public void organDonorService_healthCareProfGetOrganDonorRegistration_success() throws DAOException, MinLogException {
    	organDonorService_getOrganDonorRegistration_success(UserType.HealthCareProfessional, 1);
    }
    
    private void organDonorService_getOrganDonorRegistration_success(UserType usertype, int minlogExpectedCalls) throws DAOException, MinLogException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, usertype, messageId);

        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenReturn(new OrganDonorVO());
        // Act
        serviceUnderTest.getOrganDonorRegistration(requestContext);

        // Assert
        verify(organDonorDAO).getOrganDonorRegistration(eq(personIdentifier), any(Instant.class));
        verify(accessControl).checkGetAccess(requestContext);
        verify(minLogClient, times(minlogExpectedCalls)).logActivity(eq(requestContext), any(), eq(OrganDonorServiceImpl.ODR_SERVICE_IDENTIFIER));
        verifyNoMoreInteractions(minLogClient);
    }

    @Test
    public void organDonorService_getOrganDonorRegistration_failOnMinLogException() throws DAOException, MinLogException {
        // Arrange
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, UserType.HealthCareProfessional, messageId);

        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenReturn(new OrganDonorVO());

        // Stub MinLog to throw an Exception
        doThrow(MinLogException.class).when(minLogClient).logActivity(any(RequestContextVO.class), nullable(String.class), nullable(String.class));

        // Act
        try {
            serviceUnderTest.getOrganDonorRegistration(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            verify(organDonorDAO).getOrganDonorRegistration(eq(personIdentifier), any(Instant.class));
            verify(accessControl).checkGetAccess(requestContext);
            verify(minLogClient).logActivity(eq(requestContext), any(), eq(OrganDonorServiceImpl.ODR_SERVICE_IDENTIFIER));

            assertEquals(FaultCode.GET_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

    @Test
    public void organDonorService_hasOrganDonorRegistration_success() throws DAOException {
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, messageId);

        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenReturn(new OrganDonorVO());

        Boolean returnedBoolean = serviceUnderTest.hasOrganDonorRegistration(requestContext);

        assertEquals(true, returnedBoolean);
        verify(organDonorDAO).getOrganDonorRegistration(eq(personIdentifier), any(Instant.class));
        verify(accessControl).checkHasAccess(requestContext);
    }

    @Test
    public void organDonorService_DAOExceptionThrown_expectedODRExceptionThrown() throws DAOException {
        OrganDonorVO organDonorVO = new OrganDonorVO();
        RequestContextVO requestContext = getRequestContextExample(personIdentifier, messageId);

        when(organDonorDAO.getOrganDonorRegistration(anyString(), any(Instant.class)))
                .thenThrow(new DAOException("An error occurred reading Organdonorregistration"));

        // Act
        try {
            serviceUnderTest.createOrganDonorRegistration(organDonorVO, requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.CREATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            serviceUnderTest.updateOrganDonorRegistration(organDonorVO, requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.UPDATE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        // Act
        try {
            serviceUnderTest.deleteOrganDonorRegistration(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.DELETE_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        try {
            serviceUnderTest.getOrganDonorRegistration(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.GET_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }

        try {
            serviceUnderTest.hasOrganDonorRegistration(requestContext);
            fail();
        } catch (ODRException e) {
            // Assert
            assertEquals(FaultCode.HAS_ORGANDONORREGISTRATION_INTERNAL_SERVER_ERROR, e.getFaultCode());
        }
    }

}
