/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.dao.impl;

import dk.sundhedsdatastyrelsen.organdonor.service.AbstractSpringTest;
import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.time.Year;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;

import static org.junit.Assert.*;

@Transactional("organDonorDataSourceTxManager")
public class OrganDonorDAOImplTest extends AbstractSpringTest {
    public static final String PERSON_ID = "1254125412";

    @Autowired
    private OrganDonorDAO dao;


    @Ignore // disabled until a fix is implemented
    @Test
    public void createAndGetOrganDonorRegistration_winterTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to winter time (in Denmark)
        final Instant afterWinterTimeChange = Instant.parse("2018-10-28T00:10:00.000Z");
        final OrganDonorVO organDonorVO = buildTestOrganDonor(10);

        // Act
        dao.createOrganDonorRegistration(organDonorVO, afterWinterTimeChange);

        // Assert
        // We should be able to find that record 55 minutes later
        final OrganDonorVO actualOrganDonorVO = dao.getOrganDonorRegistration(
                organDonorVO.getPersonIdentifier(),
                afterWinterTimeChange.plus(55, ChronoUnit.MINUTES));
        assertNotNull(actualOrganDonorVO);
        assertEquals(afterWinterTimeChange, actualOrganDonorVO.getValidFrom());
    }

    @Test
    public void createAndGetOrganDonorRegistration_summerTime() throws Exception {
        // Arrange
        // Create a record a record with validFrom 10 minutes after the change to summer time (in Denmark)
        final Instant afterSummerTimeChange = Instant.parse("2018-03-25T02:10:00.000Z");
        final OrganDonorVO organDonorVO = buildTestOrganDonor(10);

        // Act
        dao.createOrganDonorRegistration(organDonorVO, afterSummerTimeChange);

        // Assert
        // We should be able to find that record 55 minutes later
        final OrganDonorVO actualOrganDonorVO = dao.getOrganDonorRegistration(
                organDonorVO.getPersonIdentifier(),
                afterSummerTimeChange.plus(55, ChronoUnit.MINUTES));
        assertNotNull(actualOrganDonorVO);
        assertEquals(afterSummerTimeChange, actualOrganDonorVO.getValidFrom());
    }

    @Test
    public void createAndGetOrganDonorRegistration_CreatesCorrectData() throws DAOException {
        final Instant date = Year.of(2018).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
        OrganDonorVO beforeTest = buildTestOrganDonor(10);
        final Integer pid = dao.createOrganDonorRegistration(beforeTest, date);

        OrganDonorVO afterTest = dao.getOrganDonorRegistration(beforeTest.getPersonIdentifier(), date);
        assertNotNull(afterTest);
        assertNotNull(afterTest.getModifiedDateTime());
        assertNotNull(afterTest.getValidFrom());
        assertNotNull(afterTest.getValidTo());

        assertEquals(pid, afterTest.getOrganDonorPID());
        assertEquals(beforeTest.getOrganDonorPID(), afterTest.getOrganDonorPID());
        assertEquals(beforeTest.getPersonIdentifier(), afterTest.getPersonIdentifier());
        assertEquals(beforeTest.getHeart(), afterTest.getHeart());
    }

    @Test
    public void createAndUpdateOrganDonorRegistration_updatesPersonCorrect() throws DAOException {
        final Instant date = Year.of(2018).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
        OrganDonorVO beforeTest = buildTestOrganDonor(10);
        int createdPid = dao.createOrganDonorRegistration(beforeTest, date);

        int newPid = beforeTest.getOrganDonorPID() + 1;
        int newVersion = beforeTest.getVersion() + 1;
        beforeTest.setHeart(false);
        beforeTest.setOrganDonorPID(newPid);
        beforeTest.setVersion(newVersion);
        final Instant newDate = Year.of(2019).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();

        dao.updateOrganDonorRegistration(createdPid, beforeTest, newDate);
        OrganDonorVO afterTest = dao.getOrganDonorRegistration(beforeTest.getPersonIdentifier(), newDate);
        assertNotNull(afterTest);
        assertEquals(newPid, afterTest.getOrganDonorPID().intValue());
        assertEquals(beforeTest.getHeart(), afterTest.getHeart());
    }

    @Test
    public void createAndDeleteOrganDonorRegistration_deletedSucces() throws DAOException {
        final Instant createdDate = Year.of(2018).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();
        OrganDonorVO beforeTest = buildTestOrganDonor(10);
        int personPid = dao.createOrganDonorRegistration(beforeTest, createdDate);
        final Instant deletedDate = Year.of(2019).atDay(1).atStartOfDay(ZoneOffset.UTC).toInstant();

        dao.deleteOrganDonorRegistration(beforeTest.getOrganDonorPID(), deletedDate);
        OrganDonorVO afterTest = dao.getOrganDonorRegistration(beforeTest.getPersonIdentifier(), deletedDate);
        OrganDonorVO deletedOrganDonorRegistration = dao.getOrganDonorRegistration(beforeTest.getPersonIdentifier(), createdDate);
        assertNull(afterTest);

        final DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.s").withZone(ZoneId.systemDefault());
        String newDate = formatter.format(deletedDate);
        assertEquals(deletedDate.atOffset(ZoneOffset.UTC).toLocalDate(), deletedOrganDonorRegistration.getValidTo().atOffset(ZoneOffset.UTC).toLocalDate());
    }

    @Test(expected = DAOException.class)
    public void createOrganDonorRegistration_WithNullTime_ThrowsException() throws Exception {
        OrganDonorVO beforeTest = buildTestOrganDonor(10);
        dao.createOrganDonorRegistration(beforeTest, null);
    }

    private static OrganDonorVO buildTestOrganDonor(int pid) {
        OrganDonorVO organDonorVO = new OrganDonorVO();
        organDonorVO.setOrganDonorPID(pid);
        organDonorVO.setPersonIdentifier(PERSON_ID);
        organDonorVO.setVersion(4);
        organDonorVO.setHeart(true);
        organDonorVO.setLungs(false);
        organDonorVO.setLiver(false);
        organDonorVO.setCorneas(false);
        organDonorVO.setKidneys(true);
        organDonorVO.setRelativeAcceptanceRequired(false);
        organDonorVO.setPermissionType(PermissionType.FULL);
        return organDonorVO;
    }
}
