/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog;

import dk.sundhedsdatastyrelsen.organdonor.domain.OrganizationTypeVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnit;
import org.mockito.junit.MockitoRule;
import org.mockito.quality.Strictness;

import java.time.Instant;
import java.time.temporal.ChronoUnit;

import static org.junit.Assert.*;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.verify;

public class MinLogDataWrapperTests {

    @Rule
    public MockitoRule rule = MockitoJUnit.rule().strictness(Strictness.STRICT_STUBS);

    @Mock
    private MinLogClient minLogClient;

    private MinLogDataWrapperImpl classUnderTest;

    @Mock
    private MinLogData minLogData;

    @Before
    public void setUp() {
        classUnderTest = new MinLogDataWrapperImpl(minLogClient);
    }

    @Test
    public void testMinLog_WithValidData_PassesDataCorrectlyToMinLogDataWrapper() throws MinLogException {
        classUnderTest.logActivity(getRequestContextExample(), "test", "system");

        ArgumentCaptor<MinLogData> argument = ArgumentCaptor.forClass(MinLogData.class);
        verify(minLogClient).logDataAdd(any(), argument.capture());
        final MinLogData minLogData = argument.getValue();
        assertFalse(minLogData.getActivity().isEmpty());
        assertEquals("test", minLogData.getActivity());
        assertEquals("cpr1", minLogData.getCitizenCpr());
        assertEquals("orgName", minLogData.getOrganizationName());
        assertEquals("orgId", minLogData.getOrganizationId());
        assertEquals(OrganizationTypeVO.SOR, minLogData.getOrganizationType());
        assertNull(minLogData.getOnBehalfOfCpr());
        assertNull(minLogData.getSessionId());
        assertEquals("system", minLogData.getSourceSystemIdentifier());
        Instant tenMinutesAgo = Instant.now().minusSeconds(60*10).truncatedTo(ChronoUnit.SECONDS);
        assertTrue(minLogData.getEventDateTime().isAfter(tenMinutesAgo));
        assertEquals("cpr2", minLogData.getUserCPR());
    }

    @Test(expected = MinLogException.class)
    public void testMinlogClient_WithInvalidCitizenCPR() throws MinLogException{
        Mockito.when(minLogData.getCitizenCpr()).thenReturn("");
        minLogClient = new MinLogClient("/minlogclient.properties");
        minLogClient.logDataAdd(getRequestContextExample().getSlaLogItem(), minLogData);
    }

    @Test(expected = MinLogException.class)
    public void testMinlogClient_WithInvalidData() throws MinLogException{
        Mockito.doThrow(MinLogException.class)
                .when(minLogClient)
                .logDataAdd(any(),any());
        minLogClient.logDataAdd(getRequestContextExample().getSlaLogItem(), null);
    }

    @Test(expected = NullPointerException.class)
    public void createMinlogNonExistingProperties() throws MinLogException{
        minLogClient = new MinLogClient("/minlog.properties");
    }
    private RequestContextVO getRequestContextExample() {
        ServiceActorVO actor = new ServiceActorVO(UserType.Citizen);
        actor.setOrganizationId("orgId");
        actor.setOrganizationName("orgName");
        actor.setOrganizationIdSource("SOR");
        actor.setPersonIdentifier("cpr2");
        RequestContextVO requestContext = new RequestContextVO(actor, null);
        requestContext.setPersonIdentifier("cpr1");
        return requestContext;
    }
}
