/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice;

import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.*;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

public class CprExistsImplTest {

	private CprExistsServiceClientFactory cprExistsServiceClientFactory;
	private AuditLogger auditLogger;
	private CprExistsServiceClient cprExistsServiceClient;
	private ServiceActorVO serviceActorVO;

	private static final String VALID_ACTIVE_CPR = "valid";
	private static final String VALID_INACTIVE_CPR = "inactive";
	private static final String INVALID_CPR = "invalid";

	private static final int CPR_EXISTS_MIN_AGE = 15;


	@Before
	public void setup() throws Exception {

		cprExistsServiceClient = Mockito.mock(CprExistsServiceClient.class);
		Mockito.when(cprExistsServiceClient.validateCpr(Mockito.eq(VALID_ACTIVE_CPR))).thenReturn(new CprStatus(false));
		Mockito.when(cprExistsServiceClient.validateCpr(Mockito.eq(VALID_INACTIVE_CPR))).thenReturn(new CprStatus(true));
		Mockito.when(cprExistsServiceClient.validateCpr(Mockito.eq(INVALID_CPR))).thenReturn(null);

		cprExistsServiceClientFactory = Mockito.mock(CprExistsServiceClientFactory.class);
		Mockito.when(cprExistsServiceClientFactory.createClient()).thenReturn(cprExistsServiceClient);

		auditLogger = Mockito.mock(AuditLogger.class);

		serviceActorVO = Mockito.mock(ServiceActorVO.class);
	}

	@Test
	public void testValidationOffAndInvalidCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.OFF, cprExistsServiceClientFactory, auditLogger,
				CPR_EXISTS_MIN_AGE);

		// When
		subject.validateCpr(INVALID_CPR);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(0)).validateCpr(Mockito.anyString());
	}

	@Test
	public void testValidationOffAndInactiveCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.OFF, cprExistsServiceClientFactory, auditLogger,
				CPR_EXISTS_MIN_AGE);

		// When
		subject.validateCpr(VALID_INACTIVE_CPR);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(0)).validateCpr(Mockito.anyString());
	}

	@Test
	public void testValidationWarningGeneratesAuditEntryOnInvalidCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.WARNING, cprExistsServiceClientFactory, auditLogger,
				CPR_EXISTS_MIN_AGE);
		String invalidPatientCpr = INVALID_CPR;

		// When
		subject.validateCpr(invalidPatientCpr);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
		Mockito.verify(auditLogger, Mockito.times(1)).logCprDoesNotExist(invalidPatientCpr);
	}

	@Test
	public void testValidationWarningGeneratesAuditEntryOnInactiveCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.WARNING, cprExistsServiceClientFactory, auditLogger,
				CPR_EXISTS_MIN_AGE);
		String inactivePatientCpr = VALID_INACTIVE_CPR;

		// When
		subject.validateCpr(inactivePatientCpr);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
		Mockito.verify(auditLogger, Mockito.times(1)).logCprInactive(inactivePatientCpr);
	}

	@Test
	public void testValidationWarningGeneratesNoAuditEntryOnValidActiveCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.WARNING, cprExistsServiceClientFactory,
				auditLogger,
				CPR_EXISTS_MIN_AGE);

		// When
		subject.validateCpr(VALID_ACTIVE_CPR);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
		Mockito.verifyZeroInteractions(auditLogger);
	}


	@Test
	public void testValidationRejectOnValidActiveCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.REJECT, cprExistsServiceClientFactory,
				auditLogger,
				CPR_EXISTS_MIN_AGE);

		// When
		subject.validateCpr(VALID_ACTIVE_CPR);

		// Then
		Mockito.verify(cprExistsServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
		Mockito.verifyZeroInteractions(auditLogger);
	}

	@Test(expected = ValidationException.class)
	public void testValidationRejectOnInValidCpr() throws ValidationException {

		// Given
		CprExistsImpl subject = new CprExistsImpl(CprExists.ValidationLevel.REJECT, cprExistsServiceClientFactory, auditLogger,
				CPR_EXISTS_MIN_AGE);
		String invalidPatientCpr = INVALID_CPR;

		// When
		try {
			subject.validateCpr(invalidPatientCpr);
		} catch (ValidationException e) {
			// Then
			Mockito.verify(cprExistsServiceClient, Mockito.times(1)).validateCpr(Mockito.anyString());
			Mockito.verifyZeroInteractions(auditLogger);
			throw e;
		}
	}

	@Test
	public void validationAgeYoungerThanMinAge() {
		int patientAge = CPR_EXISTS_MIN_AGE - 1;
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, CprExists.ValidationLevel.OFF));
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, CprExists.ValidationLevel.WARNING));
		Assert.assertThrows(ValidationException.class, () -> validateAge(patientAge, CprExists.ValidationLevel.REJECT));
	}

	@Test
	public void validationAgeAtMinAge() throws ValidationException {
		int patientAge = CPR_EXISTS_MIN_AGE;
		validateAge(patientAge, CprExists.ValidationLevel.OFF);
		validateAge(patientAge, CprExists.ValidationLevel.WARNING);
		validateAge(patientAge, CprExists.ValidationLevel.REJECT);
	}

	private void validateAge(int validationAge, CprExists.ValidationLevel vl) throws ValidationException {
		Mockito.when(cprExistsServiceClient.getAge(Mockito.eq(VALID_ACTIVE_CPR))).thenReturn(validationAge);
		new CprExistsImpl(vl, cprExistsServiceClientFactory, auditLogger, CPR_EXISTS_MIN_AGE)
				.validateAge(VALID_ACTIVE_CPR);
	}

}
