/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice;

import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.*;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.auditlog.AuditLogger;
import dk.sundhedsdatastyrelsen.organdonor.service.business.OrganDonorService;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.context.RequestContextProvider;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.FromHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.ToHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.CprExists;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.odr.PermissionType;
import org.hl7.v3.II;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnitRunner;

import javax.servlet.ServletRequest;
import javax.xml.bind.ValidationException;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import static org.junit.Assert.assertFalse;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class OrganDonorWebServiceTest {
    private static final String personIdentifier = "1234567890";

    @InjectMocks
    private OrganDonorWebService serviceUnderTest;

    @Mock
    private FromHL7CDAMapper fromHL7CDAMapper = null;

    @Mock
    private OrganDonorService businessService = null;

    @Mock
    private ToHL7CDAMapper toHL7CDAMapper = null;

    @Mock
    private CprExists cprExists;

    @Mock
    private Properties properties = null;

    @Mock
    private AuditLogger auditLogger;

    @Mock
    private RequestContextProvider requestContextProvider;

    @Mock
    WebServiceContext context;

    @Mock
    RequestContextVO requestContext;

    @Before
    public void setup() {
        MessageContext mc = mock(MessageContext.class);
        when(mc.get(MessageContext.SERVLET_REQUEST)).thenReturn(mock(ServletRequest.class));
        when(context.getMessageContext()).thenReturn(mc);
        when(requestContextProvider.getRequestContext(Mockito.anyString(), Mockito.anyBoolean(), any())).thenReturn(requestContext);
        when(requestContext.getPersonIdentifier()).thenReturn(personIdentifier);
    }

    @Test
    public void createOrganDonorRegistration_validRequest_auditLogs() throws ValidationException {
        CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest();
        OrganDonorVO organDonor = new OrganDonorVO();
        when(fromHL7CDAMapper.mapOrganDonorFromRequest(request)).thenReturn(organDonor);
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        when(requestContext.getServiceActor()).thenReturn(serviceActor);

        serviceUnderTest.createOrganDonorRegistration20180501(request);

        Mockito.verify(auditLogger).logCreateOrganDonorRegistration(organDonor, serviceActor);
    }

    @Test
    public void deleteOrganDonorRegistration_validRequest_auditLogs() {
        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest();
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        when(requestContext.getServiceActor()).thenReturn(serviceActor);

        serviceUnderTest.deleteOrganDonorRegistration20180501(request);

        Mockito.verify(auditLogger).logDeleteOrganDonorRegistration(personIdentifier, serviceActor);
    }

    @Test
    public void getOrganDonorRegistration_validRequest_auditLogs() {
        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest();
        OrganDonorVO organDonor = new OrganDonorVO();
        when(businessService.getOrganDonorRegistration(requestContext)).thenReturn(organDonor);
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        when(requestContext.getServiceActor()).thenReturn(serviceActor);

        serviceUnderTest.getOrganDonorRegistration20180501(request);

        Mockito.verify(auditLogger).logGetOrganDonorRegistration(organDonor, serviceActor);
    }

    @Test
    public void hasOrganDonorRegistration_validRequest_auditLogs() {
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest();
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        when(requestContext.getServiceActor()).thenReturn(serviceActor);

        serviceUnderTest.hasOrganDonorRegistration20180501(request);

        Mockito.verify(auditLogger).logHasOrganDonorRegistration(personIdentifier, serviceActor);
    }

    @Test
    public void updateOrganDonorRegistration_validRequest_auditLogs() throws ValidationException {
        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest();
        OrganDonorVO organDonor = new OrganDonorVO();
        when(fromHL7CDAMapper.mapOrganDonorFromRequest(request)).thenReturn(organDonor);
        ServiceActorVO serviceActor = new ServiceActorVO(UserType.Citizen);
        when(requestContext.getServiceActor()).thenReturn(serviceActor);

        serviceUnderTest.updateOrganDonorRegistration20180501(request);

        Mockito.verify(auditLogger).logUpdateOrganDonorRegistration(organDonor, serviceActor);
    }

    @Test
    public void hasOrganDonorRegistration_cprDoesNotExist_falseIsReturned() throws dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException {
        doThrow(new dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException("Failing validateCpr")).when(cprExists).validateCpr(anyString());
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest();

        HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertFalse(response.isRegistrationExists());
    }

    @Test
    public void hasOrganDonorRegistration_ageTooLow_falseIsReturned() throws dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException {
        doThrow(new dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException("Failing validateAge")).when(cprExists).validateAge(anyString());
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest();

        HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertFalse(response.isRegistrationExists());
    }


    private UpdateOrganDonorRegistrationRequest createUpdateOrganDonorRegistrationRequest() {
        UpdateOrganDonorRegistrationRequest request = new UpdateOrganDonorRegistrationRequest();
        request.setId(createDefaultII(OrganDonorWebServiceTest.personIdentifier));
        request.setOrganDonorRegistration(createDefaultOrganDonorRegistration());
        return request;
    }

    private HasOrganDonorRegistrationRequest createHasOrganDonorRegistrationRequest() {
        HasOrganDonorRegistrationRequest hasRequest = new HasOrganDonorRegistrationRequest();
        hasRequest.setId(createDefaultII(OrganDonorWebServiceTest.personIdentifier));
        return hasRequest;
    }

    private GetOrganDonorRegistrationRequest createGetOrganDonorRegistrationRequest() {
        GetOrganDonorRegistrationRequest request = new GetOrganDonorRegistrationRequest();
        request.setId(createDefaultII(OrganDonorWebServiceTest.personIdentifier));
        return request;
    }

    private DeleteOrganDonorRegistrationRequest createDeleteOrganDonorRequest() {
        DeleteOrganDonorRegistrationRequest deleteRequest = new DeleteOrganDonorRegistrationRequest();
        deleteRequest.setId(createDefaultII(OrganDonorWebServiceTest.personIdentifier));
        return deleteRequest;
    }

    private CreateOrganDonorRegistrationRequest createCreateOrganDonorRegistrationRequest() {
        CreateOrganDonorRegistrationRequest createRequest = new CreateOrganDonorRegistrationRequest();
        createRequest.setId(createDefaultII(OrganDonorWebServiceTest.personIdentifier));
        createRequest.setOrganDonorRegistration(createDefaultOrganDonorRegistration());
        return createRequest;
    }

    private static II createDefaultII(String personIdentifier) {
        return new ToHL7CDAMapper().createId("1.2.208.176.1.2", personIdentifier, "CPR");
    }

    private OrganDonorRegistration createDefaultOrganDonorRegistration() {
        OrganDonorRegistration odr = new OrganDonorRegistration();
        odr.setPermissionType(PermissionType.FULL);
        odr.setPermissionForHeart(true);
        odr.setPermissionForLungs(true);
        odr.setPermissionForLiver(true);
        odr.setPermissionForPancreas(true);
        odr.setPermissionForKidneys(true);
        odr.setPermissionForCornea(true);
        odr.setPermissionForSmallIntestine(false);
        odr.setPermissionForSkin(false);
        odr.setRequiresRelativeAcceptance(false);
        return odr;
    }

}