/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.context;

import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext;
import dk.sdsd.nsp.slalog.impl.SLALogItemImpl;
import dk.sundhedsdatastyrelsen.organdonor.domain.RequestContextVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnitRunner;

import java.util.Optional;
import java.util.UUID;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

@RunWith(MockitoJUnitRunner.class)
public class RequestContextProviderTest {

    @Mock
    private ServiceActorProvider serviceActorProvider;
    @Mock
    private SecurityContextProvider securityContextProvider;

    private final SecurityContext securityContext = Mockito.spy(new DevelopmentSecurityProvider());
    private RequestContextProvider requestContextProvider;
    private SecurityContext.User actingUser;
    private SecurityContext.Message message;

    @Before
    public void setUp() throws Exception {
        setUpSecurityContext();
        requestContextProvider = new RequestContextProvider(securityContextProvider, serviceActorProvider);
    }

    private void setUpSecurityContext() {
        message = createMessage();
        Mockito.when(securityContext.getMessage()).thenReturn(Optional.of(message));
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void getRequestContext() {
        // Arrange
        ServiceActorVO actor = new ServiceActorVO(UserType.Citizen);
        Mockito.when(serviceActorProvider.getServiceActor(Mockito.any(), Mockito.anyBoolean())).thenReturn(actor);

        // Act
        String cpr = "1234567890";
        RequestContextVO requestContext = requestContextProvider.getRequestContext(cpr, false, new SLALogItemImpl(null, null, null));

        // Assert
        assertEquals(actor, requestContext.getServiceActor());
        assertEquals(cpr, requestContext.getPersonIdentifier());
        assertEquals(message.getConversationIdentifier().get(), requestContext.getFlowId());
        assertEquals(message.getIdentifier().get(), requestContext.getMessageId());
        assertEquals(requestContext.getMessageId(), requestContext.getSlaLogItem().getMessageId());
        Mockito.verify(serviceActorProvider).getServiceActor(securityContext, false);
    }

    @Test
    public void getRequestContext_withInvalidTicket_fails() {
        // Arrange
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(createInvalidTicket()));

        try {
            // Act
            requestContextProvider.getRequestContext("1234567890", false, null);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INVALID_SECURITY_TICKET.toString(), e.getMessage());
        }
    }

    private SecurityContext.Message createMessage() {
        SecurityContext.Message message = Mockito.mock(SecurityContext.Message.class);
        String messageId = UUID.randomUUID().toString();
        Mockito.when(message.getIdentifier()).thenReturn(Optional.of(messageId));
        String conversationId = UUID.randomUUID().toString();
        Mockito.when(message.getConversationIdentifier()).thenReturn(Optional.of(conversationId));
        return message;
    }

    private SecurityContext.Ticket createInvalidTicket() {
        return Mockito.mock(SecurityContext.Ticket.class);
    }
}