/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.context;

import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor.domain.ServiceActorVO;
import dk.sundhedsdatastyrelsen.organdonor.domain.UserType;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

public class ServiceActorProviderTest {
    private final SecurityContext securityContext = Mockito.spy(new DevelopmentSecurityProvider());
    private ServiceActorProvider serviceActorProvider;
    private List<String> whitelistedCvrs;
    private String audience;

    @Before
    public void setUp() throws Exception {
        setUpSecurityContext();

        audience = securityContext.getTicket().get().getAudience().get();
        whitelistedCvrs = Arrays.asList(securityContext.getOrganisation().get().getIdentifier());
        serviceActorProvider = new ServiceActorProvider(audience, whitelistedCvrs);
    }

    private void setUpSecurityContext() {
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.empty());
    }

    @Test
    public void getServiceActor() {
        // Act
        ServiceActorVO serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        SecurityContext.User actingUser = securityContext.getActingUser().get();
        SecurityContext.Organisation organisation = securityContext.getOrganisation().get();
        assertEquals(actingUser.getIdentifier(), serviceActor.getPersonIdentifier());
        assertEquals(actingUser.getGivenName().get(), serviceActor.getFirstName());
        assertEquals(actingUser.getSurname().get(), serviceActor.getLastName());
        assertEquals(actingUser.getCredentials().get().getAuthorizationCode().get(), serviceActor.getAuthorisationNumber());
        assertEquals(actingUser.getCredentials().get().getEducationCode().get(), serviceActor.getEducationCode());
        assertEquals(organisation.getIdentifier(), serviceActor.getOrganizationId());
        assertEquals("CVR", serviceActor.getOrganizationIdSource());
        assertEquals(organisation.getName().get(), serviceActor.getOrganizationName());
        assertEquals(UserType.HealthCareProfessional, serviceActor.getUserType());
    }

    @Test
    public void getServiceActor_unknownOrganisationIdFormat() {
        // Arrange
        SecurityContext.Organisation organisation = Mockito.mock(SecurityContext.Organisation.class);
        Mockito.when(organisation.getIdentifierFormat()).thenReturn(null);
        Mockito.when(securityContext.getOrganisation()).thenReturn(Optional.of(organisation));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INGEN_CVR.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_unknownPersonIdFormat() {
        // Arrange
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(SecurityContext.User.UserType.Citizen);
        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(null);
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUser));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_USER_NOT_CPR.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithPrincipalUser_fails() {
        // Arrange
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(SecurityContext.User.UserType.HealthcareProfessional);
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalAdmin() {
        // Act
        ServiceActorVO serviceActor = serviceActorProvider.getServiceActor(securityContext, true);

        // Assert
        assertEquals(UserType.Admin, serviceActor.getUserType());
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithoutAudience() {
        // Arrange
        SecurityContext.Ticket ticketWithoutAudience = createTicketWithoutAudience();
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticketWithoutAudience));

        // Act
        serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalWithInvalidAudience() {
        // Arrange
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs);

        // Act
        serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
    }

    @Test
    public void getServiceActor_forHealthCareProfessionalNotWhitelisted() {
        // Arrange
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList()); // Given cvr no longer whitelisted

        // Act
        serviceActorProvider.getServiceActor(securityContext, false); // Succeeds
    }

    @Test
    public void getServiceActor_forCitizen() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);

        // Act
        ServiceActorVO serviceActor = serviceActorProvider.getServiceActor(securityContext, false);

        // Assert
        assertEquals(UserType.Citizen, serviceActor.getUserType());
    }

    @Test
    public void getServiceActor_forCitizenNotWhitelisted() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList()); // Given cvr no longer whitelisted

        // Act
        serviceActorProvider.getServiceActor(securityContext, false); // succeeds
    }

    @Test
    public void getServiceActor_forCitizenAdmin_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, true);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithoutAudience_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        SecurityContext.Ticket ticketWithoutAudience = createTicketWithoutAudience();
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticketWithoutAudience));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INVALID_SECURITY_TICKET.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithInvalidAudience_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs);

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_WRONG_AUDIENCE.toString() + ", Gyldigt audience: http://gyldig", e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forCitizenWithPrincipalUser_fails() {
        // Arrange
        mockActingUser(SecurityContext.User.UserType.Citizen);
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_INGEN_DELEGERING.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserAdmin_fails() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, true);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_USERTYPE_ADMIN_NOT_ALLOWED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserNotWhitelisted_fails() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        serviceActorProvider = new ServiceActorProvider(audience, Collections.emptyList()); // Given cvr no longer whitelisted

        // Act
        try {
            serviceActorProvider.getServiceActor(securityContext, false);
            fail();
        } catch (Exception e) {
            // Assert
            assertEquals(FaultCode.AUTHORIZATION_ERROR_CVR_NOT_WHITELISTED.toString(), e.getMessage());
        }
    }

    @Test
    public void getServiceActor_forSystemUserWithInvalidAudience() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        serviceActorProvider = new ServiceActorProvider("http://gyldig", whitelistedCvrs);

        // Act
        serviceActorProvider.getServiceActor(securityContext, false); // succeeds
    }

    @Test
    public void getServiceActor_forSystemWithPrincipalUser() {
        // Arrange
        // No acting user, but an organization, means system user
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.empty());
        Mockito.when(securityContext.getPrincipalUser()).thenReturn(Optional.of(Mockito.mock(SecurityContext.User.class)));

        // Act
        serviceActorProvider.getServiceActor(securityContext, false);
    }

    private void mockActingUser(SecurityContext.User.UserType userType) {
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(userType);
        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(SecurityContext.User.PersonIdentifierFormat.CPR);
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUser));
    }

    private SecurityContext.Ticket createTicketWithoutAudience() {
        SecurityContext.Ticket ticket = Mockito.mock(SecurityContext.Ticket.class);
        Mockito.when(ticket.isValid()).thenReturn(true);
        Mockito.when(ticket.getAudience()).thenReturn(Optional.empty());
        return ticket;
    }

}