/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.CreateOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.GetOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.GetOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.AbstractSpringTest;
import dk.sundhedsdatastyrelsen.organdonor.service.business.OrganDonorServiceImpl;
import dk.sundhedsdatastyrelsen.organdonor.service.business.ServiceTestHelper;
import dk.sundhedsdatastyrelsen.organdonor.service.conf.TestDAOConfig;
import dk.sundhedsdatastyrelsen.organdonor.service.integrations.minlog.MinLogDataWrapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.context.SecurityContextProvider;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.FromHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.hl7cda.ToHL7CDAMapper;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.CprExists;
import org.flywaydb.core.Flyway;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.odr.PermissionType;
import org.hl7.v3.II;
import org.junit.Before;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.transaction.annotation.Transactional;

import javax.xml.bind.ValidationException;
import java.time.Instant;
import java.util.Optional;

import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.assertThat;

@Transactional("organDonorDataSourceTxManager")
public abstract class AbstractWebServiceTest extends AbstractSpringTest {
    public static final String personIdentifier = "0501792275";
    public static final String messageId = "8572";

    public static final String healtCareProffesionalIdentifier = "18374828181";
    public static final String adminUserIdentifier = "1837482818";

    final String AUTHORITYNAME = "CPR";
    final String ROOT = "1.2.208.176.1.2";

    final PermissionType PERMISSIONTYPE = PermissionType.FULL;
    final Boolean HEART = true;
    final Boolean LUNGS = true;
    final Boolean LIVER = false;
    final Boolean PANCREAS = true;
    final Boolean KIDNEYS = false;
    final Boolean CORNEA = false;
    final Boolean SMALLINTESTINE = true;
    final Boolean SKIN = false;
    final Boolean RELATIVEACCEPTANCE = true;

    @Autowired
    public SecurityContextProvider securityContextProvider;

    @Autowired
    public MinLogDataWrapper minLogDataWrapper;

    @Autowired
    TestDAOConfig daoConfig;

    @Autowired
    public OrganDonorDAO organDonorDAO;

    @Autowired
    protected ServiceTestHelper testHelper;

    @Autowired
    protected OrganDonorServiceImpl businessService;

    @Autowired
    private ToHL7CDAMapper toHL7CDAMapper;

    @Autowired
    private FromHL7CDAMapper fromHL7CDAMapper;

    @Autowired
    protected CprExists cprExists;

    private int port = 8080;

    @Value("${server.servlet.contextPath}")
    private String contextPath;
    private String uri;
    private String adminUri;
    private boolean initialized = false;
    private Flyway flyway;

    public static final dk.sundhedsdatastyrelsen.organdonor._2018._05._01.ObjectFactory odrFactory = new dk.sundhedsdatastyrelsen.organdonor._2018._05._01.ObjectFactory();

    @Before
    public void before() {
        flyway = daoConfig.getFlywayOrganDonor();

        uri = "http://localhost:" + port + contextPath + "/odr";
        adminUri = "http://localhost:" + port + contextPath + "/odrAdmin";

        if (!initialized) {
            initialized = true;
            flyway.clean();
            flyway.migrate();
        }
    }

    protected void createOrganDonorRegistration(String personIdentifier, String messageId) throws ValidationException {
        CreateOrganDonorRegistrationRequest createRequest = odrFactory.createCreateOrganDonorRegistrationRequest();
        createRequest.setId(createDefaultII(personIdentifier));
        createRequest.setOrganDonorRegistration(createDefaultOrganDonorRegistration(PERMISSIONTYPE, HEART, LUNGS, LIVER, PANCREAS, KIDNEYS, CORNEA, SMALLINTESTINE, SKIN, RELATIVEACCEPTANCE));
        businessService.createOrganDonorRegistration(fromHL7CDAMapper.mapOrganDonorFromRequest(createRequest), testHelper.getRequestContextExample(personIdentifier, messageId));
    }

    public OrganDonorRegistration createDefaultOrganDonorRegistration(PermissionType permissionType, boolean heart, boolean lungs,
                                                                      boolean liver, boolean pancreas, boolean kidneys,
                                                                      boolean cornea, boolean smallIntestine, boolean skin,
                                                                      boolean relativeAcceptance) {
        OrganDonorRegistration odr = new OrganDonorRegistration();
        odr.setPermissionType(permissionType);
        odr.setPermissionForHeart(heart);
        odr.setPermissionForLungs(lungs);
        odr.setPermissionForLiver(liver);
        odr.setPermissionForPancreas(pancreas);
        odr.setPermissionForKidneys(kidneys);
        odr.setPermissionForCornea(cornea);
        odr.setPermissionForSmallIntestine(smallIntestine);
        odr.setPermissionForSkin(skin);
        odr.setRequiresRelativeAcceptance(relativeAcceptance);
        return odr;
    }

    protected II createDefaultII(String personIdentifier) {
        return toHL7CDAMapper.createId(ROOT, personIdentifier, AUTHORITYNAME);
    }

    protected OrganDonorVO getOrganDonorRegistration(String personIdentifier) throws DAOException {
        Instant date = Instant.now();
        return organDonorDAO.getOrganDonorRegistration(personIdentifier, date);
    }

    protected void clearAllFields(OrganDonorRegistration organDonorRegistration) {
        organDonorRegistration.setRequiresRelativeAcceptance(true);
        organDonorRegistration.setPermissionForCornea(null);
        organDonorRegistration.setPermissionForHeart(null);
        organDonorRegistration.setPermissionForKidneys(null);
        organDonorRegistration.setPermissionForLiver(null);
        organDonorRegistration.setPermissionForLungs(null);
        organDonorRegistration.setPermissionForSkin(null);
        organDonorRegistration.setPermissionForPancreas(null);
        organDonorRegistration.setPermissionForSmallIntestine(null);
    }

    protected void checkError(Exception ex, String message) {
        String error = ex.toString();
        assertMessageContainsError(error, message);
    }

    protected void assertMessageContainsError(String message, String error) {
        if (error.contains("%")) {
            String errorWithoutWildcard = error.substring(0, error.indexOf("%"));
            assertThat(message, containsString(errorWithoutWildcard));
        } else {
            assertThat(message, containsString(error));
        }
    }

    protected CreateOrganDonorRegistrationRequest createCreateOrganDonorRegistrationRequest(String personIdentifier) {
        final CreateOrganDonorRegistrationRequest createRequest = odrFactory.createCreateOrganDonorRegistrationRequest();
        createRequest.setId(createDefaultII(personIdentifier));
        createRequest.setOrganDonorRegistration(createDefaultOrganDonorRegistration(PERMISSIONTYPE, HEART, LUNGS, LIVER, PANCREAS, KIDNEYS, CORNEA,
                SMALLINTESTINE, SKIN, RELATIVEACCEPTANCE));
        return createRequest;
    }

    protected GetOrganDonorRegistrationRequest createGetOrganDonorRegistrationRequest(String personIdentifier) {
        GetOrganDonorRegistrationRequest request = odrFactory.createGetOrganDonorRegistrationRequest();
        request.setId(createDefaultII(personIdentifier));
        return request;
    }

    protected OrganDonorRegistration getAllEntries(GetOrganDonorRegistrationResponse response) {
        OrganDonorRegistration organDonorRegistration = (OrganDonorRegistration) response.getClinicalDocument().getComponent().getStructuredBody().
                getComponents().get(0).getSection().getEntries().get(0).getObservation().getValues().get(0);
        return organDonorRegistration;
    }

    protected SecurityContext mockValidSecurityContext(SecurityContext.User.UserType userType, String cpr) {
        SecurityContext securityContext = Mockito.mock(SecurityContext.class);
        SecurityContext.Ticket ticket = Mockito.mock(SecurityContext.Ticket.class);
        Mockito.when(ticket.isValid()).thenReturn(true);
        Mockito.when(ticket.getAudience()).thenReturn(Optional.of("https://fsk"));
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
        SecurityContext.Message message = Mockito.mock(SecurityContext.Message.class);
        Mockito.when(securityContext.getMessage()).thenReturn(Optional.of(message));
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);
        Mockito.when(actingUser.getUserType()).thenReturn(userType);
        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(SecurityContext.User.PersonIdentifierFormat.CPR);
        Mockito.when(actingUser.getIdentifier()).thenReturn(cpr);
        Mockito.when(securityContext.getActingUser()).thenReturn(Optional.of(actingUser));
        SecurityContext.Organisation organisation = Mockito.mock(SecurityContext.Organisation.class);
        Mockito.when(organisation.getIdentifier()).thenReturn("31908574"); // Cvr is whitelisted in application.properties
        Mockito.when(organisation.getIdentifierFormat()).thenReturn(SecurityContext.Organisation.OrganisationIdentifierFormat.CVR);
        Mockito.when(securityContext.getOrganisation()).thenReturn(Optional.of(organisation));
        return securityContext;
    }

    protected SecurityContext mockValidSecurityContextWithoutUser() {
        SecurityContext securityContext = Mockito.mock(SecurityContext.class);
        SecurityContext.Ticket ticket = Mockito.mock(SecurityContext.Ticket.class);
        Mockito.when(ticket.isValid()).thenReturn(true);
        Mockito.when(ticket.getAudience()).thenReturn(Optional.of("https://fsk"));
        Mockito.when(securityContext.getTicket()).thenReturn(Optional.of(ticket));
        SecurityContext.Message message = Mockito.mock(SecurityContext.Message.class);
        Mockito.when(securityContext.getMessage()).thenReturn(Optional.of(message));
        SecurityContext.Organisation organisation = Mockito.mock(SecurityContext.Organisation.class);
        Mockito.when(organisation.getIdentifierFormat()).thenReturn(SecurityContext.Organisation.OrganisationIdentifierFormat.CVR);
        Mockito.when(securityContext.getOrganisation()).thenReturn(Optional.of(organisation));
        return securityContext;
    }
}
