/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.CreateOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.CreateOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OrganDonorWebService;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.ValidationErrors;
import org.hl7.odr.PermissionType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.util.Optional;

import static org.junit.Assert.*;

public class CreateOrganDonorRegistrationTest extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("organDonorWebService")
    private OrganDonorWebService serviceUnderTest;

    @Autowired
    @Qualifier("organDonorAdminWebService")
    private OrganDonorWebService adminServiceUnderTest;

    @Before
    public void setUp() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void createOrganDonorRegistration_WithInvalidAudience_GivesValidationError() throws Exception {
        Mockito.when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest(personIdentifier);

        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Forkert audience i sikkerhedsbillet, Gyldigt audience: https://fsk"));
        }
    }

    @Test
    public void createOrganDonorRegistration_validRequestFullPermission_OrganDonorCreated() throws Exception {
        final CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest(personIdentifier);

        final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final OrganDonorVO createdOrganDonor = getOrganDonorRegistration(request.getId().getExtension());
        assertNotNull(createdOrganDonor);

        assertNotNull(createdOrganDonor.getOrganDonorPID());
        assertEquals(1, createdOrganDonor.getVersion().intValue());
        assertEquals(request.getId().getExtension(), createdOrganDonor.getPersonIdentifier());
        assertEquals(request.getOrganDonorRegistration().isRequiresRelativeAcceptance(), createdOrganDonor.getRelativeAcceptanceRequired());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.FULL, createdOrganDonor.getPermissionType());
        assertEquals(RELATIVEACCEPTANCE, createdOrganDonor.getRelativeAcceptanceRequired());
    }

    @Test
    public void createOrganDonorRegistration_validRequestFullWithResearchPermission_OrganDonorCreated() throws Exception {
        final CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest(personIdentifier);
        request.getOrganDonorRegistration().setPermissionType(PermissionType.FULL_WITH_RESEARCH);

        final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final OrganDonorVO createdOrganDonor = getOrganDonorRegistration(request.getId().getExtension());
        assertNotNull(createdOrganDonor);

        assertNotNull(createdOrganDonor.getOrganDonorPID());
        assertEquals(1, createdOrganDonor.getVersion().intValue());
        assertEquals(request.getId().getExtension(), createdOrganDonor.getPersonIdentifier());
        assertEquals(request.getOrganDonorRegistration().isRequiresRelativeAcceptance(), createdOrganDonor.getRelativeAcceptanceRequired());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.FULL_WITH_RESEARCH, createdOrganDonor.getPermissionType());
        assertEquals(RELATIVEACCEPTANCE, createdOrganDonor.getRelativeAcceptanceRequired());
    }

    @Test
    public void adminCreateOrganDonorRegistration_validRequestFullPermission_OrganDonorCreated() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        final CreateOrganDonorRegistrationRequest request = createCreateOrganDonorRegistrationRequest(personIdentifier);

        final CreateOrganDonorRegistrationResponse response = adminServiceUnderTest.createOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final OrganDonorVO createdOrganDonor = getOrganDonorRegistration(request.getId().getExtension());
        assertNotNull(createdOrganDonor);

        assertNotNull(createdOrganDonor.getOrganDonorPID());
        assertEquals(1, createdOrganDonor.getVersion().intValue());
        assertEquals(request.getId().getExtension(), createdOrganDonor.getPersonIdentifier());
        assertEquals(request.getOrganDonorRegistration().isRequiresRelativeAcceptance(), createdOrganDonor.getRelativeAcceptanceRequired());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.FULL, createdOrganDonor.getPermissionType());
        assertEquals(RELATIVEACCEPTANCE, createdOrganDonor.getRelativeAcceptanceRequired());
    }

    @Test
    public void createOrganDonorRegistration_validRequestLimitedPermission_OrganDonorCreated() throws Exception {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);
        createRequest.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED);

        final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
        assertNotNull(response);

        final OrganDonorVO createdOrganDonor = getOrganDonorRegistration(createRequest.getId().getExtension());
        assertNotNull(createdOrganDonor);
        assertNotNull(createdOrganDonor.getOrganDonorPID());
        assertEquals(1, createdOrganDonor.getVersion().intValue());
        assertEquals(createRequest.getId().getExtension(), createdOrganDonor.getPersonIdentifier());
        assertEquals(createRequest.getOrganDonorRegistration().isRequiresRelativeAcceptance(), createdOrganDonor.getRelativeAcceptanceRequired());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.LIMITED, createdOrganDonor.getPermissionType());
        assertEquals(HEART, createdOrganDonor.getHeart());
        assertEquals(LUNGS, createdOrganDonor.getLungs());
        assertEquals(LIVER, createdOrganDonor.getLiver());
        assertEquals(PANCREAS, createdOrganDonor.getPancreas());
        assertEquals(KIDNEYS, createdOrganDonor.getKidneys());
        assertEquals(CORNEA, createdOrganDonor.getCorneas());
        assertEquals(SMALLINTESTINE, createdOrganDonor.getSmallIntestine());
        assertEquals(SKIN, createdOrganDonor.getSkin());
        assertEquals(RELATIVEACCEPTANCE, createdOrganDonor.getRelativeAcceptanceRequired());
    }

    @Test
    public void createOrganDonorRegistration_validRequestLimitedWithResearchPermission_OrganDonorCreated() throws Exception {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);
        createRequest.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED_WITH_RESEARCH);

        final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
        assertNotNull(response);

        final OrganDonorVO createdOrganDonor = getOrganDonorRegistration(createRequest.getId().getExtension());
        assertNotNull(createdOrganDonor);
        assertNotNull(createdOrganDonor.getOrganDonorPID());
        assertEquals(1, createdOrganDonor.getVersion().intValue());
        assertEquals(createRequest.getId().getExtension(), createdOrganDonor.getPersonIdentifier());
        assertEquals(createRequest.getOrganDonorRegistration().isRequiresRelativeAcceptance(), createdOrganDonor.getRelativeAcceptanceRequired());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.LIMITED_WITH_RESEARCH, createdOrganDonor.getPermissionType());
        assertEquals(HEART, createdOrganDonor.getHeart());
        assertEquals(LUNGS, createdOrganDonor.getLungs());
        assertEquals(LIVER, createdOrganDonor.getLiver());
        assertEquals(PANCREAS, createdOrganDonor.getPancreas());
        assertEquals(KIDNEYS, createdOrganDonor.getKidneys());
        assertEquals(CORNEA, createdOrganDonor.getCorneas());
        assertEquals(SMALLINTESTINE, createdOrganDonor.getSmallIntestine());
        assertEquals(SKIN, createdOrganDonor.getSkin());
        assertEquals(RELATIVEACCEPTANCE, createdOrganDonor.getRelativeAcceptanceRequired());
    }

    @Test
    public void createOrganDonorRegistration_personIdAlreadyExist_validationError() {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);

        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            assertNotNull(response);
            final CreateOrganDonorRegistrationResponse createTwice = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, ValidationErrors.PERSON_ID_EXISTS);
        }
    }

    @Test
    public void createOrganDonorRegistration_permissionTypeFullRelativeAcceptanceNull_validationError() {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);
        createRequest.getOrganDonorRegistration().setRequiresRelativeAcceptance(null);

        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, "relativeAcceptance");
        }
    }

    @Test
    public void createOrganDonorRegistration_limitedPermissionType_validationError() {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);
        createRequest.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED);
        clearAllFields(createRequest.getOrganDonorRegistration());

        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, ValidationErrors.VALUE_REQUIRED);
        }
    }

    @Test
    public void createOrganDonorRegistration_personIdNotDefined_validationError() {
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);

        createRequest.getId().setExtension(null);
        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, "Required value was missing");
        }
    }

    @Test
    public void createOrganDonorRegistration_healthCareProfessional_accessToCreateDenied() {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        final CreateOrganDonorRegistrationRequest createRequest = createCreateOrganDonorRegistrationRequest(personIdentifier);
        try {
            final CreateOrganDonorRegistrationResponse response = serviceUnderTest.createOrganDonorRegistration20180501(createRequest);
            fail("Expected validation error.");
        } catch (Exception e) {
            checkError(e, FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED.getMessage());
        }
    }
}
