/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.DeleteOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.DeleteOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OrganDonorWebService;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.ValidationErrors;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.time.Instant;
import java.util.Optional;

import static org.junit.Assert.*;

public class DeleteOrganDonorRegistrationTest extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("organDonorWebService")
    private OrganDonorWebService serviceUnderTest;

    @Autowired
    @Qualifier("organDonorAdminWebService")
    private OrganDonorWebService adminServiceUnderTest;

    @Before
    public void setUp() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void deleteOrganDonorRegistration_WithInvalidAudience_GivesValidationError() throws Exception {
        Mockito.when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);

        try {
            final DeleteOrganDonorRegistrationResponse response = serviceUnderTest.deleteOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Forkert audience i sikkerhedsbillet, Gyldigt audience: https://fsk"));
        }
    }

    @Test
    public void deleteOrganDonorRegistration_validRequest_odrDeleted() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);
        final DeleteOrganDonorRegistrationResponse response = serviceUnderTest.deleteOrganDonorRegistration20180501(request);
        assertNotNull(response);

        Instant date = Instant.now();
        OrganDonorVO returnOdr = organDonorDAO.getOrganDonorRegistration(request.getId().getExtension(), date);
        assertNull(returnOdr);
    }

    @Test
    public void adminDeleteOrganDonorRegistration_validRequest_odrDeleted() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);

        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);
        final DeleteOrganDonorRegistrationResponse response = adminServiceUnderTest.deleteOrganDonorRegistration20180501(request);
        assertNotNull(response);

        Instant date = Instant.now();
        OrganDonorVO returnOdr = organDonorDAO.getOrganDonorRegistration(request.getId().getExtension(), date);
        assertNull(returnOdr);
    }

    @Test
    public void deleteOrganDonorRegistration_odrNonExistingForPersonId_exception() {
        String personIdentifierNonExisting = "1235493821";//NON-EXISTING
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifierNonExisting);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);
        request.getId().setExtension(personIdentifierNonExisting);
        try {
            final DeleteOrganDonorRegistrationResponse response = serviceUnderTest.deleteOrganDonorRegistration20180501(request);
            fail("Should have caused error message: " + ValidationErrors.PERSON_ID_INVALID);
        } catch (Exception e) {
            checkError(e, ValidationErrors.PERSON_ID_INVALID);
        }
    }

    @Test
    public void deleteOrganDonorRegistration_personIdNotDefined_validationError() {
        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);
        request.getId().setExtension(null);

        try {
            final DeleteOrganDonorRegistrationResponse response = serviceUnderTest.deleteOrganDonorRegistration20180501(request);
            fail("Should have caused error message: Required value was missing");
        } catch (Exception e) {
            checkError(e, "Required value was missing");
        }
    }

    @Test
    public void deleteOrganDonorRegistration_invalidPersonId_validationError() {
        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);
        request.getId().setExtension("12415242"); //Person Id is not 10 digits
        try {
            final DeleteOrganDonorRegistrationResponse response = serviceUnderTest.deleteOrganDonorRegistration20180501(request);
            fail("Should have caused error message: " + ValidationErrors.PERSON_ID_INVALID);
        } catch (Exception e) {
            checkError(e, "Person identifier not Valid. Requires 10 digits");
        }
    }


    @Test
    public void deleteOrganDonorRegistration_healthCareProfessional_accessToDeleteDenied() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);

        DeleteOrganDonorRegistrationRequest request = createDeleteOrganDonorRequest(personIdentifier);

        try {
            serviceUnderTest.deleteOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED.getMessage());
        }
    }

    private DeleteOrganDonorRegistrationRequest createDeleteOrganDonorRequest(String personIdentifier) {
        DeleteOrganDonorRegistrationRequest deleteRequest = odrFactory.createDeleteOrganDonorRegistrationRequest();
        deleteRequest.setId(createDefaultII(personIdentifier));
        return deleteRequest;
    }
}
