/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.GetOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.GetOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.util.Properties;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OrganDonorWebService;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.exceptions.ValidationException;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.CprExists;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.RequestValidator;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.ValidationErrors;
import org.hl7.odr.OrganDonorRegistration;
import org.hl7.odr.PermissionType;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.util.List;
import java.util.Optional;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

public class GetOrganDonorRegistrationTest extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("organDonorWebService")
    private OrganDonorWebService serviceUnderTest;

    @Autowired
    @Qualifier("organDonorAdminWebService")
    private OrganDonorWebService adminServiceUnderTest;

    @Before
    public void setUp() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void getOrganDonorRegistration_WithInvalidAudience_GivesValidationError() throws Exception {
        when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        try {
            final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Forkert audience i sikkerhedsbillet, Gyldigt audience: https://fsk"));
        }
    }

    @Test
    public void getOrganDonorRegistration_WithUnknownAudienceForDGWSRequest_Succeeds() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        // User type citizen => IDWS
        when(securityContextProvider.getSecurityContext().getActingUser().get().getUserType()).thenReturn(SecurityContext.User.UserType.HealthcareProfessional);
        when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
        assertNotNull(response);
    }

    @Test
    public void getOrganDonorRegistration_validRequest_correctOdrData() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final OrganDonorRegistration organDonorRegistration = getAllEntries(response);
        assertNotNull(organDonorRegistration);
        assertEquals(PermissionType.FULL.value(), organDonorRegistration.getPermissionType().value());
        assertEquals(RELATIVEACCEPTANCE, organDonorRegistration.isRequiresRelativeAcceptance());
    }

    @Test
    public void adminGetOrganDonorRegistration_validRequest_correctOdrData() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);

        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        final GetOrganDonorRegistrationResponse response = adminServiceUnderTest.getOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final OrganDonorRegistration organDonorRegistration = getAllEntries(response);
        assertNotNull(organDonorRegistration);
        assertEquals(PermissionType.FULL.value(), organDonorRegistration.getPermissionType().value());
        assertEquals(RELATIVEACCEPTANCE, organDonorRegistration.isRequiresRelativeAcceptance());
    }

    @Test
    public void getOrganDonorRegistration_validRequest_validatesClinicalDocument() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
        assertNotNull(response);

        RequestValidator validator = new RequestValidator(cprExists, new Properties());
        validator.validateWithoutThrow(response);

        final int errorCount = validator.getValidationErrors().size();
        assertEquals(errorCount + " validation errors were found: " + validator.getCombinedErrorText(), 0, errorCount);
    }

    @Test
    public void getOrganDonorRegistration_ageTooLowRequest_errorShouldBeReturned() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);

        final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
        assertNotNull(response);

        final String expectedError = "Age validation failed";
        CprExists mockedCprExists = mock(CprExists.class);
        doThrow(new ValidationException(expectedError)).when(mockedCprExists).validateAge(anyString());

        RequestValidator validator = new RequestValidator(mockedCprExists, new Properties());
        validator.validateWithoutThrow(response);

        final List<RequestValidator.RequestValidatorError> errors = validator.getValidationErrors();
        assertEquals(1, errors.size());
        assertEquals(expectedError, errors.get(0).description);
    }

    @Test
    public void getOrganDonorRegistration_personIdNonExist_validationError() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        final String personIdentifier = "1293829102"; //NON-EXISTING
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        GetOrganDonorRegistrationRequest request = createGetOrganDonorRegistrationRequest(personIdentifier);
        request.getId().setExtension(personIdentifier);
        request.setId(request.getId());

        try {
            final GetOrganDonorRegistrationResponse response = serviceUnderTest.getOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, ValidationErrors.PERSON_ID_INVALID);
        }
    }


}










