/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.HasOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.HasOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OrganDonorWebService;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.util.Optional;

import static org.junit.Assert.*;

public class HasOrganDonorRegistrationTest extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("organDonorWebService")
    private OrganDonorWebService serviceUnderTest;


    @Before
    public void setUp() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void hasOrganDonorRegistration_WithInvalidAudience_GivesValidationError() throws Exception {
        Mockito.when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        try {
            final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Forkert audience i sikkerhedsbillet, Gyldigt audience: https://fsk"));
        }
    }

    @Test
    public void hasOrganDonorRegistration_WithUnknownAudienceForDGWSRequest_Succeeds() throws Exception {
        // User type citizen => IDWS
        Mockito.when(securityContextProvider.getSecurityContext().getActingUser().get().getUserType()).thenReturn(SecurityContext.User.UserType.HealthcareProfessional);
        Mockito.when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);
        assertNotNull(response);
    }

    @Test
    public void hasOrganDonorRegistration_WithNonWhitelistedCvrInLevel3IdCard_GivesValidationError() throws Exception {
        SecurityContext securityContext = mockValidSecurityContextWithoutUser(); // No user means level 3 id card
        Mockito.when(securityContext.getOrganisation().get().getIdentifier()).thenReturn("1234"); // 1234 not whitelisted
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        try {
            final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains(FaultCode.AUTHORIZATION_ERROR_CVR_NOT_WHITELISTED.toString()));
        }
    }

    @Test
    public void hasOrganDonorRegistration_WithWhitelistedCvrInLevel3IdCard_trueIsReturned() throws Exception {
        SecurityContext securityContext = mockValidSecurityContextWithoutUser(); // No user means level 3 id card
        Mockito.when(securityContext.getOrganisation().get().getIdentifier()).thenReturn("31908574"); // Cvr is whitelisted in application.properties
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertNotNull(response);
        assertTrue(response.isRegistrationExists());
    }

    @Test
    public void hasOrganDonorRegistration_WithNonWhitelistedCvrNotLevel3IdCard_trueIsReturned() throws Exception {
        Mockito.when(securityContextProvider.getSecurityContext().getOrganisation().get().getIdentifier()).thenReturn("1234"); // 1234 not whitelisted
        createOrganDonorRegistration(personIdentifier, messageId);
        final HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertNotNull(response);
        assertTrue(response.isRegistrationExists());
    }

    @Test
    public void hasOrganDonorRegistration_odrExists_trueIsReturned() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertNotNull(response);
        assertTrue(response.isRegistrationExists());
    }

    @Test
    public void hasOrganDonorRegistration_odrNotExists_falseIsReturned() throws Exception {
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);

        final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);

        assertNotNull(response);
        assertFalse(response.isRegistrationExists());
    }

    @Test
    public void hasOrganDonorRegistration_personIdNotDefined_validationError() throws Exception {
        HasOrganDonorRegistrationRequest request = createHasOrganDonorRegistrationRequest(personIdentifier);
        request.getId().setExtension(null);

        try {
            final HasOrganDonorRegistrationResponse response = serviceUnderTest.hasOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, "Required value was missing");
        }
    }

    private HasOrganDonorRegistrationRequest createHasOrganDonorRegistrationRequest(String personIdentifier) {
        HasOrganDonorRegistrationRequest hasRequest = odrFactory.createHasOrganDonorRegistrationRequest();
        hasRequest.setId(createDefaultII(personIdentifier));
        return hasRequest;
    }

}
