/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.service.webservice.integration;

import dk.sds.nsp.security.SecurityContext;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.UpdateOrganDonorRegistrationRequest;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.UpdateOrganDonorRegistrationResponse;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import dk.sundhedsdatastyrelsen.organdonor.service.util.FaultCode;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.OrganDonorWebService;
import dk.sundhedsdatastyrelsen.organdonor.service.webservice.validation.ValidationErrors;
import org.hl7.odr.PermissionType;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.time.Instant;
import java.util.Optional;

import static org.junit.Assert.*;

public class UpdateOrganDonorRegistrationTest extends AbstractWebServiceTest {

    @Autowired
    @Qualifier("organDonorWebService")
    private OrganDonorWebService serviceUnderTest;

    @Autowired
    @Qualifier("organDonorAdminWebService")
    private OrganDonorWebService adminServiceUnderTest;

    @Before
    public void setUp() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.Citizen, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
    }

    @Test
    public void updateOrganDonorRegistration_WithInvalidAudience_GivesValidationError() throws Exception {
        Mockito.when(securityContextProvider.getSecurityContext().getTicket().get().getAudience()).thenReturn(Optional.of("blah"));
        final UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);

        try {
            final UpdateOrganDonorRegistrationResponse response = serviceUnderTest.updateOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Forkert audience i sikkerhedsbillet, Gyldigt audience: https://fsk"));
        }
    }

    @Test
    public void updateOrganDonorRegistration_updatePermissionType_newPermissionType() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);
        Instant beforeDate = Instant.now();
        OrganDonorVO beforeTest = organDonorDAO.getOrganDonorRegistration(personIdentifier, beforeDate);
        assertNotNull(beforeTest);

        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);
        request.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED);  //Updated PermissionType

        final UpdateOrganDonorRegistrationResponse response = serviceUnderTest.updateOrganDonorRegistration20180501(request);
        assertNotNull(response);

        Instant afterDate = Instant.now();
        OrganDonorVO afterTest = organDonorDAO.getOrganDonorRegistration(request.getId().getExtension(), afterDate);
        assertNotNull(afterTest);
        assertNotNull(afterTest.getOrganDonorPID());
        assertEquals(beforeTest.getPersonIdentifier(), afterTest.getPersonIdentifier());
        assertNotEquals(beforeTest.getPermissionType(), afterTest.getPermissionType());
        assertTrue(afterTest.getValidFrom().isBefore(afterTest.getValidTo()));
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.LIMITED, afterTest.getPermissionType());
        assertEquals(2, afterTest.getVersion().intValue());
    }

    @Test
    public void adminUpdateOrganDonorRegistration_updatePermissionType_newPermissionType() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);
        Instant beforeDate = Instant.now();
        OrganDonorVO beforeTest = organDonorDAO.getOrganDonorRegistration(personIdentifier, beforeDate);
        assertNotNull(beforeTest);

        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);
        request.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED);  //Updated PermissionType

        final UpdateOrganDonorRegistrationResponse response = adminServiceUnderTest.updateOrganDonorRegistration20180501(request);
        assertNotNull(response);

        Instant afterDate = Instant.now();
        OrganDonorVO afterTest = organDonorDAO.getOrganDonorRegistration(request.getId().getExtension(), afterDate);
        assertNotNull(afterTest);
        assertNotNull(afterTest.getOrganDonorPID());
        assertEquals(beforeTest.getPersonIdentifier(), afterTest.getPersonIdentifier());
        assertNotEquals(beforeTest.getPermissionType(), afterTest.getPermissionType());
        assertTrue(afterTest.getValidFrom().isBefore(afterTest.getValidTo()));
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.LIMITED, afterTest.getPermissionType());
        assertEquals(2, afterTest.getVersion().intValue());
    }

    @Test
    public void updateOrganDonorRegistration_updateOdrValues_odrValuesUpdated() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        Instant beforeDate = Instant.now();
        OrganDonorVO beforeTest = organDonorDAO.getOrganDonorRegistration(personIdentifier, beforeDate);
        assertNotNull(beforeTest);

        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);
        request.getOrganDonorRegistration().setPermissionType(PermissionType.DONT_KNOW);
        request.getOrganDonorRegistration().setPermissionForHeart(false);
        request.getOrganDonorRegistration().setPermissionForLungs(true);
        request.getOrganDonorRegistration().setPermissionForLiver(true);
        request.getOrganDonorRegistration().setPermissionForKidneys(true);
        request.getOrganDonorRegistration().setRequiresRelativeAcceptance(false);

        final UpdateOrganDonorRegistrationResponse response = serviceUnderTest.updateOrganDonorRegistration20180501(request);
        assertNotNull(response);

        Instant afterDate = Instant.now();
        OrganDonorVO afterTest = organDonorDAO.getOrganDonorRegistration(request.getId().getExtension(), afterDate);
        assertNotNull(afterTest);
        assertNotNull(afterTest.getOrganDonorPID());
        assertEquals(beforeTest.getPersonIdentifier(), afterTest.getPersonIdentifier());
        assertEquals(2, afterTest.getVersion().intValue());

        assertNotEquals(beforeTest.getPermissionType(), afterTest.getPermissionType());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.FULL, beforeTest.getPermissionType());
        assertEquals(dk.sundhedsdatastyrelsen.organdonor.domain.PermissionType.DONT_KNOW, afterTest.getPermissionType());
    }

    @Test
    public void updateOrganDonorRegistration_noOdrValuesSpecified_validationError() throws Exception {
        createOrganDonorRegistration(personIdentifier, messageId);

        Instant beforeDate = Instant.now();
        OrganDonorVO beforeTest = organDonorDAO.getOrganDonorRegistration(personIdentifier, beforeDate);
        assertNotNull(beforeTest);

        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);
        request.getOrganDonorRegistration().setPermissionType(PermissionType.LIMITED);
        clearAllFields(request.getOrganDonorRegistration());

        try {
            final UpdateOrganDonorRegistrationResponse response = serviceUnderTest.updateOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, ValidationErrors.VALUE_REQUIRED);
        }
    }

    @Test
    public void updateOrganDonorRegistration_healthCareProfessional_accessToCreateDenied() throws Exception {
        SecurityContext securityContext = mockValidSecurityContext(SecurityContext.User.UserType.HealthcareProfessional, personIdentifier);
        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(securityContext);
        createOrganDonorRegistration(personIdentifier, messageId);
        UpdateOrganDonorRegistrationRequest request = createUpdateOrganDonorRegistrationRequest(personIdentifier);

        try {
            final UpdateOrganDonorRegistrationResponse response = serviceUnderTest.updateOrganDonorRegistration20180501(request);
            fail("Expected validation error.");
        } catch (Exception e) {
            System.out.println("Actual exception message: " + e.getMessage());
            checkError(e, FaultCode.AUTHORIZATION_ERROR_PERSON_IDENTIFIER_ACCESS_DENIED.getMessage());
        }
    }


    private UpdateOrganDonorRegistrationRequest createUpdateOrganDonorRegistrationRequest(String personIdentifier) {
        UpdateOrganDonorRegistrationRequest request = odrFactory.createUpdateOrganDonorRegistrationRequest();
        request.setId(createDefaultII(personIdentifier));
        request.setOrganDonorRegistration(createDefaultOrganDonorRegistration(PERMISSIONTYPE, HEART, LUNGS, LIVER, PANCREAS, KIDNEYS, CORNEA,
                SMALLINTESTINE, SKIN, RELATIVEACCEPTANCE));
        return request;
    }
}
