/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.organdonor.datacreator.business;

import dk.sundhedsdatastyrelsen.organdonor.datacreator.DataCreatorProperties;
import dk.sundhedsdatastyrelsen.organdonor.datacreator.util.OrganDonorDataUtil;
import dk.sundhedsdatastyrelsen.organdonor.datacreator.util.PersonFileUtil;
import dk.sundhedsdatastyrelsen.organdonor.datacreator.util.RandomDataUtil;
import dk.sundhedsdatastyrelsen.organdonor.dao.OrganDonorDAO;
import dk.sundhedsdatastyrelsen.organdonor.dao.exceptions.DAOException;
import dk.sundhedsdatastyrelsen.organdonor.domain.OrganDonorVO;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.stream.Collectors;

@Component
public class OrganDonorDataCreator {

    private static final Logger log = LogManager.getLogger(OrganDonorDataCreator.class);

    private final DataCreatorProperties dataCreatorProperties;
    private final OrganDonorDAO organDonorDAO;

    private int createRecordCount = 0;
    private int updateRecordCount = 0;

    public OrganDonorDataCreator(DataCreatorProperties dataCreatorProperties, OrganDonorDAO organDonorDAO) {
        this.dataCreatorProperties = dataCreatorProperties;
        this.organDonorDAO = organDonorDAO;
    }

    public void start() throws DAOException {
        // Get list of person identifiers
        final List<String> personIdentifiers = PersonFileUtil.parsePersonFile(dataCreatorProperties.getPersonFileName());

        // Create data objects for all person identifiers
        final List<OrganDonorVO> organDonorData = personIdentifiers.stream()
                .map(personIdentifier ->
                        OrganDonorDataUtil.getNewOrganDonor(
                                personIdentifier,
                                dataCreatorProperties.getOdr().getDataTimeIntervalStart(),
                                dataCreatorProperties.getOdr().getDataTimeIntervalEnd()))
                .collect(Collectors.toList());

        // Create data in DB
        for (OrganDonorVO organDonorVO : organDonorData) {
            final int pid = (int) createOrganDonor(organDonorVO);
            organDonorVO.setOrganDonorPID(pid);
        }

        // Update data in DB
        for (OrganDonorVO organDonorVO : organDonorData) {
            updateOrganDonorRecursively(organDonorVO);
        }

        log.info("Number of first records: " + createRecordCount);
        log.info("Number of update records: " + updateRecordCount);
        log.info("Total records: " + (createRecordCount + updateRecordCount));
    }

    private long createOrganDonor(OrganDonorVO organDonor) throws DAOException {
        final long pid = organDonorDAO.createOrganDonorRegistration(organDonor, organDonor.getModifiedDateTime());
        createRecordCount++;
        return pid;
    }

    private void updateOrganDonorRecursively(OrganDonorVO oldOrganDonor) throws DAOException {
        // Run if the random decision returns true
        if (RandomDataUtil.getRandomDecision(dataCreatorProperties.getOdr().getChanceOfUpdates())) {
            final OrganDonorVO newOrganDonor = OrganDonorDataUtil.getUpdatedOrganDonorData(oldOrganDonor);

            // The method implicitly sets the organDonorPID on the newOrganDonor-object.... :/
            organDonorDAO.updateOrganDonorRegistration(oldOrganDonor.getOrganDonorPID(),
                    newOrganDonor, newOrganDonor.getModifiedDateTime());

            updateRecordCount++;
            updateOrganDonorRecursively(newOrganDonor);
        }
    }
}
